indexing

   title:      "Key priority queues: dispensers with two element entries",
               "in which only the entry with the largest key is accessed";
   cluster:    "containers";
   project:    "Visual Eiffel";
   date:       "$Date: 2005/02/15 16:21:21 $";
   revision:   "$Revision: 1.2 $";
   copyright:  "Copyright (C) 1996-2005 Object Tools Group";
   license:    "http://visual-eiffel.com/license";
   disclaimer: "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class KEY_PRIORITY_QUEUE_[G, H -> COMPARABLE]

   inherit
      KEYED_DISPENSER_[G, H]
         redefine
            pool
         end


   feature -- Operations

      put (x: G; k: H) is
            -- adds the new entry into the dispenser
         require else
            valid_item:  x /= Void
            valid_key:   k /= Void

         do
            if pool = Void then
               !!pool.make (false)
               cs := pool.cursor
            end

            pool.put (x, k)

            key  := pool.last_key_head
            item := pool.item_at_last_key_head
         end

      remove is
            -- deletes the "active entry" and sets 'item' to the element
            -- associated with the largest key among those still existing in
            -- the dispenser after the removal and 'key' to this key; if there
            -- are no more entries then 'item' and 'key' are both set to Void
         do
            if not empty then

               check item /= Void and then key /= Void end
               pool.remove (item, key)

               key  := pool.last_key_head
               item := pool.item_at_last_key_head

            else
               fault (f_remove, errhlg.sc_is_empty)
            end

         ensure then
            well_done: cs.is_finished
         end


   feature -- Queries

      table_of_elements: TABLE_[G, H] is
            -- delivers all the entries of the dispenser in the order in
            -- which they would be repeatedly removed if it really happened
         local
            item_stack:    STACK_[G]
            key_stack:     STACK_[H]
            current_key:   H
            pop_stack_out,
            push_on_stack: BOOLEAN

         do
            !!Result.make (false)
            if not empty then
               check pool /= Void and then count > 0  end

               from
                  !!item_stack
                  !!key_stack

                  -- allocate slots in advance to avoid possible
                  -- frequent automatic reallocation ...
                  Result.allocate_manually (count)

                  cs.last
               until
                  Result.count = count
               loop
                  push_on_stack := true
                  pop_stack_out := false

                  if cs.is_finished then
                     pop_stack_out := true
                     push_on_stack := false
                  else
                     current_key := pool.key_at (cs)

                     -- all keys and items go into the stacks
                     -- to reverse the order of duplicates ...
                     if not (key_stack.empty or else
                           current_key.is_equal (key_stack.item)) then

                        pop_stack_out := true
                     end
                  end

                  if pop_stack_out then
                     check not key_stack.empty end

                     from
                     invariant
                        key_stack.count = item_stack.count
                     until
                        key_stack.empty
                     loop
                        Result.put (item_stack.item, key_stack.item)
                        item_stack.remove
                        key_stack.remove
                     end
                  end

                  if push_on_stack then
                     item_stack.put (pool.item_at (cs))
                     key_stack.put (current_key)
                     cs.back
                  end
               end

               -- return to automatic memory reallocation
               Result.allocate_automatically
            end

         ensure then
            well_done: cs /= Void implies cs.is_finished
         end


   feature {NONE} -- Implementation

      pool: SORTED_TABLE_[G, H] -- the real receptacle for entries to be stored

end
