indexing

   description: "Sequences of values, all of the same type or of a %
                %conforming one, accessible through integer indices %
                %in a contiguous interval"
   cluster:     "kernel";
   project:     "Eiffel Kernel Library";
   date:        "$Date: 2005/02/01 10:20:11 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class ARRAY [G]

inherit

   ANY
      redefine
         copy,
         is_equal
      end

creation

   make,
   make_from_array

feature {NONE} -- Creation

   make (min_index, max_index: INTEGER) is
         -- Allocate array to hold values for indexes in
         -- `min_index' .. `max_index'.
         -- Set all values to default.
         -- (Make array empty if `maxindex' = `minindex' - 1).
      require
         valid_bounds: min_index <= max_index + 1
      do
         c_array_make (max_index - min_index + 1, min_index)
      ensure
         lower_set: lower = min_index
         upper_set: upper = max_index
         items_set: -- all_default
      end

feature -- Initialization

   make_from_array (a: ARRAY [G]) is
         -- Initialize from the items of `a'.
      require
         a_not_void: a /= Void
      do
         make (1, 0)
         copy (a)
      ensure
         initialized: same_array (a)
      end

feature {NONE} -- Implementation

   frozen area: POINTER
         -- Pointer to array's data
         -- (for internal use only)

   frozen b_area: POINTER
         -- Pointer to array's data
         -- (for internal use only)

feature -- Measurement

   count: INTEGER
         -- Number of available indices
         -- ensure
         --    consistent_with_bounds: Result = upper - lower + 1

feature -- Basic specifiers

   item (i: INTEGER): G is
         -- Item at index `i'.
      require
         valid_index: valid_index (i)
      external "CWC"
         alias "_ARRAY_item"
      end

   lower: INTEGER
         -- Minimum index

   upper: INTEGER is
         -- Maximum index
      do
         Result := lower + count - 1
      end

feature -- Access

   infix "@" (i: INTEGER): G is
         -- Item at index `i'.
      require
         valid_index: valid_index (i)
      external "CWC"
         alias "_ARRAY_item"
      ensure
         definition: Result = item (i)
      end

   array: ARRAY [G] is
         -- New ARRAY having the same items as `Current'
      do
         create Result.make_from_array (Current)
      ensure
         array_not_void: Result /= Void
         result_type: -- Result type is ARRAY [like item]
         same_count: Result.count = count
         same_indexes: Result.lower = lower
         first_item: count > 0 implies Result.item (lower) = item (lower)
         recurse: count > 0 implies
            Result.subarray (lower + 1, upper).is_equal
            (subarray (lower + 1, upper).array)
      end


feature -- Comparison

   is_equal (other: like Current): BOOLEAN is
         -- Do both arrays have the same `lower', `upper', and items?
         -- (Redefined from GENERAL.)
      do
         Result :=
            same_type (other) and then
            lower = other.lower and then
            same_items (other)
      ensure then
         definition: Result = (same_type (other) and then
            (lower = other.lower) and then same_items (other))
      end

   same_array (other: ARRAY [G]): BOOLEAN is
         -- Do `Current' and `other' have the same items
         -- with the same indices?
      require
         other_not_void: other /= Void
      do
         Result := array.is_equal (other.array)
      ensure
         definition: Result = array.is_equal (other.array)
      end

feature -- Status report

   all_default: BOOLEAN is
         -- Do all items have their type's default value?
      external "CWC"
         alias "_ARRAY_all_cleared"
      ensure
         definition: Result = (
            count = 0 or else (
             (item(upper) = Void or else
              item(upper) = item(upper).default)
              and subarray(lower, upper - 1).all_default
            ))
      end

   is_empty: BOOLEAN is
          -- Is array empty?
       do
          Result := count = 0
       ensure
          definition: Result = (count = 0)
       end

   same_items (other: like Current): BOOLEAN is
         -- Do both arrays have the same items?
      require
         other_not_void: other /= Void
      local
         i: INTEGER
         j: INTEGER
         n: INTEGER
      do
         Result := True
         if count = other.count then
            from
               i := lower
               j := other.lower
               n := upper
            until
               i > n or else not Result
            loop
               Result := item (i) = other.item (j)
               i := i + 1
               j := j + 1
            end
         end
      ensure
         definition: Result =
            ((count = other.count) and then
            (count = 0 or else
            (item (upper) = other.item (other.upper) and
              subarray (lower, upper-1).same_items
               (other.subarray (other.lower, other.upper - 1)))))
      end

   valid_index (i: INTEGER): BOOLEAN is
         -- Is `i' within bounds?
      do
         Result := lower <= i and then i <= upper
      ensure
         definition: Result = ((lower <= i) and (i <= upper))
      end

feature -- Element change

   clear_all is
         -- Set every item to its default value.
      external "CWC"
         alias "_ARRAY_clear_all"
      ensure
         stable_upper: upper = old upper
         stable_lower: lower = old lower
         default_items: -- all_default
      end

   force (v: G; i: INTEGER) is
         -- Make `v' the item at index `i', enlarging the
         -- array if necessary.
      do
         if i < lower then
            resize (i, upper)
         elseif i > upper then
            resize (lower, i)
         end
         put (v, i)
      ensure
         new_lower: lower = i.min (old lower)
         new_upper: upper = i.max (old upper)
         new_item: item (i) = v
         new_low_items:
            -- i < old lower implies
            --  subarray (i + 1, old lower - 1).all_default
         new_high_items:
            -- i > old upper implies
            --  subarray (old upper + 1, i - 1).all_default
         between_lower_and_i:
            subarray (old lower, ((i-1).max (old lower-1))
             .min (old upper)).same_items
             (old subarray (lower, ((i-1).max (lower-1)).min (upper)))
         between_i_and_upper:
            subarray (((i+1).min (old upper+1))
             .max (old lower), old upper).same_items
             (old subarray (((i+1).min (upper+1)).max (lower), upper))
      end

   put (v: G; i: INTEGER) is
         -- Make 'v' the item at index `i'.
      require
         valid_index: valid_index (i)
      do
         c_array_put (i, v)
      ensure
         stable_lower: lower = old lower
         stable_upper: upper = old upper
         new_item: item (i) = v
         stable_before_i:
          subarray (lower, i - 1).is_equal
             (old subarray (lower, i - 1))
         stable_after_i:
          subarray (i + 1, upper).is_equal
             (old subarray (i + 1, upper))
      end

feature -- Resizing

   resize (min_index, max_index: INTEGER) is
         -- Resize to bounds `min_index' and `max_index'.
         -- Do not lose any item whose index is in both
         -- `lower..upper' and `min_index..max_index'.
      require
         valid_bounds: min_index <= max_index + 1
      do
         c_array_resize (max_index - min_index + 1, min_index)
      ensure
         new_lower: lower = min_index
         new_upper: upper = max_index
         default_if_too_small:
            -- min_index < old lower implies subarray
            --    (min_index, max_index.min (old lower - 1)).all_default
         default_if_too_large:
            -- max_index > old upper implies subarray
            --    (min_index.max (old upper + 1), max_index).all_default
         stable_in_intersection:
            max_index < old lower or min_index > old upper or else
             subarray (min_index.max (old lower).min (max_index + 1),
              max_index.min (old upper).max(min_index - 1)).is_equal
             (old subarray (min_index.max (lower).min(upper + 1),
              max_index.min (upper).max(lower - 1)))
      end

feature -- Duplication

   copy (other: like Current) is
         -- Reinitialize by copying all the items of `other'.
         -- (Also used by `clone'.)
         -- (Redefined from GENERAL.)
      external "CWC"
         alias "_ARRAY_copy"
      end

   subarray (min, max: INTEGER): ARRAY [G] is
      -- New array consisting of items at indexes
      -- in `min .. max'
      require
         min_large_enough: lower <= min
         max_small_enough: max <= upper
         valid_bounds: min <= max + 1
      do
         Result := clone (Current)
         Result.resize (min, max)
      ensure
         subarray_not_void: Result /= Void
         lower_set: Result.lower = min
         upper_set: Result.upper = max
         items_set:
            Result.count > 0 implies
               (Result.item (max) = item (max) and
               Result.subarray (min, max - 1).is_equal
                  (subarray (min, max - 1)))
      end

feature -- ELKS 1995 compatibility

   entry (i: INTEGER): G is
      obsolete "ELKS 1995 compatibility: use 'item (i)' instead"
         -- Entry at index 'i', if in index interval.
         -- (Synonym for 'item' and 'infix "@"')
      require
         good_key: valid_index (i)
      external "CWC"
         alias "_ARRAY_item"
      end

   enter (v: G; i: INTEGER) is
      obsolete "ELKS 1995 compatibility: use 'put (i)' instead"
         -- Replace 'i-th' entry, if in index interval, by 'v'.
         -- (Synonym for 'put')
      require
         good_key: valid_index (i)
      do
         put (v, i)
      ensure
         inserted: item (i) = v
      end

   to_c: POINTER is
         -- Address of actual sequence of values,
         -- for passing to external (non-Eiffel) routines
      external "CWC"
         alias "_ARRAY_to_external"
      end

feature -- Visual Eiffel 3.2 compatibility

   append (other: ARRAY [G]) is
         -- Append all items of 'other', if not void, at end
      local
         old_count: INTEGER
         cnt: INTEGER
         oth: ARRAY [G]
      do
         if other /= Void then
            cnt := other.count
            if cnt > 0 then
               oth := other
               if oth = Current then
                  oth := clone (oth)
               end
               old_count := count
               c_array_resize (count + cnt, lower)
               array_mem_copy (to_c + old_count * size_item, oth.to_c, cnt * size_item)
            end
         end
      ensure
         higher_count: (other  = Void implies count  = old count) and then
                       (other /= Void implies count >= old count)
      end

   from_c (ptr: POINTER; cnt: INTEGER) is
         -- Creates an array from its external representation
      require
         non_void_ptr: ptr /= default_pointer
         valid_cnt: cnt >= 0
      do
         make (1, cnt)
         array_mem_copy (to_c, ptr, cnt * size_item)
      ensure
         new_lower: lower = 1
         new_count: count = cnt
      end

feature -- Eiffel/S 1.3 compatibility

   to_external: POINTER is
      obsolete "Eiffel/S 1.3 compatibility: use 'to_c' instead"
         -- Pointer to actual storage area
      external "CWC"
         alias "_ARRAY_to_external"
      end

   insert (element: G; index: INTEGER) is
         -- Insert 'element' after position 'index'.
         -- Element at position 'upper' will be lost!
         -- Note: Insertion is AFTER 'index'!
      require
         inside_bounds: index >= lower - 1 and then index < upper
      local
         real_index: INTEGER
      do
         real_index := index + 1
         if real_index /= upper then
            move (real_index, upper - 1, 1)
         end
         put (element, real_index)
      end

   move (lower_index, upper_index, distance: INTEGER) is
         -- Move Range 'lower_index' .. 'upper_index' by 'distance'
         -- positions. Neg. distance moves towards lower indices.
         -- Free places get default values
      require
         non_empty_range: lower_index <= upper_index
         inside_bounds_before: lower       <= lower_index
                                       and then
                                upper_index <= upper
         inside_bounds_after: lower       <= lower_index + distance
                                       and then
                                upper_index + distance <= upper
      do
         c_move (lower_index - lower, upper_index - lower_index + 1, distance)
      end

   remove (index: INTEGER) is
         -- Remove element at position 'index'.
         -- Elements after 'index' will move
         -- one position left (=towards lower indices).
         -- Element at position 'upper' is set to default value.
         -- Note: Position 'upper' will be set to
         -- appropriate default value
      require
         inside_bounds: index >= lower and then index <= upper
      do
         if index /= upper then
            move (index + 1, upper, -1)
         end
         c_put_default (upper)
      end

   size: INTEGER is
      obsolete "Eiffel/S 1.3 compatibility: use 'count' instead"
         -- Length of index interval
      do
         Result := count
      end

   reindex (new_lower: INTEGER) is
         -- Change index interval so that
         -- it starts at 'new_lower'.
         -- No elements will be lost
      external "CWC"
         alias "_ARRAY_reindex"
      ensure
         bounds_adjusted: lower = new_lower and then
                           upper = new_lower + count - 1
      end

   wipe_out is
      obsolete "Eiffel/S 1.3 compatibility: use 'resize (lower, lower - 1)' instead"
         -- Empty the array, discard all items
      do
         resize (lower, lower - 1)
      end

   empty: BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility: use 'is_empty' instead"
         -- Is array empty?
      do
         Result := is_empty
      end

   all_cleared: BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility: use 'all_default' instead"
         -- Are all items set to default values?
      do
         Result := all_default
      end

feature {NONE} -- Implementation

   frozen pcount: INTEGER
      -- Physical number of item in the array

   frozen size_item: INTEGER
      -- Size of array item

   c_array_make (cnt: INTEGER; lwr: INTEGER) is
      external "CWC"
         alias "_ARRAY_make"
      end

   c_array_put (i: INTEGER; v: G) is
      external "CWC"
         alias "_ARRAY_put"
      end

   c_array_resize (cnt: INTEGER; lwr: INTEGER) is
      external "CWC"
         alias "_ARRAY_resize"
      end

   array_mem_copy (dest, source: POINTER; bytes_number: INTEGER) is
      external "C"
         alias "_mem_copy"
      end

   c_move (lower_index, number, distance: INTEGER) is
         -- Eiffel/S 1.3 compatibility
      external "CWC"
         alias "_ARRAY_move"
      end

   c_put_default (index: INTEGER) is
         -- Eiffel/S 1.3 compatibility
      external "CWC"
         alias "_ARRAY_put_default"
      end

invariant

   valid_bounds: lower <= upper + 1

   consistent_with_bounds: count = upper - lower + 1
         -- postcondition of `count'

end

--   The correctness of this specification is based on the following
--   assumptions:
--
--   1. No feature of ARRAY calls a command on any of its arguments.
--
--   2. No query of ARRAY changes the value of any basic specifier.
--
--   3. No feature of ARRAY shares internal structures in any way that
--       could lead to behaviour not deducible from this
--       specification.
--
--   4. The phrase "new ARRAY" in a header comment means a
--       newly-created ARRAY to which there is no reference other than
--       from 'result'.
--
--   5. Every type has a single default value as returned by feature
--       'default' in class GENERAL.