indexing

   title:       "Bit sequences of length `count', with binary operations.";
   cluster:     "kernel";
   project:     "Eiffel Kernel Library";
   date:        "$Date: 2005/02/01 10:20:11 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

expanded class BIT

inherit

   ANY
      redefine
         out
      end

   PLATFORM
      redefine
         out
      end

feature -- Measurement

   count: INTEGER is 0
         -- Number of bits in sequence

feature -- Basic operations

   infix "^" (i: INTEGER): like Current is
         -- Result of shifting bit sequence by 'i' positions
         -- (Positive 'i' shifts right, negative 'i' shifts left
         -- bits failling off the sequence's bounds are lost)
      local
         byte_count: INTEGER
         byte0: BIT 8
         byte1: BIT 8
         distance: INTEGER
         shift0: INTEGER
         shift1: INTEGER
         step: INTEGER
         src_index: INTEGER
         dst_index: INTEGER
         move_count: INTEGER
      do
         distance := i.abs
         if distance >= count then
            -- Result contains only 0
            create Result
         else
            Result := clone (Current)
            if i /= 0 then
               byte_count := (count + 7) // 8
               byte_array1.make (Result, byte_count)
               distance := distance // 8
               if i \\ 8 = 0 then
                  -- Move bytes ...
                  if i > 0 then
                     -- ... right
                     byte_array1.move (
                        distance + 1,
                        byte_count,
                        - distance
                     )
                  else
                     -- ... left
                     byte_array1.move (
                        1,
                        byte_count - distance,
                        distance
                     )
                  end
               else
                  from
                     shift0 := i \\ 8
                     move_count := byte_count - distance
                     if i > 0 then
                        src_index := distance + 1
                        dst_index := 1
                        step      := 1
                     else
                        src_index := byte_count - distance
                        dst_index := byte_count
                        step   := -1
                     end
                     shift1 := shift0 - step * 8
                     byte1 := byte_array1.item (src_index)
                  invariant
                     synchronious_movement: src_index - dst_index = (i + 7) // 8
                     one_byte_per_cycle: step.abs = 1
                  variant
                     move_count + 1
                  until
                     move_count <= 0
                  loop
                     byte0 := byte1
                     move_count := move_count - 1
                     src_index := src_index + step
                     if move_count <= 0 then
                        byte1 := 00000000B
                     else
                        byte1 := byte_array1.item (src_index)
                     end
                     byte_array1.put(
                        (byte0 ^ shift0) or (byte1 ^ shift1),
                        dst_index
                     )
                     dst_index := dst_index + step
                  end
               end
               -- Zero bytes ...
               if i > 0 then
                  -- ... left
                  byte_array1.fill (
                     00000000B,
                     byte_count - distance + 1,
                     byte_count
                  )
               else
                  -- ... right
                  byte_array1.fill (
                     00000000B,
                     1,
                     distance
                  )
               end
            end
         end
      end

   infix "#" (i: INTEGER): like Current is
         -- Result of rotating bit sequence by 'i' positions
         -- (Positive 'i' rotates right, negative 'i' rotates left)
      local
         j: INTEGER -- nomalized 'i'
      do
         j := i \\ count
         if i > 0 then
            Result := (Current ^ j) or (Current ^ (j - count))
         else
            Result := (Current ^ j) or (Current ^ (j + count))
         end
      end

   infix "and" (other: like Current): like Current is
         -- Bit-by-bit boolean conjunction with 'other'
      require
         other_exists: other /= Void
         conformance: other.count <= count
      do
         Result := map (other, operation_and)
      end

   infix "or" (other: like Current): like Current is
         -- Bit-by-bit boolean disjunction with 'other'
      require
         other_exists: other /= Void
         conformance: other.count <= count
      do
         Result := map (other, operation_or)
      end

   infix "xor" (other: like Current): like Current is
         -- Bit-by-bit exclusive or with 'other'
      require
         other_exists: other /= Void
         conformance: other.count <= count
      do
         Result := map (other, operation_xor)
      end

   infix "implies" (other: like Current): like Current is
         -- Bit-by-bit boolean implication of 'other'
      require
         other_exists: other /= Void
         conformance: other.count <= count
      do
         Result := map (other, operation_implies)
      end

   prefix "not": like Current is
         -- Bit-by-bit negation
      do
         Result := map (Void, operation_not)
      end

feature -- Element change

   put (value: BOOLEAN; i: INTEGER) is
         -- Set the 'i'-th bit to 1 if 'value' is True, 0 if False
      require
         inside_bounds: i >= 1 and then i <= count
      local
         byte: BIT 8
         bit_offset: BIT 32
      do
         bit_offset.from_integer (i - 1)
         byte_array.make ($Current, (bit_offset ^ 3).to_integer + 1)
         byte := byte_array.item (byte_array.count)
         byte.put (value, (bit_offset and 00000111B).to_integer + 1)
         byte_array.put (byte, byte_array.count)
      ensure
         value_inserted: item (i) = value
      end

feature -- Access

   item, infix "@" (i: INTEGER): BOOLEAN is
         -- 'i'-th bit
      require
         inside_bounds: i >= 1 and then i <= count
      local
         byte: BIT 8
         bit_offset: BIT 32
      do
         bit_offset.from_integer (i - 1)
         byte_array.make ($Current, (bit_offset ^ 3).to_integer + 1)
         byte := byte_array.item (byte_array.count)
         Result := byte.item ((bit_offset and 00000111B).to_integer + 1)
      end

feature -- Convertion

   to_string: STRING is
         -- String representation of bit sequence.
         -- A zero bit is mapped to '0', a one bit
         -- to '1'. item (1) is rightmost character in string
      do
         Result := out
         Result.remove (Result.count)
      ensure
         same_count: Result.count = count
      end

   from_string (s: STRING) is
         -- Fill sequence from string 's'.
         -- 's' must contain '0' and '1' characters only
      require
         string_not_void: s /= Void
         valid_initializer: -- 's' contains '0' and '1' only
         not_too_long: s.count <= count
      local
         bit_index: INTEGER
         str_index: INTEGER
      do
         -- The first loop fills least significant part of Current
         -- with values taken from the string
         from
            bit_index := 1
            str_index := s.count
         invariant
            bit_index + str_index = s.count + 1
         variant
            str_index + 1
         until
            bit_index > count or else str_index <= 0
         loop
            inspect
               s.item (str_index)
            when '0' then put (False, bit_index)
            when '1' then put (True , bit_index)
            end
            bit_index := bit_index + 1
            str_index := str_index - 1
         end
         -- The second loop fills most significant part of Current
         -- with False's
         from
         variant
            count - bit_index + 1
         until
            bit_index > count
         loop
            put (False, bit_index)
            bit_index := bit_index + 1
         end
      end

   to_integer: INTEGER is
         -- Integer representation of bit sequence
      require
         not_too_long: count <= integer_bits
      do
         -- Built-in
      end

   from_integer (i: INTEGER) is
         -- Fill sequence from integer 'i'
      require
         not_too_long: count <= integer_bits
      do
         -- Built-in
      end

   out: STRING is
         -- New string containing printable representation of
         -- current object
      local
         i: INTEGER
      do
         from
            i := count
            create Result.make_filled ('%U', count + 1)
            Result.put ('B', count + 1)
         variant
            i + 1
         until
            i <= 0
         loop
            if item (i) then
               Result.put ('1', count + 1 - i)
            else
               Result.put ('0', count + 1 - i)
            end
            i := i - 1
         end
      end

feature {NONE} -- Implementation: operations

   operation_and,
   operation_or,
   operation_xor,
   operation_implies,
   operation_not: INTEGER is unique

   map (other: like Current; operation: INTEGER): like Current is
      require
         other_exists: operation /= operation_not implies other /= Void
         conformance: other /= Void implies other.count <= count
      local
         byte_count: INTEGER
         other_byte_count: INTEGER
         byte: BIT 8
         other_byte: BIT 8
      do
         Result := clone (Current)
         byte_count := (count + 7) // 8
         byte_array1.make (Result, byte_count)
         if other /= Void then
            other_byte_count := (other.count + 7) // 8
            byte_array2.make (other, byte_count)
         end
         from
         variant
            byte_count + 1
         until
            byte_count <= 0
         loop
            if byte_count <= other_byte_count then
               other_byte := byte_array2.item (byte_count)
               if byte_count = other_byte_count then
                  other_byte := other_byte and
                          (11111111B ^ (other_byte_count * 8 - other.count))
               end
            end
            byte := byte_array1.item (byte_count)
            inspect operation
            when    operation_and     then byte :=     byte and     other_byte
            when    operation_or      then byte :=     byte or      other_byte
            when    operation_xor     then byte :=     byte xor     other_byte
            when    operation_implies then byte :=     byte implies other_byte
            when    operation_not     then byte := not byte
            end
            byte_array1.put (byte, byte_count)
            byte_count := byte_count - 1
         end
      ensure
         result_exists: Result /= Void
         result_as_current: Result.count = count
      end

feature {NONE} -- Implementation: memory structures

   byte_array: RTS_MEMORY_MAPPED_AREA_ARRAY [BIT 8] is
      once
         create Result.make ($Current, 0)
      end

   byte_array1: RTS_MEMORY_MAPPED_OBJECT_ARRAY [BIT 8] is
      once
         create Result.make (Current, 0)
      end

   byte_array2: RTS_MEMORY_MAPPED_OBJECT_ARRAY [BIT 8] is
      once
         create Result.make (Current, 0)
      end

end