indexing

   title:       "Reference class for DOUBLE.";
   cluster:     "kernel";
   project:     "Eiffel Kernel Library";
   date:        "$Date: 2005/02/01 10:20:12 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class DOUBLE_REF

inherit

   HASHABLE
      undefine
         is_equal
      redefine
         out
      end

   NUMERIC
      undefine
         is_equal
      redefine
         out
      end

   COMPARABLE
      redefine
         out
      end

feature -- Access

   item: DOUBLE
         -- Double value

   hash_code: INTEGER is
         -- Hash code value
         -- (From HASHABLE)
      local
         res: DOUBLE
         max_int: INTEGER
      do
         from
            res     := item.abs
            max_int := 65535 *  (65535 // 2)
         invariant
            res >= 0.0
         until
            res < max_int
         loop
            res := res / max_int
         end
         Result := res.floor
      end

feature -- Element change

   set_item (d: DOUBLE) is
         -- Make 'd' the associated double value
      do
         item := d
      ensure
         item_set: item = d
      end

feature -- Comparison

   infix "<" (other: like Current): BOOLEAN is
         -- Is current double less than 'other'?
         -- (From COMPARABLE)
      do
         Result := item < other.item
      end

   one: like Current is
         -- Neutral element for "*" and "/"
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (1.0)
      ensure then
         value: Result.item = 1.0
      end

   sign: INTEGER is
         -- Sign value (0, -1 or 1)
      do
         Result := three_way_comparison (zero)
      ensure
         three_way: Result = three_way_comparison (zero)
      end

   zero: like Current is
         -- Neutral element for "+" and "-"
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (0.0)
      ensure then
         value: Result.item = 0.0
      end

feature -- Status report

   divisible (other: like Current): BOOLEAN is
         -- May current object be divided by 'other'?
         -- (From NUMERIC)
      do
         if other.item /= 0.0 then
            if not Result and then item / other.item = 0.0 then
               -- do_nothing
            end
            Result := not Result
         end
      ensure then
         not_exact_zero: Result implies (other.item /= 0.0)
      rescue
         Result := True
         retry
      end

   valid_divisor (other: like Current): BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility, use divisible"
      do
         Result := divisible (other)
      end

   exponentiable (other: DOUBLE): BOOLEAN is
         -- May current object be elevated to the power 'other'?
         -- (From NUMERIC)
      do
         if not Result and then item ^ other.item = 0.0 then
            -- do_nothing
         end
         Result := not Result
      rescue
         Result := True
         retry
      end

feature -- Conversion

   ceiling: INTEGER is
         -- Smallest integral value no smaller than current object
      do
         Result := item.ceiling
      ensure
         result_no_smaller: Result >= item
         close_enough: Result - item < one.item
      end

   floor: INTEGER is
         -- Greatest integral value no greater than current object
      do
         Result := item.floor
      ensure
         result_no_greater: Result <= item
         close_enough: item - Result < one.item
      end

   rounded: INTEGER is
         -- Rounded integral value
      do
         Result := item.rounded
      ensure
         definition: Result = sign * (abs.item + 0.5).floor
      end

   truncated_to_integer: INTEGER is
         -- Integer part
         -- (same sign, largest absolute value no greater than current object's)
      do
         Result := item.truncated_to_integer
      end

   truncated_to_real: REAL is
         -- Real part
         -- (same sign, largest absolute value no greater than current object's)
      do
         Result := item.truncated_to_real
      end

feature -- Basic operations

   abs: like Current is
         -- Absolute value
      do
         if Current < zero then
            Result := - Current
         else
            Result := Current
         end
      ensure
         non_negative: Result >= zero
         same_absolute_value: standard_equal  (Result,   Current) or
                               standard_equal  (Result, - Current)
      end

   infix "*" (other: like Current): like Current is
         -- Product by 'other'
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (item * other.item)
      end

   infix "+" (other: like Current): like Current is
         -- Sum with 'other'
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (item + other.item)
      end

   infix "-" (other: like Current): like Current is
         -- Result of subtracting 'other'
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (item - other.item)
      end

   infix "/" (other: like Current): like Current is
         -- Division by 'other'
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (item / other.item)
      end

   infix "^" (other: DOUBLE): DOUBLE is
         -- Current double to the power 'other'
         -- (From NUMERIC)
      do
         Result := item ^ other.item
      end

   prefix "+": like Current is
         -- Unary plus
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (+ item)
      end

   prefix "-": like Current is
         -- Unary minus
         -- (From NUMERIC)
      do
         Result := clone (Current)
         Result.set_item (- item)
      end

feature -- Output

   out: STRING is
         -- Printable representation of double value
         -- (From GENERAL)
      do
         create Result.make_empty
         Result.append_double (item)
      end

end