indexing

   title:       "Platform-independent universal properties.",
                "This class is an ancestor to all developer-written classes.";
   cluster:     "kernel";
   project:     "Eiffel Kernel Library";
   date:        "$Date: 2005/02/15 14:53:27 $";
   revision:    "$Revision: 1.2 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class GENERAL

feature -- Access

   generating_type: STRING is
         -- Name of the current object's generating type
         -- (type of which it is a direct instance)
      external "CWC"
         alias "_GEN_generating_type"
      ensure
         result_not_void: Result /= Void
      end

   generator: STRING is
         -- Name of the current object's generating class
         -- (base class of the type of which it is a direct instance).
         -- Every call to this function allocates a new result string
      local
         braket_pos: INTEGER
      do
         Result := generating_type
         braket_pos := Result.index_of ('[', 1)
         if braket_pos > 0 then
            Result.keep_head (braket_pos - 1)
            Result.prune_all (' ')
         end
      ensure
         result_not_void: Result /= Void
      end

   id_object (id: INTEGER): ANY is
         -- Object for which 'object_id' has returned 'id'
         -- void if none
      external "C"
         alias "_GEN_id_object"
      end

   object_id: INTEGER is
         -- Value identifying current object uniquely
         -- meaningful only for reference types
      external "CWC"
         alias "_GEN_object_id"
      end

feature -- Status report

   frozen conforms_to (other: GENERAL): BOOLEAN is
         -- Does type of current object conform to type
         -- of 'other' (as per Eiffel: The Language, chapter 13)?
      require
         other_not_void: other /= Void
      external "CWC"
         alias "_GEN_conforms_to"
      end

   frozen same_type (other: GENERAL): BOOLEAN is
         -- Is type of current object identical to type of 'other'?
      require
         other_not_void: other /= Void
      external "CWC"
         alias "_GEN_same_type"
      ensure
         definition: Result = (conforms_to (other) and
                        other.conforms_to (Current))
      end

feature -- Comparison

   frozen deep_equal (some: GENERAL; other: like some): BOOLEAN is
         -- Are 'some' and 'other' either both void
         -- or attached to isomorphic object structures?
      do
         Result := (some = Void and then other = Void) or else
                   (some /= Void and then other /= Void and then
                   c_gen_deep_equal (some, other))
      ensure
         shallow_implies_deep: standard_equal (some, other) implies Result
         same_type: Result implies some.same_type (other)
         symmetric: Result implies deep_equal (other, some)
      end

   frozen equal (some: GENERAL; other: like some): BOOLEAN is
         -- Are 'some' and 'other' either both void
         -- or attached to objects considered equal?
      do
         Result := (some = Void and other = Void) or else
                   ((some /= Void and other /= Void) and then
                   some.is_equal (other))
      ensure
         definition: Result = ((some = Void and other = Void) or else
                        ((some /= Void and other /= Void) and then
                        some.is_equal (other)))
      end

   is_equal (other: like Current): BOOLEAN is
         -- Is 'other' attached to an object considered equal
         -- to current object?
      require
         other_not_void: other /= Void
      do
         Result := standard_is_equal (other)
      ensure
         consistent: standard_is_equal (other) implies Result
         same_type: Result implies same_type (other)
         symmetric: Result implies other.is_equal (Current)
      end

   frozen standard_equal (some: GENERAL; other: like some): BOOLEAN is
         -- Are 'some' and 'other' either both void or attached to
         -- field-by-field identical objects of the same type?
         -- Always uses the default object comparison criterion
      do
         Result := (some = Void and other = Void) or else
                   ((some /= Void and other /= Void) and then
                   some.standard_is_equal (other))
      ensure
         definition: Result = ((some = Void and other = Void) or else
                        ((some /= Void and other /= Void) and then
                        some.standard_is_equal (other)))
      end

   frozen standard_is_equal (other: like Current): BOOLEAN is
         -- Is 'other' attached to an object of the same type as
         -- current object, and field-by-field identical to it?
      require
         other_not_void: other /= Void
      external "CWC"
         alias "_GEN_is_equal"
      ensure
         same_type: Result implies same_type (other)
         symmetric: Result implies other.standard_is_equal (Current)
      end

feature -- Duplication

   frozen clone (other: GENERAL): like other is
         -- Void if 'other' is void; otherwise, new object equal to 'other'
      do
         if other /= Void then
            Result := c_gen_clone (other, other.address_of_copy_procedure)
         end
      ensure
         equal: equal (Result, other)
      end

   copy (other: like Current) is
         -- Update current object using fields of object attached
         -- to 'other', so as to yield equal objects
      require
         other_not_void: other /= Void
         type_identity: same_type (other)
      do
         c_gen_copy (other)
      ensure
         is_equal: is_equal (other)
      end

   frozen deep_clone (other: GENERAL): like other is
         -- Void if 'other' is void; otherwise, new object stucture
         -- recursively duplicated from the one attached to 'other'
      external "C"
         alias "_GEN_deep_clone"
      ensure
         deep_equal: deep_equal (other, Result)
      end

   frozen standard_clone (other: GENERAL): like other is
         -- Void if 'other' is void; otherwise, new object
         -- field-by-field identical to 'other'.
         -- Always uses the default copying semantics
      do
         if other /= Void then
            Result := c_gen_standard_clone (other)
         end
      ensure
         equal: standard_equal (Result, other)
      end

   frozen standard_copy (other: like Current) is
         -- Copy every field of 'other' onto corresponding field
         -- of current object
      require
         other_not_void: other /= Void
         type_identity: same_type (other)
      external "CWC"
         alias "_GEN_copy"
      ensure
         is_standard_equal: standard_is_equal (other)
      end

feature -- Basic operations

   frozen default: like Current is
         -- Default value of current type
      external "CWC"
         alias "_GEN_default"
      end

   frozen default_pointer: POINTER is
         -- Default value of type POINTER
         -- (Avoid the need to write 'p.default'
         -- for some 'p' of type POINTER)
      do
         -- Result gets default value during its initialization
      ensure
         Result = Result.default
      end

   default_rescue is
         -- Handle exception if no Rescue clause.
         -- (Default: do nothing)
      do
      end

   frozen do_nothing is
         -- Execute a null action
      do
      end

   frozen Void: NONE is
         -- Void reference
      once
      end

feature -- Output

   io: STD_FILES is
         -- Handle to standard file setup
      once
         create Result
         Result.set_output_default
      end

   out: STRING is
         -- New string containing printable representation of
         -- current object, each field preceded by its attribute name,
         -- a colon and a space
      do
         Result := c_gen_out
      ensure
         non_void_out: Result /= Void
      end

   frozen tagged_out: STRING is
         -- New string containing printable representation of
         -- current object, each field preceded by its attribute name,
         -- a colon and a space
      do
         Result := c_gen_out
      ensure
         non_void_out: Result /= Void
      end

   print (some: GENERAL) is
         -- Write terse external representation of 'some' on
         -- standard output
      do
         if some = Void then
            io.put_string ("(Void)")
         else
            io.put_string (some.out)
         end
      end

feature -- Access to command line arguments

   command_line_arguments: ARGUMENTS is
      once
         create Result
      end

feature -- Eiffel/S 1.3 compatibility

   frozen standard_is_deep_equal (other: like Current): BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility"
      external "CWC"
         alias "_GEN_deep_equal"
      ensure
         shallow_equal: Result implies standard_is_equal (other)
      end

   is_deep_equal (other: like Current): BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility"
      do
         Result := standard_is_deep_equal (other)
      end

   frozen standard_deep_copy (other: like Current) is
      obsolete "Eiffel/S 1.3 compatibility"
      require
         same_type: same_dynamic_type (Current, other)
         source_not_void: other /= Void
      external "CWC"
         alias "_GEN_deep_copy"
      ensure
         copy_is_deep_equal: standard_is_deep_equal (other)
      end

   deep_copy (other: like Current) is
      obsolete "Eiffel/S 1.3 compatibility"
      do
         standard_deep_copy (other)
      end

   frozen standard_deep_clone (other: ANY): like other is
      obsolete "Eiffel/S 1.3 compatibility"
      external "C"
         alias "_GEN_deep_clone"
      ensure
         clone_is_deep_equal: deep_equal (Result, other)
      end

   frozen same_dynamic_type (some: GENERAL; other: GENERAL): BOOLEAN is
      obsolete "Eiffel/S 1.3 compatibility"
      external "C"
         alias "_GEN_same_type"
      end

feature {GENERAL} -- Implementation

   address_of_copy_procedure: POINTER is
         -- Returns address of feature "copy"
      do
         Result.set_item ($copy)
      end

feature {NONE} -- Implementation

   frozen c_gen_standard_clone (other: GENERAL): like other is
      external "C"
         alias "_GEN_clone"
      end

   frozen c_gen_clone (other: GENERAL; other_copy: POINTER): like other is
      external "C"
         alias "__GEN_clone"
      end

   frozen c_gen_deep_equal (some: GENERAL; other: like some): BOOLEAN is
      external "C"
         alias "_GEN_deep_equal"
      end

   c_gen_copy (other: like Current) is
      external "CWC"
         alias "_GEN_copy"
      end

   c_gen_out: STRING is
      external "CWC"
         alias "_GEN_out"
      end

invariant

   reflexive_equality: standard_is_equal (Current)
   reflexive_conformance: conforms_to (Current)
   involutive_object_id: id_object (object_id) = Current

end