indexing

   title:       "Sequences of characters, accessible through integer %
                %indices in a contiguous range starting from one.";
   cluster:     "kernel";
   project:     "Eiffel Kernel Library";
   date:        "$Date: 2005/02/15 14:53:27 $";
   revision:    "$Revision: 1.2 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class STRING

inherit

   COMPARABLE
      redefine
         compare,
         copy,
         is_equal,
         out,
         three_way_comparison
      end

   HASHABLE
      redefine
         copy,
         is_equal,
         out
      end

   DATA_OUTPUT
      rename
         put_boolean   as append_boolean,
         put_character as append_character,
         put_double    as append_double,
         put_integer   as append_integer,
         put_real      as append_real,
         put_string    as append_string
      redefine
         copy,
         is_equal,
         out
      end

creation

   make_empty,
   make_filled,
   make,
   make_from_string

creation -- Eiffel/S 1.3 compatibility

   adapt

feature {NONE} -- Creation

   make_empty is
         -- Create empty string.
      do
         make_filled ('%U', 0)
      ensure
         empty: count = 0
      end

   make_filled (c: CHARACTER; n: INTEGER) is
         -- Create string of length `n' filled with `c'.
      require
         valid_count: n >= 0
      external "CWC"
         alias "_STR_make_filled"
      ensure
         count_set: count = n
         filled: occurrences (c) = count
      end

feature -- Initialization

   make (suggested_capacity: INTEGER) is
         -- Create empty string or remove all characters from existing string.
      require
         non_negative_suggested_capacity: suggested_capacity >= 0
      do
         make_empty
      ensure
         empty_string: count = 0
      end

   from_c (c_string: POINTER) is
         -- Set the current STRING from a copy of the
         -- zero-byte-terminated memory starting at `c_string'.
      require
         c_string_exists: c_string /= default_pointer
      external "CWC"
         alias "_STR_from_c"
      ensure
         no_zero_byte: not has ('%/0/')
         characters: -- for all i in 1..count, item (i) equals the
                     -- ASCII character at address c_string + 1
         correct_count: -- the ASCII character at address
                        -- c_string + count is NUL
      end

   make_from_string (s: STRING) is
         -- Initialize from the character sequence of `s'.
      require
         s_not_void: s /= Void
      local
         ss: STRING
      do
         if s.same_type (empty_string) then
            ss := s
         else
            ss := s.string
         end
         if Current /= ss then
            make_filled ('%U', ss.count)
            c_string_copy (ss.string)
         end
      ensure
         initialized: same_string (s)
      end

feature {NONE} -- Implementation

   frozen area: POINTER
         -- Pointer to string's data
         -- (for internal use only)

feature -- Basic specifiers

   count: INTEGER
         -- Number of characters

   item (i: INTEGER): CHARACTER is
         -- Character at index `i'
      require
         valid_index: valid_index (i)
      external "CWC"
         alias "_STR_item"
      end

feature -- Access

   hash_code: INTEGER is
         -- Hash code value (vendor dependent hashing function)
      local
         i: INTEGER
         n: INTEGER
         h: BIT 32
         g: BIT 32
      do
         from
            n := count
         until
            i = n
         loop
            i := i + 1
            h.from_integer ((h ^ -4).to_integer + item (i).code)
            g := h and 11110000000000000000000000000000B
            if g /= 00000000000000000000000000000000B then
               h := h xor (g ^ 24)
            end
            h := h and not g
         end
         Result := h.to_integer.abs
      end

   index_of (c: CHARACTER; start_index: INTEGER): INTEGER is
         -- Index of first occurrence of `c' at or after `start_index'
         -- 0 if none
      require
         valid_start_index: start_index >= 1 and start_index <= count + 1
      local
         i: INTEGER
      do
         from
            i := start_index
         until
            i > count
         loop
            if c = item (i) then
               Result := i
               i := count
            end
            i := i + 1
         end
      ensure
         valid_result:
            Result = 0 or (start_index <= Result and Result <= count)
         zero_if_absent:
            (Result = 0) = not substring (start_index, count).has (c)
         found_if_present:
            substring (start_index, count).has (c) implies item (Result) = c
         none_before:
            substring (start_index, count).has (c) implies not substring (start_index, Result - 1).has (c)
      end

   string: STRING is
         -- New STRING having the same character sequence as `Current'
      do
         create Result.make_empty
         Result.append_string (Current)
      ensure
         string_not_void: Result /= Void
         string_type: Result.same_type ("")
         first_item: count > 0 implies Result.item (1) = item (1)
         recurse: count > 1 implies Result.substring (2, count).is_equal (substring (2, count).string)
      end

   substring (start_index, end_index: INTEGER): like Current is
         -- New object containing all characters
         -- from `start_index' to `end_index' inclusive
      require
         valid_start_index: 1 <= start_index
         valid_end_index: end_index <= count
         meaningful_interval: start_index <= end_index + 1
      do
         Result := clone (Current)
         if end_index >= start_index then
            Result.keep_head (end_index - start_index + 1)
            Result.c_string_fill_from_external (Current, start_index, end_index)
         else
            Result.wipe_out
         end
      ensure
         substring_not_void: Result /= Void
         substring_count: Result.count = end_index - start_index + 1
         first_item: Result.count > 0 implies
            Result.item (1) = item (start_index)
         recurse:
            Result.count > 0 implies
            Result.substring (2, Result.count).is_equal
               (substring (start_index + 1, end_index))
      end

   substring_index (other: STRING; start_index: INTEGER): INTEGER is
         -- Index of first occurrence of `other' at or after `start_index'
         -- 0 if none
      require
         other_non_void: other /= Void
         valid_start_index: start_index >= 1 and start_index <= count + 1
      local
         other_count: INTEGER
         c: CHARACTER
         i: INTEGER
         j: INTEGER
         done: BOOLEAN
      do
         other_count := other.count
         if other_count > 0 then
            c := other.item (1)
            from
               Result := index_of (c, start_index)
            until
               done
            loop
               i := Result + other_count - 1
               if Result = 0 or else i > count then
                  -- There is no matching string
                  Result := 0
                  done := true
               else
                  from
                     j := other_count
                  until
                     j <= 0
                  loop
                     if other.item (j) /= item (i) then
                        j := 0
                     end
                     i := i - 1
                     j := j - 1
                  end
                  if j = 0 then
                     -- The matching substring is found
                     done := true
                  else
                     Result := index_of (c, Result + 1)
                  end
               end
            end
         else
            Result := start_index
         end
      ensure
         valid_result: Result = 0 or else
            (start_index <= Result and Result <= count - other.count + 1)
         zero_if_absent: (Result = 0) =
            not substring (start_index, count).has_substring (other)
         at_this_index: Result >= start_index implies
            other.same_string (substring (Result, Result + other.count - 1))
         none_before: Result > start_index implies
            not substring (start_index, Result + other.count - 2).has_substring (other)
      end

   infix "@" (i: INTEGER): CHARACTER is
         -- Character at index `i'
      require
         valid_index: valid_index (i)
      external "CWC"
         alias "_STR_item"
      ensure
         definition: Result = item (i)
      end

feature -- Measurement

   occurrences (c: CHARACTER): INTEGER is
         -- Number of times `c' appears in the string
      local
         i: INTEGER
      do
         from
         until
            i = count
         loop
            i := i + 1
            if item (i) = c then
               Result := Result + 1
            end
         end
      ensure
         zero_if_empty: count = 0 implies Result = 0
         recurse_if_not_found_at_first_position:
            (count > 0 and then item (1) /= c) implies
            Result = substring (2, count).occurrences (c)
         recurse_if_found_at_first_position:
            (count > 0 and then item (1) = c) implies
            Result = 1 + substring (2, count).occurrences (c)
      end

feature -- Comparison

   is_equal (other: like Current): BOOLEAN is
         -- Is `other' attached to an object considered equal
         -- to current object?
         -- (Redefined from GENERAL.)
      do
         if count = other.count then
            Result := compare (other) = 0
         end
      ensure then
         definition: Result = (same_type (other) and then
            count = other.count and then
            (count > 0 implies (item (1) = other.item (1) and
            substring (2, count).is_equal (other.substring (2, count)))))
      end

   same_string (other: STRING): BOOLEAN is
         -- Do `Current' and `other' have the same character sequence?
      require
         other_not_void: other /= Void
      do
         Result := string.is_equal (other.string)
      ensure
         definition: Result = string.is_equal (other.string)
      end

   infix "<" (other: like Current): BOOLEAN is
         -- Is string lexicographically lower than `other'?
         -- (Inherited from COMPARABLE.)
      do
         Result := compare (other) < 0
      ensure then
         definition: Result = (
            count = 0 and      other.count > 0 or
            count > 0 and then other.count > 0 and then (
               item (1) < other.item (1) or
               item (1) = other.item (1) and
               substring (2, count) < other.substring (2, other.count)))
      end

   three_way_comparison (other: like Current): INTEGER is
         -- If string lexicographically equal to 'other', 0; if smaller, -1; if greater, 1
      external "CWC"
         alias "_STR_compare"
      end

feature -- Status report

   has (c: CHARACTER): BOOLEAN is
         -- Does `Current' contain `c'?
      local
         i: INTEGER
      do
         from
         until
            i = count
         loop
            i := i + 1
            if item (i) = c then
               Result := True
               i :=  count
            end
         end
      ensure
         false_if_empty: count = 0 implies not Result
         true_if_first: count > 0 and then item (1) = c implies Result
         recurse:
            (count > 0 and then item (1) /= c) implies
            (Result = substring (2, count).has (c))
      end

   has_substring (other: STRING): BOOLEAN is
         -- Does `Current' contain `other'?
      require
         other_not_void: other /= Void
      do
         Result := substring_index (other, 1) > 0
      ensure
         false_if_too_small: count < other.count implies not Result
         true_if_initial: (count >= other.count and then
                          other.same_string (substring (1, other.count)))
                          implies Result
         recurse: (count >= other.count and then
                  not other.same_string (substring (1, other.count))) implies
                  (Result = substring (2, count).has_substring (other))
      end

   is_boolean: BOOLEAN is
         -- Does `Current' represent a BOOLEAN?
      local
         n: INTEGER
      do
         n := count
         if n = 4 then
            if
               item (1) = 't' and then
               item (2) = 'r' and then
               item (3) = 'u' and then
               item (4) = 'e'
            then
               Result := true
            end
         elseif n = 5 then
            if
               item (1) = 'f' and then
               item (2) = 'a' and then
               item (3) = 'l' and then
               item (4) = 's' and then
               item (5) = 'e'
            then
               Result := true
            end
         end
      ensure
         is_boolean: Result = (same_string ("true") or same_string ("false"))
      end

   is_double: BOOLEAN is
         -- Does `Current' represent a DOUBLE?
      local
         value: DOUBLE
         conversion_result: reference INTEGER
      do
         conversion_result := Conversion_no_error
         value := as_double_converted (conversion_result)
         Result := conversion_result <= Conversion_no_error -- No error or underflow
      ensure
         syntax_and_range:
            -- Result is true if and only if the following two
            -- conditions are satisfied:
            --
            -- 1. In the following BNF grammar, the value of
            --    `Current' can be produced by "Real_literal":
            --
            --    Real_literal    = Mantissa [Exponent_part]
            --    Exponent_part   = "E" Exponent | "e" Exponent
            --    Exponent        = Integer_literal
            --    Mantissa        = Decimal_literal
            --    Decimal_literal = Integer_literal ["." Integer]
            --    Integer_literal = [Sign] Integer
            --    Sign            = "+" | "-"
            --    Integer         = Digit | Digit Integer
            --    Digit           = "0"|"1"|"2"|"3"|"4"|"5"|"6"|"7"|"8"|"9"
            --
            -- 2. The numerical value represented by `Current'
            --    is within the range that can be represented
            --    by an instance of type DOUBLE.
      end

   is_empty: BOOLEAN is
         -- Is string empty?
      do
         Result := count = 0
      ensure
         definition: Result = (count = 0)
      end

   is_integer: BOOLEAN is
         -- Does `Current' represent an INTEGER?
      local
         i: INTEGER
         j: INTEGER
         c: INTEGER
         maximum_count: INTEGER
         maximum_literal: STRING
         normalized_number: STRING
         exit: BOOLEAN
      do
         c := count
         i := c
         if i > 0 then
            from
            until
               exit
            loop
               inspect item (i)
               when '0' .. '9' then
                  i := i - 1
                  exit := i <= 0
               else
                  exit := True
               end
            end
            if i = 0 then
               -- Valid integer literal without sign
               maximum_count := maximum_integer_literal_count
               maximum_literal := maximum_integer_literal
            elseif i = 1 and then c > 1 then
               if item (1) = '+' then
                  -- Valid integer literal with plus sign
                  maximum_count := maximum_signed_integer_literal_count
                  maximum_literal := maximum_signed_integer_literal
               elseif item (1) = '-' then
                  -- Valid integer literal with minus sign
                  maximum_count := minimum_integer_literal_count
                  maximum_literal := minimum_integer_literal
               end
            end
            if maximum_count > 0 then
               -- Valid integer literal
               from
                  -- Skip leading zeroes
                  j := i + 1
               invariant
                  valid_index: valid_index (i + 1) and i + 1 <= j and j - 1 <= count
                  zeroes_before: substring (i + 1, j - 1).occurrences ('0') = j - i - 1
               until
                  j >= c or else item (j) /= '0'
               loop
                  j := j + 1
               end
               c := c - (j - i - 1)
               if c < maximum_count then
                  Result := True
               elseif c = maximum_count then
                  if j - i - 1 > 0 then
                     -- There are leading zeroes
                     -- Remove them
                     normalized_number := substring (j, count)
                     if i > 0 then
                        -- Keep the leading sign
                        normalized_number.precede (item (1))
                     end
                  else
                     normalized_number := Current
                  end
                  Result := normalized_number <= maximum_literal
               end
            end
         end
      ensure
         syntax_and_range:
            -- Result is true if and only if thefollowing two
            -- conditions are satisfied:
            --
            -- 1. In the following BNF grammar, the value of
            --    `Current' can be produced by "Integer_literal":
            --
            --    Integer_literal = [Sign] Integer
            --    Sign            = "+" | "-"
            --    Integer         = Digit | Digit Integer
            --    Digit           = "0"|"1"|"2"|"3"|"4"|"5"|"6"|"7"|"8"|"9"
            --
            -- 2. The integer value represented by `Current'
            --    is within the (inclusive) range
            --    minimum_integer .. maximum_integer
            --    where `minimum_integer' and `maximum_integer'
            --    are the constants defined in class PLATFORM.
      end

   is_real: BOOLEAN is
         -- Does `Current' represent a REAL?
      do
         Result := is_double
      ensure
         syntax_and_range:
            -- Result is true if and only if the following two
            -- conditions are satisfied:
            --
            -- 1. In the following BNF grammar, the value of
            --    `Current' can be produced by "Real_literal":
            --
            --    Real_literal    = Mantissa [Exponent_part]
            --    Exponent_part   = "E" Exponent | "e" Exponent
            --    Exponent        = Integer_literal
            --    Mantissa        = Decimal_literal
            --    Decimal_literal = Integer_literal ["." Integer]
            --    Integer_literal = [Sign] Integer
            --    Sign            = "+" | "-"
            --    Integer         = Digit | Digit Integer
            --    Digit           = "0"|"1"|"2"|"3"|"4"|"5"|"6"|"7"|"8"|"9"
            --
            -- 2. The numerical value represented by `Current'
            --    is within the range that can be represented
            --    by an instance of type REAL.
      end

   valid_index (i: INTEGER): BOOLEAN is
         -- Is `i' within the bounds of the string?
      do
         Result := i > 0 and i <= count
      ensure
         definition: Result = (1 <= i and i <= count)
      end

feature -- Element change

   append_character (c: CHARACTER) is
         -- Append `c' at end.
      external "CWC"
         alias "_STR_append_character"
      ensure then
         new_count: count = old count + 1
         appended: item (count) = c
         stable_before: substring (1, count - 1).is_equal (old clone (Current))
      end

   append_string (s: STRING) is
         -- Append a copy of `s' at end.
      external "CWC"
         alias "_STR_append_string"
      ensure then
         appended: is_equal (old clone (Current) + old clone (s))
      end

   fill_with (c: CHARACTER) is
         -- Replace every character with `c'.
      external "CWC"
         alias "_STR_fill"
      ensure
         same_count: old count = count
         filled: occurrences (c) = count
      end

   insert_character (c: CHARACTER; i: INTEGER) is
         -- Insert `c' at index `i', shifting characters between
         -- ranks `i' and `count' rightwards.
      require
         valid_insertion_index: 1 <= i and i <= count + 1
      external "CWC"
         alias "_STR_insert_character"
      ensure
         one_more_character: count = old count + 1
         inserted: item (i) = c
         stable_before_i: substring (1, i - 1).is_equal (old substring (1, i - 1))
         stable_after_i: substring (i + 1, count).is_equal (old substring (i, count))
      end

   insert_string (s: STRING; i: INTEGER) is
         -- Insert `s' at index `i', shifting characters between ranks
         -- `i' and `count' rightwards.
      require
         string_not_void: s /= Void
         valid_insertion_index: 1 <= i and i <= count + 1
      local
         j, k, n: INTEGER
      do
         if Current = s then
            insert_string (clone (s), i)
         else
            from
               j := i
               n := s.count
            until
               k = n
            loop
               k := k + 1
               insert_character (s.item (k), j)
               j := j + 1
            end
         end
      ensure
         inserted: is_equal (old substring (1, i - 1) +
                             old clone (s) +
                             old substring (i, count))
      end

   put (c: CHARACTER; i: INTEGER) is
         -- Replace character at index `i' by `c'.
      require
         valid_index: valid_index (i)
      external "CWC"
         alias "_STR_put"
      ensure
         stable_count: count = old count
         replaced: item (i) = c
         stable_before_i:
            substring (1, i - 1).is_equal
               (old substring (1, i - 1))
         stable_after_i:
            substring (i + 1, count).is_equal
               (old substring (i + 1, count))
      end

   replace_substring (s: like Current; start_index, end_index: INTEGER) is
         -- Replace the substring from `start_index' to `end_index',
         -- inclusive, with `s'.
      require
         string_not_void: s /= Void
         valid_start_index: 1 <= start_index
         valid_end_index: end_index <= count
         meaningful_interval: start_index <= end_index + 1
      local
         res: like Current
      do
         if start_index <= 1 then
            res := clone (Current)
            res.wipe_out
         else
            res := substring (1, start_index - 1)
         end
         res.append_string (s)
         if end_index < count then
            res.append_string (substring (end_index + 1, count))
         end
         copy (res)
      ensure
         replaced: is_equal (
            old (substring (1, start_index - 1) +
            s +
            substring (end_index + 1, count)))
      end

feature -- Removal

   keep_head (n: INTEGER) is
         -- Remove all the characters except for the first `n'
         -- if `n' >= `count', do nothing.
      require
         n_non_negative: n >= 0
      external "CWC"
         alias "_STR_head"
      ensure
         kept: is_equal (old substring (1, n.min (count)))
      end

   keep_tail (n: INTEGER) is
         -- Remove all the characters except for the last `n'
         -- if `n' > `count', do nothing.
      require
         n_non_negative: n >= 0
      do
         from
         until
            n >= count
         loop
            remove (1)
         end
      ensure
         kept: is_equal (old substring (count - n.min (count) + 1, count))
      end

   remove (i: INTEGER) is
         -- Remove 'i-th' character, shifting characters between
         -- ranks i + 1 and `count' leftwards.
      require
         valid_removal_index: valid_index (i)
      external "CWC"
         alias "_STR_remove"
      ensure
         removed: is_equal (old substring (1, i - 1) + old substring (i + 1, count))
      end

   remove_head (n: INTEGER) is
         -- Remove the first `n' characters
         -- if `n' > `count', remove all.
      require
         n_non_negative: n >= 0
      do
         if n >= count then
            wipe_out
         else
            keep_tail (count - n)
         end
      ensure
         removed: is_equal (old substring (n.min (count) + 1, count))
      end

   remove_substring (start_index, end_index: INTEGER) is
         -- Remove all characters from `start_index'
         -- to `end_index' inclusive.
      require
         valid_start_index: 1 <= start_index
         valid_end_index: end_index <= count
         meaningful_interval: start_index <= end_index + 1
      do
         replace_substring (empty_string, start_index, end_index)
      ensure
         removed: is_equal (old substring (1, start_index - 1) +
                            old substring (end_index + 1, count))
      end

   remove_tail (n: INTEGER) is
         -- Remove the last `n' characters
         -- if `n' > `count', remove all.
      require
         n_non_negative: n >= 0
      do
         if n >= count then
            wipe_out
         else
            keep_head (count - n)
         end
      ensure
         removed: is_equal (old substring (1, count - n.min (count)))
      end

   wipe_out is
         -- Remove all characters.
      do
         keep_head (0)
      ensure
         empty_string: count = 0
      end

feature -- Conversion

   as_lower: like Current is
         -- New object with all letters in lower case
      do
         Result := clone (Current)
         Result.to_lower
      ensure
         length: Result.count = count
         anchor: count > 0 implies Result.item (1) = item (1).as_lower
         recurse: count > 1 implies Result.substring (2, count).is_equal
                  (substring (2, count).as_lower)
      end

   as_upper: like Current is
         -- New object with all letters in upper case
      do
         Result := clone (Current)
         Result.to_upper
      ensure
         length: Result.count = count
         anchor: count > 0 implies Result.item (1) = item (1).as_upper
         recurse: count > 1 implies Result.substring (2, count).is_equal
                  (substring (2, count).as_upper)
      end

   to_boolean: BOOLEAN is
         -- Boolean value
         -- "true" yields `true', "false" yields `false'
      require
         represents_a_boolean: is_boolean
      do
         Result := count = 4
      ensure
         to_boolean: Result = same_string ("true")
      end

   to_double: DOUBLE is
         -- Double value
         -- for example, when applied to "123.0", will yield 123.0 (double)
      require
         represents_a_double: is_double
      local
         conversion_result: reference INTEGER
      do
         conversion_result := Conversion_no_error
         Result := as_double_converted (conversion_result)
      end

   to_integer: INTEGER is
         -- Integer value
         -- for example, when applied to "123", will yield 123
      require
         represents_an_integer: is_integer
      external "CWC"
         alias "_STR_to_integer"
      ensure
         single_digit:
            count = 1 implies
            Result = ("0123456789").index_of (item (1), 1) - 1
         minus_sign_followed_by_single_digit:
            count = 2 and item (1) = '-' implies
            Result = - substring (2, 2).to_integer
         plus_sign_followed_by_single_digit:
            count = 2 and item (1) = '+' implies
            Result = substring (2, 2).to_integer
         recurse_to_reduce_length:
            count > 2 or count = 2 and not (("+-").has (item (1))) implies
            Result // 10 = substring (1, count - 1).to_integer and
            (Result \\ 10).abs = substring (count, count).to_integer
      end

   to_lower is
         -- Convert all letters to lower case.
      external "CWC"
         alias "_STR_to_lower"
      ensure
         length_and_content: is_equal (old as_lower)
      end

   to_real: REAL is
         -- Real value
         -- for example, when applied to "123.0", will yield 123.0
      require
         represents_a_real: is_real
      local
         conversion_result: reference INTEGER
      do
         conversion_result := Conversion_no_error
         Result := as_double_converted (conversion_result).truncated_to_real
      end

   to_upper is
         -- Convert all letters to upper case.
      external "CWC"
         alias "_STR_to_upper"
      ensure
         length_and_content: is_equal (old as_upper)
      end

feature -- Duplication

   copy (other: like Current) is
         -- Reinitialize by copying the characters of `other'.
         -- (This is also used by clone.)
         -- (From GENERAL.)
      do
         c_string_copy (other)
      end

   infix "+" (other: STRING): like Current is
         -- New object which is a clone of `Current' extended
         -- by the characters of `other'.
         -- Create a new object which is the concatenation of Current and other
      require
         other_exists: other /= Void
      do
         Result := clone (Current)
         Result.append_string (other)
      ensure
         result_not_void: Result /= Void
         result_count: Result.count = count + other.count
         initial: Result.substring (1, count).is_equal (Current)
         final: Result.substring (count + 1, count + other.count).same_string (other)
      end

feature -- Output

   out: STRING is
         -- New STRING containing terse printable representation
         -- of current object
      do
         create Result.make_from_string (Current)
      ensure then
         same_items: same_type ("") implies Result.same_string (Current)
      end

feature -- Initialization: Visual Eiffel 3.2 compatibility

   frozen remake (n: INTEGER) is
      obsolete "Visual Eiffel 3.2 compatibility, use wipe_out or make"
         -- Allocate space for at least 'n' characters
      require
         non_negative_size: n >= 0
      external "CWC"
         alias "_STR_make"
      ensure
         empty_string: count = n
      end

feature -- Initialization: Eiffel/S 1.3 compatibility

   adapt (other: STRING) is
      obsolete "Eiffel/S 1.3 compatibility, use make_from_string"
         -- Initialize Current's text from string 's'
      require
         not_void_other: other /= Void
      do
         make_from_string (other)
      end

feature -- Access: ELKS'1995 compatibility

   item_code (i: INTEGER): INTEGER is
         -- Numeric code of character at position 'i'
      require
         valid_index: valid_index (i)
      do
         Result := item (i).code
      end

feature -- Comparison: Visual Eiffel 3.2 compatibility

   same_as (other: STRING): BOOLEAN is
         -- Is string made of same character sequence as other?
         -- Case insensitive comparison.
      require
         other_exists: other /= Void
      local
         current_str: STRING
         other_str: STRING
      do
         current_str := clone (Current)
         current_str.to_lower
         other_str := clone (other)
         other_str.to_lower
         Result := current_str.is_equal (other_str)
      end

feature -- Comparison: Eiffel/S 1.3 compatibility

   compare (other: like Current): INTEGER is
      obsolete "Eiffel/S 1.3 compatibility, use three_way_comparison"
      require else
         not_void_other: other /= Void
      external "CWC"
         alias "_STR_compare"
      end

feature -- Status report: Visual Eiffel specific

   is_open: BOOLEAN is True
         -- May write data?

feature -- Status report: ELKS'1995 compatibility

   empty: BOOLEAN is
      obsolete "ELKS'1995 compatibility, use is_empty"
         -- Is string empty?
      do
         Result := count = 0
      end

feature -- Element change: ELKS'1995 compatibility

   append_boolean (b: BOOLEAN) is
         -- Append the string representation of 'b' at end
      external "CWC"
         alias "_STR_append_boolean"
      end

   append_double (d: DOUBLE) is
         -- Append the string representation of 'd' at end
      external "CWC"
         alias "_STR_append_double"
      end

   append_integer (i: INTEGER) is
         -- Append the string representation of 'i' at end
      external "CWC"
         alias "_STR_append_integer"
      end

   append_real (r: REAL) is
         -- Append the string representation of 'r' at end
      external "CWC"
         alias "_STR_append_real"
      end

   fill_blank is
         -- Fill with blanks
      do
         fill (' ')
      ensure
         --allblank: For every i in 1..count, item (i) = Blank
      end

   head (n: INTEGER) is
      obsolete "ELKS'1995 compatibility, use keep_head"
         -- Remove all characters except for the first 'n'
         -- do nothing if n >= count
      require
         non_negative_argument: n >= 0
      external "CWC"
         alias "_STR_head"
      ensure
         new_count: count = n.min (old count)
         --first_kept: For every i in 1..n, item (i) = old item (i)
      end

   insert (s: like Current; i: INTEGER) is
         -- Add 's' to the left of position 'i'
      require
         string_exists: s /= Void
         good_key: valid_index (i)
      local
         j, k, n: INTEGER
      do
         if Current = s then
            insert (clone (s), i)
         else
            from
               j := i
               n := s.count
            until
               k = n
            loop
               k := k + 1
               insert_character (s.item (k), j)
               j := j + 1
            end
         end
      ensure
         new_count: count = old (count + s.count)
      end

   left_adjust is
         -- Remove leading white space
      local
         i, max_blank_index: INTEGER
         is_non_blank: BOOLEAN
      do
         from
         until
            is_non_blank or else (i = count)
         loop
            i := i + 1
            if item (i) <= ' ' then
               max_blank_index := i
            else
               is_non_blank := True
            end
         end
         if max_blank_index > 0 then
            tail (count - max_blank_index)
         end
      ensure
         new_count: (count /= 0) implies (item (1).code > (' ').code)
      end

   precede (ch: CHARACTER) is
         -- Add 'ch' at front
      external "CWC"
         alias "_STR_precede"
      ensure
         new_count: count = old count + 1
      end

   prepend (s: STRING) is
         -- Prepend a copy of 's' at front
      require
         argument_not_void: s /= Void
      external "CWC"
         alias "_STR_prepend"
      ensure
         new_count: count = old (count + s.count)
      end

   right_adjust is
         -- Remove trailing white space
      local
         c: CHARACTER
         i, last_non_blank_index: INTEGER
         is_non_blank: BOOLEAN
      do
         from
            i := count
         until
            is_non_blank or else (i < 1)
         loop
            c := item (i)
            if c > ' ' then
               is_non_blank := True
               last_non_blank_index := i
            end
            i := i - 1
         end
         head (last_non_blank_index)
      ensure
         new_count: (count /= 0) implies (item (count) /= ' ')
      end

   set (t: like Current; n1, n2: INTEGER) is
         -- Set current string to substring of 't' from indices 'n1'
         -- to 'n2', or to empty string if no such substring
      require
         argument_not_void: t /= Void
         meaningful_origin: 1 <= n1
         meaningful_interval: n1 <= n2
         meaningful_end: n2 <= t.count
      do
         if (n1 >= 1) and (n1 <= n2) and (n2 <= t.count) then
            copy (t.substring (n1, n2))
         else
            wipe_out
         end
      ensure
         is_substring: is_equal (old clone (t.substring (n1, n2)))
      end

   tail (n: INTEGER) is
      obsolete "ELKS'1995 compatibility, use keep_tail"
         -- Remove all characters except for the last 'n'
      require
         non_negative_element: n >= 0
      do
         from
         until
            n >= count
         loop
            remove (1)
         end
      ensure
         new_count: count = n.min (old count)
      end

feature -- Element change: Visual Eiffel 3.3 compatibility

   fill (c: CHARACTER) is
      obsolete "Visual Eiffel 3.3 compatibility, use fill_with"
         -- Replace every character with `c'.
      external "CWC"
         alias "_STR_fill"
      ensure
         same_count: old count = count
         filled: occurrences (c) = count
      end

   put_substring (s: like Current; start_pos, end_pos: INTEGER) is
         obsolete "Visual Eiffel 3.3 compatibility, use replace_substring"
         -- Copy the characters of 's' to positions 'start_pos'..'end_pos'
      require
         string_exists: s /= Void
         index_small_enough: end_pos <= count
         order_respected: start_pos <= end_pos
         index_large_enough: start_pos > 0
      do
         replace_substring (s, start_pos, end_pos)
      ensure
         new_count: count = old (count + s.count) - end_pos + start_pos - 1
      end

feature -- Element change: Eiffel/S 1.3 compatibility

   append (other: STRING) is
      obsolete "Eiffel/S 1.3 compatibility, use append_string"
         -- Append 'other' to Current
      require
         not_void_other: other /= Void
      external "CWC"
         alias "_STR_append_string"
      end

   extend (ch: CHARACTER) is
      obsolete "Eiffel/S 1.3 compatibility, use append_character"
         -- Append 'ch' to string
      external "CWC"
         alias "_STR_append_character"
      end

feature -- Removal: ELKS'1995 compatibility

   prune (c: CHARACTER) is
         -- Remove first occurrence of 'c', if any
      local
         i: INTEGER
      do
         if count /= 0 then
            i := index_of (c, 1)
            if i /= 0 then
               remove (i)
            end
         end
      end

   prune_all (c: CHARACTER) is
         -- Remove all occurrences of 'c'
      local
         i: INTEGER
      do
         from
            i := 1
         until
            i > count
         loop
            i := index_of (c, i)
            if i = 0 then
               i := count + 1
            else
               remove (i)
            end
         end
      ensure
         changed_count: count = (old count) - (old occurrences (c))
         --removed: For every 'i' in 1..count, item (i) /= c
         no_more_occurrences: not has (c)
      end

feature -- Resizing: ELKS'1995 compatibility

   resize (newsize: INTEGER) is
         -- Rearrange string so that it can accommodate
         -- at least 'newsize' characters.
         -- Do not lose any previously entered characters
      require
         new_size_non_negative: newsize >= 0
      local
         s: STRING
      do
         if newsize > count then
            create s.make_filled ('%U', newsize - count)
            append_string (s)
         elseif newsize < count then
            head (newsize)
         end
      end

feature -- Conversion: ELKS'1995 compatibility

   mirror is
         -- Reverse the order of characters.
         -- "Hello world" -> "dlrow olleH"
      local
         tmp: CHARACTER
         i, j: INTEGER
      do
         from
            i := 1
            j := count
         until
            i >= j
         loop
            tmp := item (i)
            put (item (j), i)
            put (tmp, j)
            i := i + 1
            j := j - 1
         end
      ensure
         same_count: count = old count
         reversed: -- For every 'i' in 1..count, item (i) = old item (count+1-i)
      end

   mirrored: like Current is
         -- Mirror image of string
         -- result for "Hello world" is "dlrow olleH"
      do
         Result := clone (Current)
         Result.mirror
      ensure
         same_count: Result.count = count
         --reversed: For every 'i' in 1..count, Result.item (i) = item (count+1-i)
      end

   to_pointer: POINTER is
         -- A pointer to a C from of current string.
         -- Useful only for interfacing with C software
      external "CWC"
         alias "_STR_to_external"
      end

feature -- Conversion: Visual Eiffel 3.2 compatibility

   to_external, to_c: POINTER is
         -- A pointer to a C from of current string.
         -- Useful only for interfacing with C software
      external "CWC"
         alias "_STR_to_external"
      end

feature -- Output: Visual Eiffel specific

   code_sequence: STRING is
         -- Printable representation of the string with all the special
         -- characters replaced by the special character codes
         -- This is a newly created string
         -- Examples:
         --   code_sequence ("ABC")    = "ABC"     -- printed as ABC
         --   code_sequence ("A%B")    = "A%%B"    -- printed as A%B
         --   code_sequence ("%Ax")    = "@x"      -- printed as @x
         --   code_sequence ("5%%")    = "5%%%%"   -- printed as 5%%
         --   code_sequence ("%"")     = "%%%""    -- printed as %"
         --   code_sequence ("%/130/") = "%%/130/" -- printed as %/130/
         -- So, the output enclosed in double quotes gives the original string
      local
         ch: CHARACTER
         i: INTEGER
      do
         from
         until
            i >= count
         loop
            i := i + 1
            ch := item (i)
            inspect ch
            when '%/0/' .. '%/31/', '%'', '%"', '%%', '%/127/' .. '%/255/' then
               -- This is a special character
               if Result = Void then
                  if i = 1 then
                     Result := ""
                  else
                     Result := substring (1, i - 1)
                  end
               end
                  check
                     non_void_result: Result /= Void
                  end
               Result.append_string (ch.code_sequence)
            else
               if Result /= Void then
                  Result.append_character (ch)
               end
            end
         end
         if Result = Void then
            Result := out
         end
      ensure
         non_void_result: Result /= Void
         long_enough: Result.count >= count
      end

feature {NONE} -- Implementation: literals

   maximum_integer_literal: STRING is
         -- String representing maximum integer
      local
         p: expanded PLATFORM
      once
         Result := p.maximum_integer.out
      ensure
         non_void_result: Result /= Void
         integer_result: Result.is_integer
         no_leading_zero: Result.item (1) /= '0'
      end

   maximum_signed_integer_literal: STRING is
         -- String representing maximum signed integer,
         -- i.e. integer with leading '+'
      once
         Result := clone (maximum_integer_literal)
         Result.insert_character ('+', 1)
      ensure
         non_void_result: Result /= Void
         definition: Result.is_equal ("+" + maximum_integer_literal)
      end

   minimum_integer_literal: STRING is
         -- String representing minimum integer
      local
         p: expanded PLATFORM
      once
         Result := p.minimum_integer.out
      ensure
         non_void_result: Result /= Void
         integer_result: Result.is_integer
         negative_result: Result.item (1) = '-'
         no_leading_zero: Result.item (2) /= '0'
      end

   maximum_integer_literal_count: INTEGER is
         -- Length of `maximum_integer_literal'
      local
         p: expanded PLATFORM
         i: INTEGER
      once
         from
            i := p.maximum_integer
         until
            i = 0
         loop
            Result := Result + 1
            i := i // 10
         end
      ensure
         definition: Result = maximum_integer_literal.count
      end

   maximum_signed_integer_literal_count: INTEGER is
         -- Length of `maximum_signed_integer_literal'
      once
         Result := maximum_integer_literal_count + 1
      ensure
         definition: Result = maximum_signed_integer_literal.count
      end

   minimum_integer_literal_count: INTEGER is
         -- Length of `minimum_integer_literal'
      local
         p: expanded PLATFORM
         i: INTEGER
      once
         from
            Result := 1
            i := p.minimum_integer
         until
            i = 0
         loop
            Result := Result + 1
            i := i // 10
         end
      ensure
         definition: Result = minimum_integer_literal.count
      end

feature {NONE} -- Implementation: conversion constants

   Conversion_negative_underflow,
   Conversion_positive_underflow,
   Conversion_no_error,
   Conversion_negative_overflow,
   Conversion_positive_overflow,
   Conversion_not_a_number: INTEGER is unique
         -- Result of the conversion

feature {NONE} -- Implementation: conversion

   as_double_converted (conversion_result: reference INTEGER): DOUBLE is
         -- Double value
         -- `conversion_result' contains the result of conversion
      require
         conversion_result /= Void
      local
         i: INTEGER
         j: INTEGER
         sign: BOOLEAN
         exit: BOOLEAN
         c: CHARACTER
         mantissa: DOUBLE
         exp: INTEGER
         dot: INTEGER
         exp_sign: INTEGER
      do
         conversion_result.set_item (Conversion_not_a_number)
         j := count
         if j > 0 then
            inspect
               item (1)
            when '-' then
               i := 2
               sign := True
            when '+' then
               i := 2
            else
               i := 1
            end
            if i <= j then
               c := item (i)
               -- Valid number should have digit in `c'
               if '0' <= c and then c <= '9' then
                  -- Collect mantissa digits
                  mantissa := c.code - 48 -- ('0').code = 48
                  from
                  until
                     exit
                  loop
                     i := i + 1
                     if i > j then
                        exit := True
                     else
                        c := item (i)
                        if '0' <= c and then c <= '9' then
                           if mantissa < 1.0e20 then
                              -- The number of significant digits is not exceeded yet
                              mantissa := mantissa * 10 + c.code - 48 -- ('0').code = 48
                           else
                              exp := exp + 1
                           end
                           exp := exp - dot
                        elseif c = '.' then
                           if dot = 0 and then i < j then
                              c := item (i + 1)
                              if '0' <= c and then c <= '9' then
                                 dot := 1
                              else
                                 -- Not a valid number literal
                                 exit := True
                              end
                           else
                              -- Not a valid number literal
                              exit := True
                           end
                        else
                           exit := True
                        end
                     end
                  end
                  if i <= j then
                     -- Exponent part
                     if c /= 'e' and then c /= 'E' or else i >= j then
                        -- Not a valid number literal
                        exit := False
                     else
                        exp_sign := 1
                        i := i + 1
                        inspect item (i)
                        when '-' then
                           exp_sign := -1
                           i := i + 1
                        when '+' then
                           i := i + 1
                        else
                        end
                        if i > j then
                           -- Not a valid number literal
                           exit := False
                        else
                           -- Skip leading zeroes
                           from
                           until
                              item (i) /= '0' or else i >= j
                           loop
                              i := i + 1
                           end
                           from
                           until
                              j < i
                           loop
                              c := item (j)
                              if '0' <= c and then c <= '9' then
                                 if exp_sign >=  1000000000 or else
                                    exp_sign <= -1000000000
                                 then
                                    -- Exponent number is too large
                                    if exp_sign < 0 then
                                       if sign then
                                          conversion_result.set_item (Conversion_negative_underflow)
                                          Result := - (0.0)
                                       else
                                          conversion_result.set_item (Conversion_positive_underflow)
                                       end
                                    else
                                       if sign then
                                          conversion_result.set_item (Conversion_negative_overflow)
                                       else
                                          conversion_result.set_item (Conversion_positive_overflow)
                                       end
                                    end
                                    exit := False
                                 else
                                    exp := exp + (c.code - 48) * exp_sign
                                    exp_sign := exp_sign * 10
                                 end
                              else
                                 -- Not a valid number literal
                                 exit := False
                                 j := i - 1
                              end
                              j := j - 1
                           end
                        end
                     end
                  end
                  if exit then
                     -- Valid number literal
                     -- Mantissa and exponent parts are in
                     -- `mantissa' and `exp' respectively
                     if exp >= 308 then
                        -- Exponent is too large
                        if sign then
                           conversion_result.set_item (Conversion_negative_overflow)
                        else
                           conversion_result.set_item (Conversion_positive_overflow)
                        end
                     else
                        conversion_result.set_item (Conversion_no_error)
                        if exp < - 306 then
                           mantissa := mantissa * 1.0e-308
                           exp := exp + 308
                        end
                        Result := mantissa * 10 ^ exp
                        if Result = 0 and then mantissa /= 0 then
                           if sign then
                              conversion_result.set_item (Conversion_negative_underflow)
                           else
                              conversion_result.set_item (Conversion_positive_underflow)
                           end
                        end
                        if sign then
                           Result := - Result
                        end
                     end
                  end
               end
            end
         end
      end

feature {NONE} -- Implementation

   c_string_copy (other: like Current) is
      external "CWC"
         alias "_STR_copy"
      end

   empty_string: STRING is ""

feature {STRING} -- Implementation

   c_string_fill_from_external (other:STRING; from_index, to_index:INTEGER) is
      external "CWC"
         alias "_STR_from_external"
      end

invariant

   non_negative_count: count >= 0

end