indexing

   title:       "Recursive directory traversal";
   cluster:     "pool";
   project:     "Visual Eiffel Library";
   date:        "$Date: 2005/02/15 16:22:04 $";
   revision:    "$Revision: 1.2 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class RECURSIVE_DIRECTORY_INPUT_STREAM

inherit

	INPUT_STREAM [STRING]

create

	make,
	make_files,
	make_directories

feature {NONE} -- Creation

	make (dir: DIRECTORY) is
			-- Create on an object that lists both files and directories
			-- in the directory `dir' and its subdirectories (recursively).
		require
			non_void_dir: dir /= Void
		do
			directory := dir
			create path_prefix.make_empty
			create_reader (dir)
			create recursive_path_prefix.make_empty
			create recursive_reader.make_directories (dir)
			create directories.make (1, 0)
			is_file_list := True
			is_directory_list := True
		ensure
			directory_set: directory = dir
			is_directory_list: is_directory_list
			is_file_list: is_file_list
			is_open: is_open
		end

	make_files (dir: DIRECTORY) is
			-- Create on an object that lists only files
			-- in the directory `dir' and its subdirectories (recursively).
		require
			non_void_dir: dir /= Void
		do
			is_file_list := True
			is_directory_list := False
			make (dir)
			is_file_list := True
			is_directory_list := False
		ensure
			directory_set: directory = dir
			is_file_list: is_file_list
			is_open: is_open
		end

	make_directories (dir: DIRECTORY) is
			-- Create on an object that lists only directories
			-- in the directory `dir' and its subdirectories (recursively).
		require
			non_void_dir: dir /= Void
		do
			is_file_list := False
			is_directory_list := True
			make (dir)
			is_file_list := False
			is_directory_list := True
		ensure
			directory_set: directory = dir
			is_directory_list: is_directory_list
			is_open: is_open
		end

feature -- Status

	directory: DIRECTORY
			-- Accociated directory

	is_file_list: BOOLEAN
			-- Are files included in the list?

	is_directory_list: BOOLEAN
			-- Are directories included in the list?

	is_open: BOOLEAN is
			-- Is directory stream open?
		do
			Result := reader.is_open
		end

	item: STRING
			-- Name of the last entry read

	end_of_stream: BOOLEAN is
			-- Are all entries read?
		do
			if
				reader.end_of_stream and then
				recursive_reader.end_of_stream and then
				directories.count = 0
			then
				Result := true
			end
		end

feature -- Status setting

	close is
			-- Stop directory traversal and release
			-- associated resources.
		do
			if reader.is_open then
				reader.close
			end
			if recursive_reader.is_open then
				recursive_reader.close
			end
		end


feature -- Access

	read is
			-- Read next directory entry.
		local
			dir: DIRECTORY
		do
			from
				read_until_normal_line (reader)
			until
				not reader.end_of_stream or else end_of_stream
			loop
				read_until_normal_line (recursive_reader)
				if recursive_reader.end_of_stream then
					if directories.count > 0 then
						recursive_path_prefix := directories.item (directories.count)
						directories.resize (1, directories.count - 1)
						create dir.make (fs.concat_paths (directory.name, recursive_path_prefix))
						create recursive_reader.make_directories (dir)
					end
				else
					path_prefix := recursive_path_prefix + recursive_reader.item
					path_prefix.append_character (directory.directory_separator)
					directories.force (path_prefix, directories.count + 1)
					create dir.make (fs.concat_paths (directory.name, path_prefix))
					create_reader (dir)
					read_until_normal_line (reader)
				end
			end
			if not reader.end_of_stream then
				item := reader.item
				item.prepend (path_prefix)
			end
		end

feature {NONE} -- Implementation

	create_reader (dir: DIRECTORY) is
			-- Create `reader' that reads files and/or directories
			-- according to `is_file_list' and `is_directory_list'.
		require
			non_void_dir: dir /= Void
		do
			if is_file_list = is_directory_list then
				create reader.make (dir)
			elseif is_file_list then
				create reader.make_files (dir)
			else
				create reader.make_directories (dir)
			end
		ensure
			non_void_reader: reader /= Void
		end

	read_until_normal_line (r: DIRECTORY_INPUT_STREAM) is
			-- Read file name entries until normal ones.
			-- (Skip special file name entries.)
		require
			non_void_stream: r /= Void
			not_end_of_stream: not r.end_of_stream
		do
			from
				r.read
			until
				r.end_of_stream or else
				not r.item.is_equal (current_directory_name) and then
				not r.item.is_equal (parent_directory_name)
			loop
				r.read
			end
		end

	fs: FILE_SYSTEM is
		once
			create Result.make
		ensure
			non_void_fs: Result /= Void
		end

feature {NONE} -- Implementation: data

	path_prefix: STRING
			-- Path prefix to be appended before reader's results

	reader: DIRECTORY_INPUT_STREAM
			-- Current directory content reader

	recursive_path_prefix: STRING
			-- Path prefix to be appended before recursive_reader's results

	recursive_reader: DIRECTORY_INPUT_STREAM
			-- Current reader to traverse subdirectories

	directories: ARRAY [STRING]
			-- Directories to be processed

	current_directory_name: STRING is "."
			-- Name of the current directory

	parent_directory_name: STRING is ".."
			-- Name of the parent directory

invariant

	non_void_directory: directory /= Void

	non_void_path_prefix: path_prefix /= Void
	non_void_recursive_path_prefix: recursive_path_prefix /= Void
	non_void_reader: reader /= Void
	non_void_recursive_reader: recursive_reader /= Void
	non_void_directories: directories /= Void
	standard_directories: directories.lower = 1

end
