indexing

   title:       "The class to store objects in the array in the file";
   cluster:     "pool";
   project:     "Visual Eiffel Library";
   date:        "$Date: 2005/02/01 10:20:16 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class PERSISTENT_ARRAY [G]

inherit

   MEMORY
      rename
         dispose as close
      redefine
         close
      end

creation

   make_create_read_write,
   make_general_create_read_write,
   make_open_read,
   make_open_read_write

feature -- Operations

   put (element: G; index: INTEGER) is
         -- Replace 'index'-th entry by 'element'
      require
         connected: is_connected
         positive_index: index > 0
         extendible: extendible
      do
         PERS_put_f (index, handle, hFile, element)
      end -- put

   item (index: INTEGER): G is
         -- Entry at index 'index', if in interval
      require
         connected: is_connected
         positive_index: index >= 1
      do
         Result := PERS_pop_f (index, handle, hFile)
      end -- item

   close is
         -- Close the file
      do
         if is_connected then
            PERS_Disconnect (handle, hFile)
            handle := 0
         end -- if
      rescue
         handle := 0
      end -- close

   put_range (range: ARRAY [G]; lower, upper: INTEGER) is
         -- Put the first object in array 'range'
         -- (i.e. range.item (range.lower)) at position
         -- 'lower' in the current object, the next object
         -- at position 'lower' + 1, etc.
      require
         connected: is_connected
         positive_index: lower > 0
         valid_range: lower <= upper - 1
         range_not_void: range /= Void
         range_big_enough: upper - lower <= range.upper - range.lower
         extendible: extendible
      local
         i, k: INTEGER
      do
         from
            i := lower
            k := range.lower
         until
            i > upper
         loop
            put (range.item (k), i)
            i := i + 1;
            k := k + 1;
         end -- loop
      end -- put_range

   item_range (lower, upper: INTEGER): ARRAY [G] is
         -- Build an array with index range [lower, upper]
         -- containing the objects from corresponding
         -- positions in the current object.
      require
         connected: is_connected
         positive_index: lower > 0
         valid_range: lower <= upper - 1
      local
         i: INTEGER
      do
         !!Result.make (lower, upper)
         from
            i := lower
         until
            i > upper
         loop
            Result.put (item (i), i)
            i := i + 1
         end -- loop
      end -- item_range

feature -- Status

   is_connected: BOOLEAN is
         -- Is the file open?
      do
         Result := handle /= 0
      end

   extendible: BOOLEAN is
         -- May new items be added?
      do
         Result := mode < READ_FILE
      end

   count: INTEGER is
         -- Number of items in the array
      require
         connected: is_connected
      do
         Result := PERS_count  (handle, hFile)
      end

   is_general: BOOLEAN is
         -- Was "general" store used?
      require
         connected: is_connected
      do
         Result := c_is_general( handle, hFile )
      end

feature {NONE} -- Creation procedures

   make_create_read_write (file_name: STRING) is
         -- Create file object with 'file_name' as file name
         -- and open file for both reading and writing;
         -- create it if it does not exist
      require
         non_void_file_name: file_name /= Void
      do
         mode := CREATE_FILE
         handle := PA_ConnectTo (file_name.to_external, mode, $hFile)
      end -- make_create_read_write

   make_general_create_read_write (file_name: STRING) is
         -- Create file object with 'file_name' as file name
         -- and open file for both reading and writing;
         -- create it if it does not exist
         -- Use "general" store mechanism
      require
         non_void_file_name: file_name /= Void
      do
         mode := CREATE_GENERAL_FILE
         handle := PA_ConnectTo (file_name.to_external, mode, $hFile)
      end

   make_open_read (file_name: STRING) is
         -- Create file object with 'file_name' as file name
         -- and open file in read mode
      require
         non_void_file_name: file_name /= Void
      do
         mode := READ_FILE
         handle := PA_ConnectTo (file_name.to_external, mode, $hFile)
      end

   make_open_read_write (file_name: STRING) is
         -- Create file object with 'file_name' as file name
         -- and open file for both reading and writing
      require
         non_void_file_name: file_name /= Void
      do
         mode := READ_WRITE_FILE
         handle := PA_ConnectTo (file_name.to_external, mode, $hFile)
      end

feature {NONE} -- Data

   handle: INTEGER
   mode: INTEGER
   hFile: INTEGER

feature {NONE} -- Mode

   CREATE_FILE:         INTEGER is 0
   CREATE_GENERAL_FILE: INTEGER is 1
   READ_WRITE_FILE:     INTEGER is 2
   READ_FILE:           INTEGER is 3

feature {NONE} -- External function

   PA_ConnectTo (fname: POINTER; md: INTEGER; hF: POINTER): INTEGER is
      external "CWC"
         alias "_PA_ConnectTo"
      end

   PERS_Disconnect (hook, hF: INTEGER) is
      external "C"
         alias "_PADisconnect"
      end

   PERS_put_f (index, hnd, hF: INTEGER; element: G) is
      external "C"
         alias "_PAput_f"
      end

   PERS_pop_f  (index, hnd, hF: INTEGER): G is
      external "C"
         alias "_PApop_f"
      end

   PERS_count(hnd, hF: INTEGER): INTEGER is
      external "C"
         alias "_PAcount"
      end

   c_is_general (hook, hF: INTEGER): BOOLEAN is
      require
         connected: is_connected
      external "C"
         alias "_isGeneral"
      end

end -- class PERSISTENT_ARRAY
