indexing
   title:       "Ordered sequence of values of the conforming type  ",
                "accessible through integer indices in a contiguous ",
                "interval                                           ";
   cluster:     "pool";
   project:     "Visual Eiffel Library";
   date:        "$Date: 2005/02/01 10:20:16 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class SORTED_ARRAY [G -> COMPARABLE]

inherit

   ARRAY [G]
      rename
         make   as array_make,
         resize as array_resize,
         remove as array_remove,
         count  as array_count
      export
         {NONE} array_make, array_resize, array_remove, array_count
      redefine
         append
      end

creation

   make,
   fill,
   init

feature -- Creation

   make is
      do
         array_make (1, 0)
      ensure
         empty_array: count = 0
      end

   fill (arr: ARRAY [G]) is
      require
         non_void_array: arr /= Void
         valid_lower: arr.lower = 1
         array_sorted: is_sorted (arr)
      do
         make
         c_array_copy (arr)
      ensure
         the_same_capacity: arr.count = count
      end

   init (a: ARRAY [G]) is
      require
         non_void_array: a /= Void
      do
         make
         c_array_copy (a)
         if array_count > 1 then
            sort_it (1, array_count)
         end
      end

feature -- Measurement

   count: INTEGER is
         -- Number of items in the array
      do
         Result := array_count
      ensure
         non_negative_count: Result >= 0
      end

feature -- Operations

   append (arr: like Current) is
         -- Allows to append one SORTED_ARRAY (may be Void) to another
      local
         i: INTEGER
      do
         if arr /= Void then
            from
               i := arr.count
            until
               i <= 0
            loop
               add (arr.item (i))
               i := i - 1
            end
         end
      end

   substract (arr: like Current) is
         -- Allows to substract one SORTED_ARRAY (may be Void) from the current.
      local
         i: INTEGER
      do
         if arr /= Void then
            from
               i := arr.count
            until
               i <= 0
            loop
               remove (arr.item (i))
               i := i - 1
            end
         end
      end

feature -- Removal

   delete (pos: INTEGER) is
      require
         positive_pos: pos > 0
         valid_pos: pos <= count
      local
         c: INTEGER
      do
         c := count
         if pos < c then
            move (pos + 1, c, - 1)
         end
         array_resize (1, c - 1)
      ensure
         count_decreased: count = old count - 1
      end

   remove (x: G) is
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek (x)
         if pos > 0 then
            delete (pos)
         end
      end

   removed (x: G): BOOLEAN is
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek (x)
         if pos > 0 then
            delete (pos)
            Result := True
         end
      ensure
         done: Result implies (count = old count - 1)
         absent: not Result implies (count = old count)
      end

feature -- Element change

   append_item (x: G) is
      require
         non_void_element: x /= Void
         largest_element: count > 0 implies item (count) < x
      local
         new_count: INTEGER
      do
         new_count := count + 1
         array_resize (1, new_count)
         put (x, new_count)
      end

   add (x: G) is
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek (x)
         if pos <= 0 then
            add_after (x, pos)
         end
      end

   added (x: G): BOOLEAN is
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek (x)
         if pos <= 0 then
            add_after (x, pos)
            Result := True
         end
      ensure
         done: Result implies (count = old count + 1)
      end

   replace (x: G) is
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek (x)
         if pos <= 0 then
            add_after (x, pos)
         else
            put (x, pos)
         end
      end

   add_after (x: G; pos: INTEGER) is
      require
         non_void_element: x /= Void
         negative_position: pos <= 0
         valid_position: - pos <= count
      do
         array_resize (1, count + 1)
         insert (x, - pos)
      ensure
         added: seek (x) = - pos + 1
      end

   add_it (x: G): G is
      require
         non_void_element: x /= Void
      local
         pos : INTEGER
      do
         pos := seek (x)
         if pos <= 0 then
            add_after (x, pos)
            Result := x
         else
            Result := item (pos)
         end
      end

feature -- Searching

   search (x: G): like x is
         -- Search for an element that is equal to `x'
         -- Return default value (i.e. Void for reference types) if not found
      require
         non_void_element: x /= Void
      local
         pos: INTEGER
      do
         pos := seek( x )
         if pos > 0 then
            Result := item( pos )
         end
      end

   seek (x: G): INTEGER is
         -- This is used to add new elements into the array.
         -- Allows to test if the element is present in the structure or
         -- where it should be put. So if there is no such an
         -- element in the structure seek will return negative number which means
         -- the position after which this element must be in the order of the sort.
      require
         non_void_element: x /= Void
      local
         l: INTEGER;
         r: INTEGER;
         m: INTEGER
      do
         r := count
         if r > 1 then
            from
               l := 1
            until
               l >= r   -- while l < r do
            loop
               m := (l + r) // 2
               if item (m) < x then
                  l := m + 1
               else
                  r := m
               end
            end

            if x.is_equal (item (r)) then
               Result := r
            elseif r = count then
               if x < item (r) then
                  Result := 1 - r
               else
                  Result := - r
               end -- if
            else
               Result := 1 - r
            end
         elseif r = 1 then
            if x.is_equal (item (1))  then
               Result := 1
            elseif x >= item (1) then
               Result := -1
            end
         end
      end

feature -- Resizing

   resize (n: INTEGER) is
      require
         good_array_size: n >= 0 and then n <= count
      do
         array_resize (1, n)
      end

feature -- Conversion

   to_array: ARRAY [G] is
      do
         Result := Current
      end

feature -- Sorting

   qsort is
      do
         if count > 1 then
            sort_it (1, count)
         end
      end

feature {NONE} -- Implementation

   sort_it (l, r: INTEGER) is
      local
         to_quit : BOOLEAN;
         i, j    : INTEGER;
         w, x    : G
      do
         from
            i := l
            j := r
            x := item ((l + r) // 2)
         until
            to_quit
         loop
            from
            until
               item (i) >= x
            loop
               i := i + 1
            end
            from
            until
               x >= item( j )
            loop
               j := j - 1
            end
            if i <= j then
               w := item (i)
               put (item (j), i)
               put (w, j)
               i := i + 1
               j := j - 1
            end
            to_quit := i > j
         end
         if l < j then
            sort_it (l, j)
         end
         if i < r then
            sort_it (i, r)
         end
      end

   is_sorted (arr: ARRAY [G]): BOOLEAN is
         -- Is `arr' sorted?
      require
         non_void_array: arr /= Void
         valid_lower_index: arr.lower = 1
      local
         i, n: INTEGER
      do
         Result := True
         from
            i := 1
            n := arr.count - 1
         until
            i > n
         loop
            if arr.item (i) >= arr.item (i + 1) then
               i      := n + 1
               Result := False
            else
               i := i + 1
            end
         end
      end

feature {NONE} -- Implementation

   c_array_copy (other: ARRAY [G]) is
         -- Reinitialize by copying all the items of `other'.
         -- (Also used by `clone'.)
         -- (Redefined from GENERAL.)
      external "CWC"
         alias "_ARRAY_copy"
      end

invariant

   sorted: -- is_sorted (Current)

end -- class SORTED_ARRAY
