indexing
   description: "Lists implemented by resizable arrays                       ";
   title      : "EiffelBase: library of reusable components for Eiffel.      ",
                "Based on EiffelBase library for ISE Eiffel,                 ",
                "used with permission.                                       ";
   cluster    : EiffelBase;
   project    : "EiffelBase Library";
   copyright  : "(c) 1986-1997 Interactive Software Engineering Inc.         ",
                "Modifications and extensions:                               ",
                "(c) 1997 Object Tools                                       ",
                "All rights reserved. Duplication and distribution prohibited",
                "May be used only with Visual Eiffel under terms of user     ",
                "license                                                     ";
   key        : sequence;
   address    : "mail_to:support@object-tools.com                            ",
                "http://www.object-tools.com                                 ";

class ARRAYED_LIST [G] inherit

   ISE_ARRAY [G]
      rename
         force as force_i_th,
         item as i_th,
         make as array_make,
         put as put_i_th,
         wipe_out as array_wipe_out,
         count as array_count,
         bag_put as put
      export
         {NONE}
            all;
         {ARRAYED_LIST}
            array_make;
         {ANY}
            capacity
      undefine
         linear_representation, prunable, put,
         prune, consistent, is_equal,
         occurrences,
         extendible, has
      redefine
         extend, setup, copy, prune_all, full, valid_index
      end;

   ISE_ARRAY [G]
      rename
         force as force_i_th,
         item as i_th,
         make as array_make,
         put as put_i_th,
         count as array_count,
         bag_put as put
      export
         {NONE}
            all;
         {ARRAYED_LIST}
            array_make;
         {ANY}
            capacity
      undefine
         linear_representation, prunable, full, put,
         prune, consistent, is_equal,
         occurrences,
         extendible, has
      redefine
         wipe_out, extend,
         setup, copy, prune_all, valid_index
      select
         wipe_out
      end;

   DYNAMIC_LIST [G]
      undefine
         valid_index, infix "@", i_th, put_i_th,
         force
      redefine
         first, last, swap, wipe_out,
         go_i_th, move, prunable, start, finish,
         prune,
         setup, copy, put_left, merge_left,
         merge_right, duplicate, prune_all
      select
         count
      end;

creation

   make, make_filled

feature -- Initialization

   make (n: INTEGER) is
         -- Allocate list with `n' items.
         -- (`n' may be zero for empty list.)
      require
         valid_number_of_items: n >= 0
      do
         array_make (1, n)
      ensure
         correct_position: before
      end;

   make_filled (n: INTEGER) is
         -- Allocate list with `n' items.
         -- (`n' may be zero for empty list.)
         -- This list will be full.
      require
         valid_number_of_items: n >= 0
      do
         array_make (1, n)
         count := n
      ensure
         correct_position: before
         filled: full
      end;

feature -- Access

   item: G is
         -- Current item
      require else
         index_is_valid: valid_index (index)
      do
         Result := area_item (index - 1);
      end;

   first: like item is
         -- Item at first position
      do
         Result := area_item (0);
      end;

   last: like item is
         -- Item at last position
      do
         Result := area_item (count - 1)
      end;

   index: INTEGER;
         -- Index of `item', if valid.

   cursor: CURSOR is
         -- Current cursor position
      do
         !ARRAYED_LIST_CURSOR! Result.make (index)
      end;

feature -- Measurement

   count: INTEGER;
      -- Number of items.

feature -- Status report

   prunable: BOOLEAN is
         -- May items be removed? (Answer: yes.)
      do
         Result := true
      end;

   full: BOOLEAN is
         -- Is structure filled to capacity? (Answer: no.)
      do
         Result := (count = capacity)
      end;

   valid_cursor (p: CURSOR): BOOLEAN is
         -- Can the cursor be moved to position `p'?
      local
         al_c: ARRAYED_LIST_CURSOR
      do
         al_c ?= p;
         if al_c /= Void then
            Result := valid_cursor_index (al_c.index)
         end
      end;

   valid_index (i: INTEGER): BOOLEAN is
         -- Is `i' a valid index?
      do
         Result := (1 <= i) and (i <= count)
      end

feature -- Cursor movement

   move (i: INTEGER) is
         -- Move cursor `i' positions.
      do
         index := index + i;
         if (index > count + 1) then
            index := count + 1
         elseif (index < 0) then
            index := 0
         end
      end;

   start is
         -- Move cursor to first position if any.
      do
         index := 1
      ensure then
         after_when_empty: empty implies after
      end;

   finish is
         -- Move cursor to last position if any.
      do
         index := count
      --| Temporary patch. Start moves the cursor
      --| to the first element. If the list is empty
      --| the cursor is before. The parents (CHAIN, LIST...)
      --| and decendants (ARRAYED_TREE...) need to be revised.
      ensure then
         before_when_empty: empty implies before
      end;

   forth is
         -- Move cursor one position forward.
      do
         index := index + 1
      end;

   back is
         -- Move cursor one position backward.
      do
         index := index - 1
      end;

   go_i_th (i: INTEGER) is
         -- Move cursor to `i'-th position.
      do
         index := i;
      end;

   go_to (p: CURSOR) is
         -- Move cursor to position `p'.
      local
         al_c: ARRAYED_LIST_CURSOR
      do
         al_c ?= p;
            check
               al_c /= Void
            end;
         index := al_c.index
      end;

feature -- Transformation

   swap (i: INTEGER) is
         -- Exchange item at `i'-th position with item
         -- at cursor position.
      local
         old_item: G
      do
         old_item := item;
         replace (area_item (i - 1));
         area_put (old_item, i - 1);
      end;

feature -- Element change

   put_front (v: G) is
         -- Add `v' to the beginning.
         -- Do not move cursor.
      do
         if empty then
            extend (v)
         else
            insert (v, 1)
         end;
      end;

   force, extend (v: G) is
         -- Add `v' to end.
         -- Do not move cursor.
      do
         count := count + 1;
         force_i_th (v, count)
      end;

   put_left (v: G) is
         -- Add `v' to the left of current position.
         -- Do not move cursor.
      do
         if after or empty then
            extend (v);
            index := index + 1
         else
            insert (v, index)
         end
      end;

   put_right (v: G) is
         -- Add `v' to the right of current position.
         -- Do not move cursor.
      do
         if index = count then
            extend (v)
         else
            insert (v, index + 1)
         end;
      end;


   replace (v: G) is
         -- Replace current item by `v'.
      do
         put_i_th (v, index)
      end;

   merge_left (other: like Current) is
      local
         i, l_count: INTEGER;
      do
         if not other.empty then
            resize (1, count + other.count);
            from
               i := count - 1;
               l_count := other.count
            until
               i < index - 1
            loop
               area_put (area_item (i), i + l_count);
               i := i - 1
            end;
            from
               other.start;
               i := index - 1
            until
               other.after
            loop
               area_put (other.item, i);
               i := i + 1;
               other.forth
            end;
            index := index + l_count;
            count := count + l_count;
            other.wipe_out
         end
      end;

   merge_right (other: like Current) is
      local
         old_index: INTEGER;
      do
         old_index := index;
         index := index + 1;
         merge_left (other);
         index := old_index
      end;

feature -- Removal

   prune (v: G) is
         -- Remove first occurrence of `v', if any,
         -- after cursor position.
         -- Move cursor to right neighbor
         -- (or `after' if no right neighbor or `v'does not occur)
      do
         if before then index := 1 end;
         if object_comparison then
            if v /= Void then
               from
               until
                  after or else (item /= Void and then v.is_equal (item))
               loop
                  forth;
               end
            end
         else
            from
            until
               after or else item = v
            loop
               forth;
            end
         end;
         if not after then remove end;
      end;

   remove is
         -- Remove current item.
         -- Move cursor to right neighbor
         -- (or `after' if no right neighbor)
      local
         i,j: INTEGER;
         default_value: G;
      do
         if not off then
            from
               i := index - 1;
            until
               i >= count - 1
            loop
               j := i + 1;
               area_put (area_item (j), i);
               i := j;
            end;
            put_i_th (default_value, count);
            count := count - 1
         end
      end;

   prune_all (v: G) is
         -- Remove all occurrences of `v'.
         -- (Reference or object equality,
         -- based on `object_comparison'.)
         -- Leave cursor `after'.
      local
         i: INTEGER;
         val, default_value: G;
      do
         if object_comparison then
            if v /= void then
               from
                  start
               until
                  after or else (item /= Void and then v.is_equal (item))
               loop
                  index := index + 1;
               end;
               from
                  if not after then
                     i := index;
                     index := index + 1
                  end
               until
                  after
               loop
                  val := item;
                  if val /= Void and then not v.is_equal (val) then
                     put_i_th (val, i);
                     i := i + 1
                  end;
                  index := index + 1;
               end
            end
         else
            from
               start
            until
               after or else (item = v)
            loop
               index := index + 1;
            end;
            from
               if not after then
                  i := index;
                  index := index + 1
               end
            until
               after
            loop
               val := item;
               if val /= v then
                  put_i_th (val, i);
                  i := i + 1;
               end;
               index := index + 1
            end
         end;
         if i > 0 then
            index := i
            from
            until
               i >= count
            loop
               put_i_th (default_value, i);
               i := i + 1;
            end;
            count := index - 1;
         end
      ensure then
         is_after: after;
      end;

   remove_left is
         -- Remove item to the left of cursor position.
         -- Do not move cursor.
      do
         index := index - 1;
         remove;
      end;

   remove_right is
         -- Remove item to the right of cursor position
         -- Do not move cursor
      do
         index := index + 1;
         remove;
         index := index - 1;
      end;

   wipe_out is
         -- Remove all items.
      do
         count := 0;
         index := 0;
         array_wipe_out;
      end;

feature -- Duplication

   setup (other: like Current) is
         -- Prepare current object so that `other'
         -- can be easily copied into it.
         -- It is not necessary to call `setup'
         -- (since `consistent' is always true)
         -- but it will make copying quicker.
      do
         if other.empty then
            wipe_out
         else
            resize (1, other.count)
         end
      end;

   copy (other: like Current) is
      local
         c: like cursor;
      do
         wipe_out;
         c := other.cursor;
         from
            other.start
         until
            other.after
         loop
            extend (other.item);
            other.forth
         end;
         other.go_to (c);
      end;

   duplicate (n: INTEGER): like Current is
         -- Copy of sub-list beginning at current position
         -- and having min (`n', `count' - `index' + 1) items.
      local
         pos: INTEGER
      do
         !! Result.make (n.min (count - index + 1));
         from
            Result.start;
            pos := index
         until
            Result.count = Result.capacity
         loop
            Result.extend (item);
            forth;
         end;
         Result.start;
         go_i_th (pos);
      end;


feature {NONE} --Internal

   insert (v: G; pos: INTEGER) is
         -- Add `v' at `pos', moving subsequent items
         -- to the right.
      require
         index_small_enough: pos <= count;
         index_large_enough: pos >= 1;
      local
         i,j: INTEGER;
         last_item: G;
      do
         if index >= pos then
            index := index + 1
         end;
         last_item := last;
         count := count + 1;
         force_i_th (last_item, count);
         from
            i := count - 2
         until
            i < pos
         loop
            j := i - 1;
            area_put (area_item (j), i);
            i := j;
         end;
         put_i_th (v, pos);
      ensure
         new_count: count = old count + 1;
         insertion_done: i_th (pos) = v
      end;

   new_chain: like Current is
         -- unused
      do
      end;

invariant

   prunable: prunable;

end -- class ARRAYED_LIST
