/*
 * The search template for mispelled.
 */

/* Searching has to distinguish lots of cases, eg:
 *  - forward or backward search
 *  - case sensitive or insensitive search
 *  - small distance table (based on chars) or large one (based on ints)
 *	- single-line vs. multi-line mode
 * Since it's speed what counts here, we cannot distinuish the cases in every
 * simple step (inside the searching routine) but must dispatch to different
 * searching routines at the top level. 
 * Since all cases must be combined we get a lot of routines, which are almost
 * identical. To ease maintenance of the code, all routines are generated
 * from the following template.
 *
 * To build the routines, you have to define the following macros:
 *  - define SEARCH to the name of the search routine
 *  - define DTABLE to the type of distance table used
 *  - define BACKWARD if routine is for backward search
 *	- define IGNORE_CASE if routine is for case-insensitive search
 *	- define MULTI_LINE if routine is for multi-line mode
 */
 
/* The following macros are defined for convenience and to make the code more
 * readable. 
 */
#undef INCR1
#undef INCR
#undef PADD
#undef LESS_EQUAL
#ifdef BACKWARD
#define INCR1(p)			(p)--
#define INCR(p,v)			(p) -= (v)
#define PADD(p,v)			((p) - (v))
#define LESS_EQUAL(p1,p2)	((p1) >= (p2))
#else
#define INCR1(p)			(p)++
#define INCR(p,v)			(p) += (v)
#define PADD(p,v)			((p) + (v))
#define LESS_EQUAL(p1,p2)	((p1) <= (p2))
#endif

#undef CASE_ADAPT
#ifdef IGNORE_CASE
#define CASE_ADAPT(c)		(tolower(c))
#else
#define CASE_ADAPT(c)		(c)
#endif

#undef IS_NEWLINE
#ifdef EIF_WINDOWS
#define	IS_NEWLINE	(*t == '\r' && (i + 1 < plen ? *(PADD(t, 1)) == '\n' : 1))
#else 
#define IS_NEWLINE	(*t == '\n')
#endif

/*
 * Finally, this is the generic searching routine.
 */

static EIF_POINTER SEARCH	(
	unsigned char *text, 		/* The text we are searching through */
	int tlen,					/* Length of the text */
	unsigned char *pattern, 	/* The pattern string */
	int plen,					/* The length of the pattern */
	int mm_max, 				/* Allowed number of mismatches */
	DTABLE *dt					/* The shift tables for the pattern */
)
{
	/* The fast searching algorithm. It searches in `text' for
	 * `pattern' and returns a pointer to the position where `pattern'
	 * matches with at most `mm_max' mismatches. `text' and `pattern' are
	 * pointers to the first character in a forward search and pointers to
	 * the last character in a backward search. All INCR macros are defined
	 * as to increment pointers in a forward search and decrement them in a
	 * backward search. So simply read the code as if we do a forward search.
	 */

	register1 unsigned char *p;		/* Pattern string pointer */
	register2 unsigned char *t;		/* Text pointer */
	register4 unsigned char *tx;	/* Another text pointer (start of search) */
	register3 int i;				/* Position within pattern string */
	register5 int mm;				/* Number of mismatches so far */
	register6 int s;				/* Current shifting minimum */

	mm = 0;							/* No mismatches so far */
	mm_max++;						/* Number of shift tables needed */
	tx = text;						/* Where we start scanning */
	INCR(text, tlen - plen);		/* Last position, where match can occur */

	while(LESS_EQUAL(tx, text)) {	/* While enough text still left */
		for(p = pattern, t = tx, i = 0; i < plen; INCR1(p), INCR1(t), i++)
			if(*p != CASE_ADAPT(*t))	/* Scan pattern string */
#ifdef MULTI_LINE
				if(++mm >= mm_max || IS_NEWLINE)
#else 
				if(++mm >= mm_max)	/* Did we have enough mismatches? */
#endif
					break;			/* We did, so stop */

		if(i == plen) {			/* Got substring (whole pattern matched) */
			mismatches = mm;
			return (EIF_POINTER) tx;	/* Start of substring found */
		}

		/* Compute the minimum delta shift value */
		for(i = 0, s = plen + 1; i < mm_max; i++) {
			mm = dt[i][*(PADD(tx, plen - i))];
			if(mm < s)
				s = mm;
		}
		INCR(tx, s);		/* Go to next position in text */
		mm = 0;				/* Reset mismatches count */
	}

	return NULL;		/* No substring found */
}
