/*
 *******************************************************************************
 *                                                      R E G U L A R . C
 *******************************************************************************
 *
 * Description:
 *      Pattern matching with regular expressions.
 *
 *******************************************************************************
 */

/*
 * The following implementation of pattern matching by regular expressions is
 * based on a freely redistributable reimplementation  of the Version 8 regexp
 * routines, which was originally written by Henry Spencer. The notion "based"
 * is used here, since the original C code was heavily modified and extended,
 * partly to make it more flexible and powerful than the original V8 regexp and
 * partly to adapt it to serve our needs in a clean object-oriented interface
 * usable from Eiffel.
 *
 * To get more information (e.g. a description of the overall syntax) see the
 * separate documentation.
 *
 * NOTE: There are two major principals in the current implementation, which
 * should be kept, even if you make any modifications or extensions:
 *      1.      There is no dynamic memory allocation / management on the C-side. Every
 *              memory is provided by the caller (Eiffel side).
 *      2.      The memory is structured in a way, that the Eiffel-side knows enough
 *              to store and retrieve the memory in a platform independant way.
 */

/* LINTLIBRARY */

#include <limits.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "eif_match.h"

typedef struct regexp {
        char firstchar;         /* Char that must begin a match; '\0' if none obvious */
        char anchored;          /* Is the match anchored (at beginning-of-line only)? */
        char musthave[2];       /* Pointer to string that match must include (or 0) */
        char ignorecase;        /* Shall we ignore case? */
        char backward;          /* Is the pattern compiled for backward search? */
        char program[1];        /* Unwarranted chumminess with compiler. */
} Regexp_T;

/* The internal fields in `Regexp_T' are present to pass info from compile
 * to execute that permits the execute phase to run lots faster on simple
 * cases.
 * `firstchar' and `anchored' permit very fast decisions on suitable starting
 * points for a match, cutting down the work a lot. 
 * `musthave' is an offset from the field `program' into the code. If it's
 * not 0, program[offset] is the first character of a string, that must be
 * included in any match, so `musthave' permits fast rejection of lines
 * that can never match. Since the "musthave" tests are costly enough,
 * eif_match_reg_compile() supplies a `musthave' only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or + at
 * the start of the r.e., which can involve a lot of backup).
 * `ignorecase' marks the code, since a compiled pattern for case-sensitive
 * match is different from an insensitive one.
 * `program' marks the beginning of the compiled pattern. It is essentially a
 * linear encoding of a nondeterministic finite-state automaton (NFA).
 * Note, that all fields are single chars (even `musthave' is splitted into
 * chars). This has the great advantage, that a compiled r.e. can be safely
 * exchanged between different architectures, which is fine if we want to
 * transmit it by Eiffel Net or store it in a file by use of STORABLE.
 */

/*
 * Utility definitions.
 */
#ifdef CHAR_BIT
#define NBITS CHAR_BIT
#define MASK ((unsigned char) ((1<<NBITS) - 1))
#else
#define NBITS 8
#define MASK ((unsigned char) 0xFF)
#endif

#ifdef UCHAR_MAX
#define NSUBEXP UCHAR_MAX
#else
#define NSUBEXP 255
#endif

#define UCHARAT(p)              ((int) *(unsigned char *) (p))
#define MUST(p)                 ((UCHARAT(p) << NBITS) + UCHARAT((p) + 1))
#define ISMULT(c)               ((c) == '*' || (c) == '+' || (c) == '?')
//#define ISWORDCHAR(c)          (isalnum(c) || (c) == '_')
#define ISWORDCHAR(c)          (((c) >= '0' && (c) <= '0') || ((c) >= 'A' && (c) <= 'Z') || ((c) >= 'a' && (c) <= 'z') || ((c) == '_'))
#define META                    "^$.[()|?+*\\{<>"

/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
#define MAGIC   0xB4

/* The "program" (means the structure of the NFA) has the following form:
 * It consists of a quantity of nodes, where each node is an opcode plus a
 * "next" pointer, possibly plus an operand. "Next" pointers of all nodes
 * except BRANCH implement concatenation; a "chain" of BRANCH nodes (chained
 * up by "next" pointers) denotes alternatives. (Here we have one of the subtle
 * syntax dependencies:  an individual BRANCH (as opposed to a collection of
 * them) is never concatenated with anything because of operator precedence.)
 * The operand of a BRANCH node is the first node of the branch, so you can
 * regard it as a link to a sub-NFA. (NB this is *not* a tree structure:  the
 * tail of the branch connects to the thing following the set of BRANCHes.)
 * The operand of some types of node is a literal string or a single character.
 * The opcodes are:
 */

/* definition   number  opnd?   meaning */
#define END                     0       /* no   End of program. */
#define BOL                     1       /* no   Match "" at beginning of line. */
#define EOL                     2       /* no   Match "" at end of line. */
#define ANY                     3       /* no   Match any one character. */
#define ANYOF           4       /* str  Match any character in this string. */
#define ANYBUT          5       /* str  Match any character not in this string. */
#define BRANCH          6       /* node Match this alternative, or the next... */
#define BACK            7       /* no   Match "", "next" ptr points backward. */
#define EXACTLY         8       /* str  Match this string. */
#define NOTHING         9       /* no   Match empty string. */
#define STAR            10      /* node Match this (simple) thing 0 or more times. */
#define PLUS            11      /* node Match this (simple) thing 1 or more times. */
#define EXACTLYNUL      12      /* no   Match a NUL character. */
#define ANYOFNUL        13      /* str  Match any character in this string or NUL. */
#define ANYBUTNUL       14      /* str  Match any character not in string and not NUL */
#define OPEN            15      /* byte Match "", mark as start of subexpr. #byte */
#define CLOSE           16      /* byte Match "", mark as end of subexpr. #byte */
#define BOW                     17      /* no   Match "" at beginning of a word. */
#define EOW                     18      /* no   Match "" at end of a word. */
#define ANCHEND         99      /* no   Anchored end, a union of END and EOL */

/*
 * Opcode notes:
 *
 * BOL,EOL
 *              BOL and EOL match on the beginning and end of the search string. If
 *              in multi-line mode, they also match on the beginning and end of each
 *              line. Notice that "beginning" and "end" takes into account the search
 *              direction. If we search forward, BOL meens the beginning of the string,
 *              if we search backward, BOL meens the end of the string.
 *
 * BRANCH       
 *              The set of branches constituting a single choice are hooked
 *              together with their "next" pointers, since precedence prevents
 *              anything being concatenated to any individual branch.  The
 *              "next" pointer of the last BRANCH in a choice points to the
 *              thing following the whole choice.  This is also where the
 *              final "next" pointer of each individual branch points; each
 *              branch starts with the operand node of a BRANCH node.
 *
 * BACK         
 *              Normal "next" pointers all implicitly point forward; BACK
 *              exists to make loop structures possible.
 *
 * STAR,PLUS    
 *              '?', and complex '*' and '+', are implemented as circular
 *              BRANCH structures using BACK.  Simple cases (one character
 *              per match) are implemented with STAR and PLUS for speed
 *              and to minimize recursive plunges.
 *
 * OPEN,CLOSE
 *              Mark parenthesized subexpressions for access of subpatterns. The subs
 *              are numbered at compile time and the number is stored as an operand of
 *              the node.
 *
 * EXACTLY,EXACTLYNUL
 *              A sequence of ordinary characters is sampled and stored as operand
 *              of a EXACTLY node. This is smaller to store and more efficient to run.
 *              The characters are stored as a C-string, which meens they cannot contain
 *              NUL-characters. If the sequence contains NUL-characters, it is splitted
 *              in separate NUL-character-free nodes and concatenated with EXACTLYNUL
 *              nodes.
 *
 * ANY
 *              If in multi-line mode, ANY will not match a newline.
 *
 * ANYOF,ANYBUT,ANYOFNUL,ANYBUTNUL
 *              Denotes a character class. Since all possible (or all excluded)
 *              characters are stored in a C-string, the same problems arises as with
 *              EXACTLY nodes; therefor we have special *NUL nodes which handle
 *              that cases. If in multi-line mode, ANYBUT* will not match a newline.
 *
 * BOW,EOW
 *              These are similar to BOL and EOL but match word boundaries. Like BOL
 *              and EOL they depend on the search direction (see there). A word
 *              boundary can be viewed as an imaginary spot between two characters,
 *              with a word character (alphanumeric or `_') on one side and a non-
 *              word character on the other. If the spot is at the beginning or end
 *              of the input string, the not present character is interpreted like
 *              a non-word character.
 *              
 * ANCHEND
 *              Is used for short term anchoring of the search pattern during execution.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define OP(p)           (*(p))
#define NEXT(p)         ((UCHARAT((p) + 1) << NBITS) + UCHARAT((p) + 2))
#define OPERAND(p)      ((p) + 3)

/* Flags for eif_match_reg_execute() */
#define MATCH_PREFIX    1
#define MATCH_SUFFIX    2
#define MULTI_LINE              4

/* Macro like functions
 * they provide a portable interface for different Eiffel compilers
 * as the macro extension is not supported by all compilers
 */

EIF_INTEGER eif_match_MATCH_PREFIX(void)
{
        return MATCH_PREFIX;
}

EIF_INTEGER eif_match_MATCH_SUFFIX(void)
{
        return MATCH_SUFFIX;
}

EIF_INTEGER eif_match_MULTI_LINE(void)
{
        return MULTI_LINE;
}

/*
 * Error codes returned by eif_match_reg_compile()
 * Used in class REGEXP_ERRORS
 */

/* The following are common errors from the user */
#define RE_UNMATCHED_PAREN      1       /* unmatched () */
#define RE_UNMATCHED_BRACK      2       /* unmatched [] */
#define RE_INVALID_RANGE        3       /* invalid [] range */
#define RE_UNMATCHED_BRACE      4       /* unmatched {} */
#define RE_INVALID_CLASS        5       /* {x} denotes no valid class */
#define RE_POSSIBLY_EMPTY       6       /* operand of a * or + could be empty */
#define RE_EMPTY_REPEAT         7       /* *, + or ? follows nothing */
#define RE_NESTED_REPEAT        8       /* nested *, +, ? */
#define RE_INCOMPLETE           9       /* trailing \ */

/* This errors may occur, but rather seldom */
#define RE_TOO_MANY_PAREN       90      /* too manny () */
#define RE_LONG_POINTER         91      /* pointer to next too large to be stored */

/*
 * Global work variables for regcomp().
 */
static char *regparse;          /* Input-scan pointer. */
static char *regparseend;       /* Points one behind input pattern string */
static int  regnpar;            /* () count. */
static char *regcode;           /* Code-emit pointer. */
static long regsize;            /* Real code size. */
static long regmaxsize; /* Size of the code buffer. */
static char ignore_case;        /* Shall we ignore case? */
static EIF_INTEGER error;       /* Kind of error that occured (0 if no error) */

/*
 * Global work variables for regexec().
 */
static int multi_line;          /* Are we in mult-line mode ? */
static int direction;           /* Direction of search. Forw = 1 / Backw = -1 */
static char *reginput;          /* String-input pointer. */
static char *reginputend;       /* Points one char behind/before relev. part  */
static char *reginputbol;       /* Start of input string (first/last char) */
static char *reginputeol;       /* Points one char behind/before input string */
static EIF_INTEGER *regstart;   /* Pointer to start_loc array. */
static EIF_INTEGER *regend;     /* Pointer to end_loc array. */
static EIF_INTEGER reglen;              /* Length of location arrays. */
static EIF_INTEGER highest;     /* Highest location register used. */

#ifdef EIF_WINDOWS
#define NEWLINE_SIZE            2       /* how many characters make a newline */
#define IS_NEWLINE_AT(s)        (direction > 0 ? \
        (*(s) == '\r' && ((s) + 1) != reginputeol && *((s) + 1) == '\n') : \
        (*(s) == '\n' && ((s) - 1) != reginputeol && *((s) - 1) == '\r'))
#else
#define NEWLINE_SIZE            1       /* how many characters make a newline */
#define IS_NEWLINE_AT(s)        (*(s) == '\n')
#endif

/*
 * Forward declarations for regcomp()'s friends.
 */
static char *reg(int paren, int *flagp);
static char *regbranch(int *flagp);
static char *regpiece(int *flagp);
static char *regatom(int *flagp);
static char *regnode(char op);
static char *regnext(register1 char *p);
static void regescaped(char c);
static void regstdclass(unsigned char c);
static void regc(char c);
static void regb(char b);
static void reginsert(char op, char *opnd);
static void regtail(char *p, char *val);
static void regoptail(char *p, char *val);
static void regoptimize(Regexp_T *r, int expensive);
static int strncspn(char *s1, int s1len, char *s2);

/*
 * Forward declarations for eif_match_reg_reverse()'s friends.
 */
static char *reverse(char *from, char *to, char **annotate);
static char *simple_rev(char *from, char *to);

/*
 * Forwards.
 */
static int regtry(Regexp_T *prog, char *string, EIF_INTEGER stringlen);
static int      regmatch(char *prog);
static int regrepeat(char *p);
static char *contains(char *s1, long s1len, char *s2, long s2len);

#ifdef DEBUG
int regnarrate = 0;             /* if regnarrate != 0, you will get more debug info */
static void regdump();
static char *regprop();
#endif

/*
 * Flags to be passed up and down.
 */
#define HASWIDTH        01      /* Known never to match null string. */
#define SIMPLE          02      /* Simple enough to be STAR/PLUS operand. */
#define SPSTART         04      /* Starts with * or +. */
#define WORST           0       /* Worst case. */

/*
 * Error handling
 */
#define FAIL(m)         { if (!error) { error = m; } return NULL; }
#define ERROR(m)        { if (!error) error = m; }

/* The following errors should never occur. They are not placed in the header
 * file, since all of them raise an exception.
 */
#define RE_NULL_ARG             100     /* NULL argument pointer */
#define RE_INTERNAL             101     /* some internal error */
#define RE_NOT_VALID    196     /* not a valid compiled pattern */

#ifdef DEBUG
static raise_error (char *msg)
{
        fprintf(stderr, "matching exception: %s\n", msg);
}
#else
#define raise_error(x)
#endif

EIF_INTEGER eif_match_reg_compile       (
        char *pattern,                  /* the regular expression pattern */
        EIF_INTEGER plen,               /* length of the expression pattern */
        char *code,                     /* code buffer for the compiled regular expr */
        EIF_INTEGER clen,               /* size of the code buffer */
        EIF_BOOLEAN ignore,     /* shall we ignore case ? */
        EIF_INTEGER *nsubs              /* pointer to store number of subexpressions */
)
{
        /* Compile the regular expression in `pattern' of length `plen' and store
         * the result in buffer `code', which has at least size `clen'. If `ignore'
         * is true, we compile the pattern so that any later application will be
         * case-insensitive.
         * The length of the pattern is solely defined by `plen'. The pattern
         * itself can include any NUL characters.
         * If `nsubs' is not NULL, the number of subexpressions is stored in *nsubs.
         * The return values are as follows:
         *      = 0:    if everything was fine
         *      > 0:    if an error occured; return value is the error code
         *      < 0:    if the code buffer was too small; we return the needed size,
         *                      to store the compiled pattern (as negative)
         */

        int flags;

        if(pattern == NULL || code == NULL) {
                raise_error("NULL argument pointer");
                return RE_NULL_ARG;     /* NULL argument */
        }

        /* Initialize */
        error = 0;
        ignore_case = ignore;
        regparse = pattern;
        regparseend = pattern + plen;
        regcode = ((Regexp_T *) code)->program + 1;
        regmaxsize = clen;
        regsize = regcode - code;
        regnpar = 1;

        /* Compile */
        (void) reg(0, &flags);
        if(error)
                return error;

        /* Check, if we had enough space */
        if(regsize > clen) 
                return -regsize;

        /* Optimize. */
        regoptimize((Regexp_T *) code, flags & SPSTART);
        ((Regexp_T *) code)->backward = 0;

        if(nsubs != NULL)
                *nsubs = regnpar - 1;
        return 0;
}

static char *reg        (
        int paren,      /* Parenthesized? */
        int *flagp
)
{
        /* Implementation is by a recursive descent parser, where `reg' is the top
         * level rule. The caller must absorb opening parenthesis.
         * Combining parenthesis handling with the base level of regular expression
         * is a trifle forced, but the need to tie the tails of the branches to what
         * follows makes it hard to avoid.
         */

        register1 char *ret;
        register2 char *br;
        register3 char *ender;
        register4 int parno = 0;
        int flags;

        *flagp = HASWIDTH;      /* Tentatively. */

        /* Make an OPEN node, if parenthesized. */
        if(paren) {
                if(regnpar > NSUBEXP)
                        FAIL(RE_TOO_MANY_PAREN) /* too many () */
                parno = regnpar++;
                ret = regnode(OPEN);
                regb((char) parno);
        } else
                ret = NULL;

        /* Pick up the branches, linking them together. */
        br = regbranch(&flags);
        if(br == NULL)
                return NULL;
        if(ret != NULL)
                regtail(ret, br);       /* OPEN -> first. */
        else
                ret = br;
        if(!(flags & HASWIDTH))
                *flagp &= ~HASWIDTH;
        *flagp |= flags & SPSTART;
        while(regparse < regparseend && *regparse == '|') {
                regparse++;
                br = regbranch(&flags);
                if(br == NULL)
                        return NULL;
                regtail(ret, br);       /* BRANCH -> BRANCH. */
                if(!(flags & HASWIDTH))
                        *flagp &= ~HASWIDTH;
                *flagp |= flags & SPSTART;
        }

        /* Make a closing node, and hook it on the end. */
        if(paren) {
                ender = regnode(CLOSE);
                regb((char) parno);
        } else
                ender = regnode(END);
        regtail(ret, ender);

        /* Hook the tails of the branches to the closing node. */
        for(br = ret; br != NULL; br = regnext(br))
                regoptail(br, ender);

        /* Check for proper termination. */
        if(paren && (regparse >= regparseend || *regparse++ != ')')) {
                FAIL(RE_UNMATCHED_PAREN)        /* unmatched () */
        } else if(!paren && regparse < regparseend) {
                if(*regparse == ')') {
                        FAIL(RE_UNMATCHED_PAREN)        /* unmatched () */
                } else {
                        raise_error("internal error");
                        FAIL(RE_INTERNAL)       /* junk on end, should never happen */
                }
        }

        return ret;
}

static char *regbranch(int *flagp)
{
        /* Implements the concatenation operator 
         * (one alternative of an | operator).
         */

        register1 char *ret;
        register2 char *chain;
        register3 char *latest;
        int flags;

        *flagp = WORST;         /* Tentatively. */

        ret = regnode(BRANCH);
        chain = NULL;
        while(regparse < regparseend && *regparse != '|' && *regparse != ')') {
                latest = regpiece(&flags);
                if(latest == NULL)
                        return NULL;
                *flagp |= flags & HASWIDTH;
                if(chain == NULL)       /* First piece. */
                        *flagp |= flags & SPSTART;
                else
                        regtail(chain, latest);
                chain = latest;
        }
        if(chain == NULL)       /* Loop ran zero times. */
                (void) regnode(NOTHING);

        return ret;
}

static char *regpiece(int *flagp)
{
        /* Something followed by possible [*+?].
         * Note that the branching code sequences used for ? and the general cases
         * of * and + are somewhat optimized:  they use the same NOTHING node as
         * both the endmarker for their branch list and the body of the last
         * branch. It might seem that this node could be dispensed with entirely,
         * but the endmarker role is not redundant.
         */

        register1 char *ret;
        register2 char op;
        register3 char *next;
        int flags;

        ret = regatom(&flags);
        if(ret == NULL)
                return NULL;

        if(regparse >= regparseend || !ISMULT(*regparse)) {
                *flagp = flags;
                return ret;
        }

        op = *regparse;
        if(!(flags & HASWIDTH) && op != '?')
                FAIL(RE_POSSIBLY_EMPTY)         /* *+ operand could be empty */
        *flagp = (op != '+') ? (WORST | SPSTART) : (WORST | HASWIDTH);

        if(op == '*' && (flags & SIMPLE))
                reginsert(STAR, ret);
        else if(op == '*') {
                /* Emit x* as (x&|), where & means "self". */
                reginsert(BRANCH, ret);                         /* Either x */
                regoptail(ret, regnode(BACK));          /* and loop */
                regoptail(ret, ret);                            /* back */
                regtail(ret, regnode(BRANCH));          /* or */
                regtail(ret, regnode(NOTHING));         /* null. */
        } else if(op == '+' && (flags & SIMPLE))
                reginsert(PLUS, ret);
        else if(op == '+') {
                /* Emit x+ as x(&|), where & means "self". */
                next = regnode(BRANCH);                         /* Either */
                regtail(ret, next);
                regtail(regnode(BACK), ret);            /* loop back */
                regtail(next, regnode(BRANCH));         /* or */
                regtail(ret, regnode(NOTHING));         /* null. */
        } else if(op == '?') {
                /* Emit x? as (x|) */
                reginsert(BRANCH, ret);                         /* Either x */
                regtail(ret, regnode(BRANCH));          /* or */
                next = regnode(NOTHING);                        /* null. */
                regtail(ret, next);
                regoptail(ret, next);
        }
        regparse++;
        if(regparse < regparseend && ISMULT(*regparse))
                FAIL(RE_NESTED_REPEAT)  /* nested *?+ */

        return ret;
}

static char *regatom(int *flagp)
{
        /* The lowest level.
         * Optimization:  gobbles an entire sequence of ordinary characters so that
         * it can turn them into a single node, which is smaller to store and
         * faster to run.  Backslashed characters are exceptions, each becoming a
         * separate node; the code is simpler that way and it's not worth fixing.
         */

        register1 char *ret = NULL;
        int flags;

        *flagp = WORST;         /* Tentatively. */

        if(regparse >= regparseend) {
                raise_error("internal error");
                FAIL(RE_INTERNAL)               /* Supposed to be caught earlier. */
        }

        switch(*regparse++) {
                case '^':
                        ret = regnode(BOL);
                        break;
                case '$':
                        ret = regnode(EOL);
                        break;
                case '<':
                        ret = regnode(BOW);
                        break;
                case '>':
                        ret = regnode(EOW);
                        break;
                case '.':
                        ret = regnode(ANY);
                        *flagp |= HASWIDTH | SIMPLE;
                        break;
                case '{':
                        if(regparse >= regparseend)
                                FAIL(RE_UNMATCHED_BRACE);
                        ret = regnode(ANYOF);
                        regstdclass(*regparse++);
                        regb('\0');
                        if(regparse >= regparseend || *regparse != '}')
                                FAIL(RE_UNMATCHED_BRACE)        /* unmatched {} */
                        regparse++;
                        *flagp |= HASWIDTH | SIMPLE;
                        break;
                case '[': {
                        register2 int class;
                        register3 int classend;

                        if(regparse >= regparseend) 
                                FAIL(RE_UNMATCHED_BRACK)        /* unmatched [] */
                        if(*regparse == '^') {  /* Complement of range. */
                                ret = regnode(ANYBUT);
                                regparse++;
                                if(regparse >= regparseend) 
                                        FAIL(RE_UNMATCHED_BRACK)        /* unmatched [] */
                        } else
                                ret = regnode(ANYOF);
                        if(*regparse == ']' || *regparse == '-')
                                regb(*regparse++);
                        while(regparse < regparseend && *regparse != ']') {
                                if(*regparse == '-') {
                                        regparse++;
                                        if(regparse >= regparseend || *regparse == ']')
                                                regb('-');
                                        else {  /* it's a range */
                                                /* If we are case insensitive there are two
                                                 * possibilities to interpret a range (e.g. [K-y]).
                                                 * First one means, that the range itself is
                                                 * influenced (interpret as [K-Y] or [k-y]),
                                                 * second one means, that only the text is affected
                                                 * (interpret as all ascii's betrween ascii 'K'
                                                 * and ascii 'y'). So first interpretation would
                                                 * not match 'Z', while the second does. Although
                                                 * 'egrep' prefers the first style, I implemented
                                                 * the second, since it's more powerful.
                                                 */
                                                class = UCHARAT(regparse - 2) + 1;
                                                classend = UCHARAT(regparse);
                                                if(class > classend + 1)
                                                        FAIL(RE_INVALID_RANGE)  /* invalid [] range */
                                                for(/* empty */; class <= classend; class++)
                                                        regc((char) class);
                                                regparse++;
                                        }
                                } else {        /* plain character */
                                        switch(*regparse) {
                                                case '\0':
                                                        *ret = (*ret == ANYOF ? ANYOFNUL :
                                                                (*ret == ANYBUT ? ANYBUTNUL : *ret));
                                                        break;
                                                case '\\':
                                                        if(++regparse >= regparseend)
                                                                FAIL(RE_INCOMPLETE)     /* trailing \ */
                                                        if(*regparse == '\0') 
                                                                *ret = (*ret == ANYOF ? ANYOFNUL :
                                                                        (*ret == ANYBUT ? ANYBUTNUL : *ret));
                                                        else 
                                                                regescaped(*regparse);
                                                        break;
                                                case '{':
                                                        if(++regparse >= regparseend)
                                                                FAIL(RE_UNMATCHED_BRACE);
                                                        regstdclass(*regparse++);
                                                        if(regparse >= regparseend || *regparse != '}')
                                                                        /* unmatched {} */
                                                                FAIL(RE_UNMATCHED_BRACE);
                                                        break;
                                                default:
                                                        regc(*regparse);
                                        }
                                        regparse++;
                                }
                        }
                        regb('\0');
                        if(regparse >= regparseend || *regparse != ']')
                                FAIL(RE_UNMATCHED_BRACK)        /* unmatched [] */
                        regparse++;
                        *flagp |= HASWIDTH|SIMPLE;
                        break;
                }
                case '(':
                        ret = reg(1, &flags);
                        if(ret == NULL)
                                return NULL;
                        *flagp |= flags & (HASWIDTH | SPSTART);
                        break;
                case '|':
                case ')':
                        raise_error("internal error");
                        FAIL(RE_INTERNAL);              /* Supposed to be caught earlier. */
                        break;
                case '?':
                case '+':
                case '*':
                        FAIL(RE_EMPTY_REPEAT);  /* ?+* follows nothing */
                        break;
                case '\\':
                        if(regparse >= regparseend)
                                FAIL(RE_INCOMPLETE);    /* trailing \ */
                        if(*regparse == '\0') {
                                ret = regnode(EXACTLYNUL);
                        } else {
                                ret = regnode(EXACTLY);
                                regescaped(*regparse);
                                regb('\0');
                        }
                        regparse++;
                        *flagp |= HASWIDTH | SIMPLE;
                        break;
                case '\0':
                        ret = regnode(EXACTLYNUL);
                        *flagp |= HASWIDTH | SIMPLE;
                        break;
                default: {
                        register int len;
                        register char ender;

                        regparse--;
                        len = strncspn(regparse, regparseend - regparse, META);
                        ender = *(regparse + len);
                        if(len > 1 && ISMULT(ender))
                                len--;          /* Back off clear of ?+* operand. */
                        *flagp |= HASWIDTH;
                        if(len == 1)
                                *flagp |= SIMPLE;
                        ret = regnode(EXACTLY);
                        while(len > 0) {
                                regc(*regparse++);
                                len--;
                        }
                        regb('\0');
                        break;
                }
        }

        return ret;
}

static void regescaped(char c)
{
        /* Emit escaped character `c' */

        switch(c) {
                case 't':       /* tab */
                        regc('\t');
                        break;
                case 'n':       /* newline */
                        regc('\n');
                        break;
                case 'r':       /* carriage return */
                        regc('\r');
                        break;
                case 'b':       /* backspace */
                        regc('\b');
                        break;
                case '\f':      /* form feed */
                        regc('\f');
                        break;
                default:
                        regc(c);
        }
}

static void regstdclass(unsigned char c)
{
        /* Expand standard class `c' to corresponding character string */

        switch(c) {
                case 'd':       /* digit */
                        for(c = '0'; c <= '9'; c++)
                                regc(c);
                        break;
                case 'c':       /* capital letter */
                        for(c = 'A'; c <= 'Z'; c++)
                                regc(c);
                        break;
                case 's':       /* small letter */
                        for(c = 'a'; c <= 'z'; c++)
                                regc(c);
                        break;
                case 'l':       /* letter */
                        for(c = 'a'; c <= 'z'; c++)
                                regc(c);
                        for(c = 'A'; c <= 'Z'; c++)
                                regc(c);
                        break;
                case 'a':       /* alphanumeric */
                        for(c = 'a'; c <= 'z'; c++)
                                regc(c);
                        for(c = 'A'; c <= 'Z'; c++)
                                regc(c);
                        for(c = '0'; c <= '9'; c++)
                                regc(c);
                        break;
                case 'w':       /* word character */
                        for(c = 'a'; c <= 'z'; c++)
                                regc(c);
                        for(c = 'A'; c <= 'Z'; c++)
                                regc(c);
                        for(c = '0'; c <= '9'; c++)
                                regc(c);
                        regc('_');
                        break;
                case 'b':       /* blank */
                        regc(' ');
                        for(c = '\011'; c <= '\015'; c++)
                                regc(c);
                        break;
                case 'p':       /* printable */
                        for(c = '\040'; c <= '\177'; c++)
                                regc(c);
                        break;
                default:
                        ERROR(RE_INVALID_CLASS)
        }
}


EIF_INTEGER eif_match_reg_reverse       (
        char *prog,             /* forward compiled regular expression */
        char *code,             /* code buffer for the backward compiled r.e. */
        EIF_INTEGER clen        /* size of the code buffer */
)
{
        /* Reverse the compiled regular expression in `prog', so that it can
         * match backwards and store the result in buffer `code', which must
         * is of size `clen'.
         * As in eif_match_reg_compile() return a value
         *      = 0:    if everything was fine
         *      > 0:    if an error occured; return value is the error code
         *      < 0:    if the code buffer was too small; we return the needed size,
         *                      to store the compiled pattern (as negative)
         */
        
        Regexp_T *src;
        Regexp_T *dst;

        if(prog == NULL || code == NULL) {
                raise_error("NULL argument pointer");
                return RE_NULL_ARG;    /* NULL argument */
        }

        /* Check validity of program. */
        src = (Regexp_T *) prog;
        if(UCHARAT(src->program) != MAGIC) {
                raise_error("not a valid compiled pattern");
                return RE_NOT_VALID;
        }

        /* Initialize */
        error = 0;
        ignore_case = src->ignorecase;
        dst = (Regexp_T *) code;
        regcode = dst->program + 1;
        regmaxsize = clen;
        regsize = regcode - code;
        regnpar = 1;

        /* Reverse */
        (void) reverse(src->program + 1, NULL, NULL);
        if(error)
                return error;

        /* Check, if we had enough space */
        if(regsize > clen)
                return -regsize;

        /* optimize */
        dst->backward = !src->backward;
        regoptimize(dst, MUST(src->musthave));
        /* FIXME: */
        /* `MUST(src->musthave)' is really not appropriate and should be changed */

        return 0;
}

static char *reverse(char *from, char *to, char **annotate)
{
        register char *startp;
        char *endp;
        char *cyclep;
        char *ret;
        char *next;

        /* first we look for a chain of BRANCHES */
        startp = from; 
        while(startp != to && OP(startp) != BRANCH)
                startp = regnext(startp);
        if(startp == to)
                /* there are no branches, so there are no cycles */
                return simple_rev(from, to);
        endp = regnext(startp);
        while(endp != to && OP(endp) == BRANCH)
                endp = regnext(endp);
        /* now startp and endp brackets a list of alternatives */

        /* reverse all behind endp by a recursive call to self */
        if(endp != to) {
                ret = reverse(endp, to, &endp);
                if(endp < startp)       /* `endp' modified by reverse() */
                        startp = endp;
        } else
                ret = NULL;

        /* Before we can reverse the chain of branches, we have to check for
         * cycles, which may lead to a place before `startp'.
         * The NFA contains a cycle on that level if and only if the last node
         * of the sub-NFA of the first branch contains a BACK node. The next-
         * pointer of that BACK node leads either back to the BRANCH node 
         * (a * cycle) or to a node before `startp' (a + cycle).
         * Remember: since the ends of the branches are connected to the node
         * following the alternatives (if we do not have a cycle), we can stop
         * if we have reached `endp'.
         */
        
        for(cyclep = OPERAND(startp); cyclep < endp; cyclep = regnext(cyclep))
                if(OP(cyclep) == BACK)  /* yes, there is a cycle */
                        if((cyclep = regnext(cyclep)) < startp) {  /* it's a + cycle */
                                if(ret != NULL)
                                        /* prelink with current position; we can do that, since
                                         * the following code will generate at least one node
                                         */
                                        regtail(ret, regcode);
                                ret = regcode;  /* remember current position */
                                if(reverse(cyclep, startp, &cyclep) == NULL)
                                        ret = regnode(NOTHING);
                                if(cyclep < from && annotate != NULL) 
                                        /* we had a long jump backward, so propagate the result */
                                        *annotate = cyclep;

                                /* since we know that it's a + cycle, we build the sub-NFA
                                 * the same way as in regpiece();
                                 */
                                regtail(ret, regnode(BRANCH));
                                regtail(regnode(BACK), ret);
                                regtail(ret, regnode(BRANCH));
                                regtail(ret, regcode); /* prelink with following NOTHING node */
                                ret = regnode(NOTHING);
                                break;
                        } else {        /* it's a * cycle */
                                if (ret != NULL)          /* prelink with following BRANCH node */
                                        regtail(ret, regcode);
                                ret = regnode(BRANCH);
                                if((next = reverse(OPERAND(startp),startp, NULL)) == NULL)
                                        next = regnode(NOTHING);
                                regtail(next, regnode(BACK));
                                regtail(next, ret);
                                regtail(ret, regnode(BRANCH));
                                regtail(ret, regcode); /* prelink with following NOTHING node */
                                ret = regnode(NOTHING);
                                break;
                        }
        
        if(cyclep == endp) {    /* there is no cycle, so reverse each branch */
                /* since we have at least one branch, we can prelink the last node */
                if(ret != NULL)
                        regtail(ret, regcode);
                ret = regnode(BRANCH);
                reverse(OPERAND(startp),endp,NULL);
                for(
                                cyclep = regnext(startp);
                                cyclep != endp;
                                cyclep = regnext(cyclep)
                ) {
                        /* reverse other branches */
                        regtail(ret, regnode(BRANCH));
                        reverse(OPERAND(cyclep), endp, NULL);
                }
        }

        if(from != startp) {
                /* We must probably handle nodes from `from' to `startp', but first,
                 * check if not already done during handling of + cycles.
                 */
                if(from < cyclep) {
                        next = regcode;  /* remember current position for later linking */
                        if(simple_rev(from, startp) == NULL)
                                next = NULL;    /* nothing to link to */
                } else /* already done */
                        next = NULL;
        } else 
                /* We are probably on the top level, so include an END node 
                 * if appropriate.
                 */
                if(ret == ((Regexp_T *) (regcode - regsize))->program + 1)
                        /* yes, we are on the top level */
                        next = regnode(END);
                else
                        next = NULL;

        /* If ret is a BRANCH node, we still must link the ends of the branches */
        if(OP(ret) == BRANCH) {
                if(next == NULL)
                        /* there is nothing we can link to, so create a NOTHING node */
                        next = regnode(NOTHING);
                regtail(ret, next);             /* connect both chains */
                do {
                        regoptail(ret,next);
                        ret = regnext(ret);
                } while(OP(ret) == BRANCH);

                return ret;
        }

        /* connect both chains */
        if(next != NULL) {
                regtail(ret, next);
                return next;
        }

        return ret;
}

static char *simple_rev(register1 char *from, register2 char *to)
{
        register3 char *next = NULL;
        register4 char *ret;

        if(!from || from == to)
                return NULL;
        else
                ret = simple_rev(regnext(from), to);
        
        switch(OP(from)) {
                case END:
                case BACK:
                case BRANCH:    /* that should never occur */
                        return ret;
                case STAR:
                case PLUS:
                        next = regnode(OP(from));
                        if(simple_rev(OPERAND(from), NULL) == NULL)
                                (void) regnode(NOTHING);
                        break;
                case BOL:
                        next = regnode(EOL);
                        break;
                case EOL:
                        next = regnode(BOL);
                        break;
                case BOW:
                        next = regnode(EOW);
                        break;
                case EOW:
                        next = regnode(BOW);
                        break;
                case EXACTLYNUL:
                case ANY:
                        next = regnode(OP(from));
                        break;
                case NOTHING:
                        /* We do not need any NOTHING node at the begin of a chain, so omit
                         * generation. If later on it turns out that it was not redundant
                         * (because of an empty chain) we can generate it then.
                         */
                        if(ret == NULL)
                                return NULL;
                        next = regnode(NOTHING);
                        break;
                case ANYOF:
                case ANYBUT:
                case ANYOFNUL:
                case ANYBUTNUL: 
                case EXACTLY: {
                        register5 char *s;
                        register6 char *e;

                        next = regnode(OP(from));
                        s = OPERAND(from);
                        e = s + strlen(s);
                        while(e > s)
                                regb(*--e);
                        regb('\0');
                        break;
                }
                case OPEN:
                case CLOSE:
                        next = regnode(OP(from));
                        regb((char) UCHARAT(OPERAND(from)));
                        break;
        }

        if(ret != NULL)
                regtail(ret, next);

        return next;
}

static char *regnode    (
        char op         /* kind of node */
)
{
        /* Emit a node and return a pointer to that node */

        register1 char *ret;
        register2 char *ptr;

        ret = regcode;
        regsize += 3;
        if(regsize <= regmaxsize) {
                ptr = regcode;
                *ptr++ = op;
                *ptr++ = '\0';          /* Null "next" pointer. */
                *ptr++ = '\0';
                regcode = ptr;
        }
        return ret;
}

static void regb(char b)
{
        /* Emit a byte of code */

        if(regsize <= regmaxsize) 
                *regcode++ = b;
        regsize++;
}

static void regc(char c)
{
        /* Emit one character */

        if(regsize <= regmaxsize) 
                *regcode++ = (ignore_case ? toupper(c) : c);
        regsize++;
}

static void reginsert   (
        char op,        /* kind of node */
        char *opnd      /* pointer to the operand */
)
{
        /* Insert an operator in front of already-emitted operand (this meens
         * we relocate the operand).
         */

        register1 char *src;
        register2 char *dst;
        register3 char *place;

        regsize += 3;
        if(regsize <= regmaxsize) {
                src = regcode;
                regcode += 3;
                dst = regcode;
                while(src > opnd)
                        *--dst = *--src;
                place = opnd;           /* Op node, where operand used to be. */
                *place++ = op;
                *place++ = '\0';
                *place++ = '\0';
        }
}

static void regtail(char *p, char *val)
{
        /* Set the next-pointer at the end of a node chain */

        register1 char *scan;
        register2 char *temp;
        register3 int offset;

        if(regsize > regmaxsize)
                return;

        /* Find last node. */
        scan = p;
        for(;;) {
                temp = regnext(scan);
                if(temp == NULL)
                        break;
                scan = temp;
        }

        if(OP(scan) == BACK)
                offset = scan - val;
        else
                offset = val - scan;
        *(scan + 2) = offset & MASK;
        offset >>= NBITS;
        *(scan + 1) = offset & MASK;
        offset >>= NBITS;
        if(offset)
                ERROR(RE_LONG_POINTER)
}

static void regoptail(char *p, char *val)
{
        /* regtail() on operand of first argument; nop if operandless.
         * "Operandless" and "op != BRANCH" are synonymous in practice.
         */

        if(p == NULL || OP(p) != BRANCH || OPERAND(p) == val)
                return;
        regtail(OPERAND(p), val);
}

static void regoptimize(Regexp_T *r, int expensive)
{
        /* Dig out information for optimizations. */

        register1 char *scan;
        register2 EIF_INTEGER longest;
        register3 EIF_INTEGER len;


        r->ignorecase = ignore_case;    /* Worst-case defaults. */
        r->firstchar = '\0';                    /*      - no first char obvious */
        r->anchored = 0;                                /*      - not anchored */
        r->musthave[0] = 0;                             /*      - no pointer to neccessary string */
        r->musthave[1] = 0;
        r->program[0] = (char) MAGIC;

        scan = r->program + 1;                          /* First BRANCH. */
        if(OP(regnext(scan)) == END) {          /* Only one top-level choice. */
                scan = OPERAND(scan);

                /* Starting-point info. */
                if(OP(scan) == EXACTLY)
                        r->firstchar = *OPERAND(scan);
                else if(OP(scan) == PLUS) {
                        scan = OPERAND(scan);
                        if(OP(scan) == EXACTLY)
                                r->firstchar = *OPERAND(scan);
                } else if(OP(scan) == BOL)
                        r->anchored++;

                /*
                 * If there's something expensive in the r.e., find the
                 * longest literal string that must appear and make it the
                 * musthave.  Resolve ties in favor of later strings, since
                 * the firstchar check works with the beginning of the r.e.
                 * and avoiding duplication strengthens checking.  Not a
                 * strong reason, but sufficient in the absence of others.
                 */
                if(expensive) {
                        longest = 0;
                        len = 0;
                        for(/* empty */; scan != NULL; scan = regnext(scan))
                                if(OP(scan) == EXACTLY && (int) strlen(OPERAND(scan)) >= len) {
                                        longest = OPERAND(scan) - r->program;
                                        len = strlen(OPERAND(scan));
                                }
                        if(!(longest >> 2 * NBITS)) {
                                /* pointer can be stored in 2 chars */
                                r->musthave[0] = (longest >> NBITS) & MASK;
                                r->musthave[1] = longest & MASK;
                        }
                }
        }
}

/*
 * regexec and friends
 */


EIF_BOOLEAN eif_match_reg_execute       (
        char *prog,             /* The compiled regular expression */
        char *string,           /* String to match against (may include NUL bytes) */
        EIF_INTEGER stringlen,  /* Length of the string */
        EIF_INTEGER from,               /* Index of first relevant character in `string' */
        EIF_INTEGER to,                 /* Index of last relevant character in `string' */
        EIF_INTEGER *startloc,  /* Array to store location of matched expr and */
        EIF_INTEGER *endloc,    /* subs. One for start position, one for ends. */
        EIF_INTEGER lenloc,             /* Length of each location array */
        EIF_INTEGER flags               /* Some flags to control execution, see below */
)
{
        /* Match a regexp against a string. The string may include NUL-bytes,
         * its significant part is denoted by `from' and `to', where `from' is 
         * the first character, `to' is the last character that is considered.
         * Counting starts at one, so 1 is the first character in the string.
         * The case sensitiveness of the match is compiled into the program `prog',
         * so if you want to change it, you must compile again.
         * You can ask for the location of the matched part by providing pointers
         * `startloc' and `endloc'. You can even ask for matched subexpressions.
         * Subexpressions are determined by parenthesised parts of the r.e. and
         * are numbered at compile time. startloc[i] is the begin of the i'th
         * matched subexpression in the string, endloc[i] the end. startloc[0]
         * resp. endloc[0] denotes the top level. `lenloc' determines the size
         * of the location arrays. If you are not interested in location, provide
         * NULL for `startloc' / `endloc' or 0 for `lenloc'. 
         * Locations are indices in the string, not pointers, and counting starts
         * at 1 (first character in string). Locations are with respect to the
         * whole string, and denote the first resp. last matched character. So
         * startloc[i] >= from and endloc[i] <= to for all i.
         * The pattern can be temporarily anchored to the beginning or end of the
         * string by using `flags'. If `flags' has set the `MATCH_PREFIX' bit, the
         * match only succeeds, if startloc[0] will be 1; if the `MATCH_SUFFIX' bit
         * is set, the match only succeeds, if endloc[0] will be `stringlen' after
         * the match.
         * If `flags' has the `MULTI_LINE' bit set, matching will be in multi line
         * mode.
         */

        register1 char *relstr;         /* pointer to the relevant part of string */
        register2 long relstrlen;       /* length of the relevant part */
        register3 char *s;
        register4 Regexp_T *r;
        register5 char startc;
        register6 long len;
        char *end_node = NULL;
        char temp_anch = 0;
        char *mh = NULL;
        long mhlen = 0;

        /* Be paranoid... */
        if(prog == NULL || string == NULL) {
                raise_error("NULL argument pointer");
                ERROR(RE_NULL_ARG)
                return EIF_FALSE;
        }
        r = (Regexp_T *) prog;

        /* Check validity of program. */
        if(UCHARAT(r->program) != MAGIC) {
                raise_error("not a valid compiled pattern");
                return EIF_FALSE;
        }

        /* Initialize globals */
        ignore_case = r->ignorecase;
        multi_line = (flags & MULTI_LINE);
        if(r->backward) {
                direction = -1;
                reginputbol = string + stringlen - 1;
                reginputeol = string - 1;
        } else {
                direction = 1;
                reginputbol = string;
                reginputeol = string + stringlen;
        }
        if(startloc == NULL || endloc == NULL)
                reglen = 0;
        else {
                regstart = startloc;
                regend = endloc;
                reglen = lenloc;
        }
        highest = reglen - 1;                   /* force initialization */
        relstr = string + from - 1;             /* pointer to the relevant part */
        relstrlen = to - from + 1;              /* length of the relevant part */

        /* If there is a "musthave" string, look for it. */
        s = r->program + MUST(r->musthave);
        if(s > r->program) {
                mh = s;
                mhlen = strlen(s);
                if(!multi_line && contains(relstr, relstrlen, mh, mhlen) == NULL)
                        return EIF_FALSE;
        }

        /* set temporary anchors */
        if(
                        (direction > 0 && (flags & MATCH_SUFFIX)) || 
                        (direction < 0 && (flags & MATCH_PREFIX))
        ) {
                for(
                                end_node = r->program + 1; 
                                OP(end_node) != END; 
                                end_node = regnext(end_node)
                )
                        /* empty */ ;
                *end_node = ANCHEND;
        }
        if(
                        (direction > 0 && (flags & MATCH_PREFIX)) ||
                        (direction < 0 && (flags & MATCH_SUFFIX))
        )
                temp_anch = 1;

        /* Simplest case:  anchored match need be tried only once. */
        if((!multi_line && r->anchored) || temp_anch)
                if(regtry(r, relstr, relstrlen))
                        goto success;
                else
                        goto failure;

        /* More complex: anchored match in multi-line mode */
        if(multi_line && r->anchored) {
                if(r->backward) {
                        s = relstr + relstrlen - 1;
                        if(mh == NULL) { /* no hint where to start */
                                if(regtry(r, relstr, relstrlen))
                                        goto success;
                                for(len = relstrlen; len >= NEWLINE_SIZE; len--, s--)
                                        if(
                                                        IS_NEWLINE_AT(s) &&
                                                        regtry(r, relstr, len - NEWLINE_SIZE)
                                        )
                                                goto success;
                        } else {
                                register7 char *se = contains(s, relstrlen, mh, mhlen);

                                if(se != NULL && regtry(r, relstr, relstrlen))
                                        goto success;
                                for(len = relstrlen; se && len >= NEWLINE_SIZE; len--, s--)
                                        if(
                                                        IS_NEWLINE_AT(s) && (se <= s - NEWLINE_SIZE ||
                                                        (se = contains(relstr, len - NEWLINE_SIZE,
                                                                mh, mhlen))) &&
                                                        regtry(r, relstr, len-NEWLINE_SIZE)
                                        )
                                                goto success;
                        }
                } else {        /* forward */
                        s = relstr;
                        if(mh == NULL) { /* no hint where to start */
                                if(regtry(r, relstr, relstrlen))
                                        goto success;
                                for(len = relstrlen; len >= NEWLINE_SIZE; len--, s++)
                                        if(
                                                        IS_NEWLINE_AT(s) && 
                                                        regtry(r, s + NEWLINE_SIZE, len - NEWLINE_SIZE)
                                        )
                                                goto success;
                        } else {
                                register char *se = contains(s, relstrlen, mh, mhlen);

                                if(se != NULL && regtry(r, relstr, relstrlen))
                                        goto success;
                                for(len = relstrlen; se && len >= NEWLINE_SIZE; len--, s++)
                                        if(
                                                        IS_NEWLINE_AT(s) && (se >= s + NEWLINE_SIZE ||
                                                        (se = contains(s + NEWLINE_SIZE, len - NEWLINE_SIZE,
                                                                mh, mhlen))) &&
                                                        regtry(r, s + NEWLINE_SIZE, len - NEWLINE_SIZE)
                                        )
                                                goto success;
                        }
                }
                goto failure;
        }

        /* Messy cases:  unanchored match. */
        if((startc = r->firstchar))

                /* We know what char it must start with. */
                if(r->backward) {
                        s = relstr + relstrlen - 1;             /* goto last relevant char */
                        if(ignore_case) {
                                for(len = relstrlen; len > 0; len--, s--)
                                        if(toupper(*s) == startc && regtry(r, relstr, len))
                                                goto success;
                        } else {
                                for(len = relstrlen; len > 0; len--, s--)
                                        if(*s == startc && regtry(r, relstr, len))
                                                goto success;
                        }
                } else {        /* forward */
                        s = relstr;
                        if(ignore_case) {
                                for(len = relstrlen; len > 0; len--, s++)
                                        if(toupper(*s) == startc && regtry(r, s, len))
                                                goto success;
                        } else {
                                for (len = relstrlen; len > 0; len--, s++)
                                        if(*s == startc && regtry(r, s, len))
                                                goto success;
                        }
                }
        else 

                /* We don't know anything special -- general case. */
                if(r->backward) {
                        for(len = relstrlen; len >= 0; len--)
                                if(regtry(r, relstr, len))
                                        goto success;
                } else {
                        for(s = relstr, len = relstrlen; len >= 0; len--, s++)
                                if(regtry(r, s, len))
                                        goto success;
                }

        /* fall into failure */
failure:
        if(end_node != NULL)
                *end_node = END;
        return EIF_FALSE;

success:
        if (end_node != NULL)
                *end_node = END;
        return EIF_TRUE;
}

static int regtry(Regexp_T *prog, char *string, EIF_INTEGER stringlen)
{
        /* Try match at specific point. Return 0 if fail, 1 if succeed */

        register1 int i;

        if(direction > 0) {                                             /* forward */
                reginput = string;
                reginputend = string + stringlen;
        } else {                                                                /* backward */
                reginput = string + stringlen - 1;
                reginputend = string - 1;
        }
        for(i = highest; i > 0; i--) {
                regstart[i] = 0;
                regend[i] = 0;
        }
        highest = 0;
        if(!regmatch(prog->program + 1))                /* did not match */
                return 0;
        if(reglen < 1)                                  /* no location arrays, but we matched */
                return 1;
        if(direction > 0) {                                     /* forward */
                regstart[0] = string - reginputbol + 1;
                regend[0] = reginput - reginputbol;
        } else {                                                                /* backward */
                regstart[0] = reginput - reginputeol + 1;
                regend[0] = string - reginputeol + stringlen - 1;
        }
        return 1;
}

static int      regmatch(char *prog)
{
        /* Main matching routine, return 0 if fail, 1 if succeed.
         * Conceptually the strategy is simple:  check to see whether the current
         * node matches, call self recursively to see whether the rest matches,
         * and then act accordingly.  In practice we make some effort to avoid
         * recursion, in particular by going through "ordinary" nodes (that don't
         * need to know whether the rest of the match failed) by a loop instead of
         * by recursion.
         */

        register1 char *scan;   /* Current node. */
        char *next;                             /* Next node. */

        scan = prog;

#ifdef DEBUG
        if(scan != NULL && regnarrate)
                fprintf(stderr, "%s(\n", regprop(scan));
#endif

        while(scan != NULL) {

#ifdef DEBUG
                if(regnarrate)
                        fprintf(stderr, "%s...\n", regprop(scan));
#endif

                next = regnext(scan);

                switch(OP(scan)) {
                        case BOL: {
                                register2 char *last = reginput - direction * NEWLINE_SIZE;

                                if(
                                        reginput == reginputbol || 
                                        (multi_line && (last - reginputbol) * direction >= 0 &&
                                        IS_NEWLINE_AT(last))
                                )
                                        break;  /* ok */
                                else 
                                        return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case EOL: 
                                if(
                                        reginput == reginputeol ||
                                        (multi_line && IS_NEWLINE_AT(reginput))
                                )
                                        break;  /* ok */
                                else
                                        return 0;
                                /* NOTREACHED */
                                break;
                        case BOW: {
                                register2 char *last = reginput - direction;

                                if(reginput == reginputbol)
                                        if(reginput != reginputeol && ISWORDCHAR(*reginput))
                                                break; /* ok */
                                        else
                                                return 0;
                                else 
                                        if(
                                                        reginput != reginputeol &&
                                                        ISWORDCHAR(*reginput) &&
                                                        !ISWORDCHAR(*last)
                                        )
                                                break;  /* ok */
                                        else
                                                return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case EOW: {
                                register2 char *last = reginput - direction;

                                if(reginput == reginputeol)
                                        if(reginput != reginputbol && ISWORDCHAR(*last))
                                                break;  /* ok */
                                        else
                                                return 0;
                                else
                                        if(
                                                        reginput != reginputbol &&
                                                        !ISWORDCHAR(*reginput) &&
                                                        ISWORDCHAR(*last)
                                        )
                                                break;  /* ok */
                                        else
                                                return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case ANY:
                                if(
                                                reginput == reginputend || 
                                                (multi_line && IS_NEWLINE_AT(reginput))
                                )
                                        return 0;
                                reginput += direction;
                                break;
                        case EXACTLY: {
                                register2 char *opnd;
                                register3 char *save;

                                save = reginput;
                                if(ignore_case) {
                                        for(
                                                        opnd = OPERAND(scan); 
                                                        *opnd && reginput != reginputend;
                                                        opnd++, reginput += direction
                                        )
                                                if(*opnd != toupper(*reginput)) {
                                                        reginput = save;
                                                        return 0;
                                                }
                                } else {
                                        for(
                                                        opnd = OPERAND(scan); 
                                                        *opnd && reginput != reginputend;
                                                        opnd++, reginput += direction)
                                                if(*opnd != *reginput) {
                                                        reginput = save;
                                                        return 0;
                                                }
                                }
                                if(*opnd) {
                                        save = reginput;
                                        return 0;
                                }
                                break;
                        }
                                /* NOTREACHED */
                                break;
                        case EXACTLYNUL:
                                if(reginput == reginputend || *reginput != '\0')
                                        return 0;
                                reginput += direction;
                                break;
                        case ANYOFNUL:
                                if(reginput == reginputend)
                                        return 0;
                                if(
                                        *reginput == '\0' ||
                                        strchr(OPERAND(scan),
                                                (ignore_case ? toupper(*reginput) : *reginput)) != NULL
                                ) {
                                        reginput += direction;
                                        break;
                                }
                                return 0;
                        case ANYOF:
                                if(
                                        reginput == reginputend ||
                                        *reginput == '\0' ||
                                        strchr(OPERAND(scan),
                                                (ignore_case ? toupper(*reginput) : *reginput)) == NULL
                                )
                                        return 0;
                                reginput += direction;
                                break;
                        case ANYBUTNUL:
                                if(
                                        reginput == reginputend || 
                                        *reginput == '\0' ||
                                        (multi_line && IS_NEWLINE_AT(reginput)) ||
                                        strchr(OPERAND(scan),
                                                (ignore_case ? toupper(*reginput) : *reginput)) != NULL
                                )
                                        return 0;
                                reginput += direction;
                                break;
                        case ANYBUT:
                                if(
                                        reginput == reginputend ||
                                        (multi_line && IS_NEWLINE_AT(reginput)) ||
                                        (*reginput != '\0' && strchr(OPERAND(scan),
                                                (ignore_case ? toupper(*reginput) : *reginput)) != NULL)
                                )
                                        return 0;
                                reginput += direction;
                                break;
                        case NOTHING:
                                break;
                        case BACK:
                                break;
                        case OPEN: {
                                register2 int no;
                                register3 char *save;

                                no = UCHARAT(OPERAND(scan));
                                save = reginput;
        
                                if(regmatch(next)) {
                                        /*
                                         * Don't set regstart if some later
                                         * invocation of the same parentheses
                                         * already has.
                                         */
                                        if(no < reglen && regstart[no] == 0) {
                                                regstart[no] = (direction > 0 ?
                                                                                save - reginputbol + 1 :
                                                                                save - reginputeol + 1);
                                                if(no > highest)
                                                        highest = no;
                                        }
                                        return 1;
                                } else
                                        return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case CLOSE: {
                                register2 int no;
                                register3 char *save;

                                no = UCHARAT(OPERAND(scan));
                                save = reginput;

                                if(regmatch(next)) {
                                        /*
                                         * Don't set regend if some later
                                         * invocation of the same parentheses
                                         * already has.
                                         */
                                        if(no < reglen && regend[no] == 0)
                                                regend[no] = (direction > 0 ?
                                                                                save - reginputbol :
                                                                                save - reginputeol);
                                        return 1;
                                } else
                                        return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case BRANCH: {
                                register2 char *save;

                                if(OP(next) != BRANCH)          /* No choice. */
                                        next = OPERAND(scan);   /* Avoid recursion. */
                                else {
                                        do {
                                                save = reginput;
                                                if(regmatch(OPERAND(scan)))
                                                        return 1;
                                                reginput = save;
                                                scan = regnext(scan);
                                        } while(scan != NULL && OP(scan) == BRANCH);
                                        return 0;
                                }
                                break;
                        }
                                /* NOTREACHED */
                                break;
                        case STAR:
                                /* FALLTHROUGH */
                        case PLUS: {
                                register2 char nextch;
                                register3 int no;
                                register4 char *save;
                                register5 int min;

                                /*
                                 * Lookahead to avoid useless match attempts
                                 * when we know what character comes next.
                                 */
                                nextch = '\0';
                                if(OP(next) == EXACTLY)
                                        nextch = *OPERAND(next);
                                min = (OP(scan) == STAR) ? 0 : 1;
                                save = reginput;
                                no = regrepeat(OPERAND(scan));
                                while(no >= min) {
                                        /* If it could work, try it. */
                                        if(
                                                nextch == '\0' || 
                                                (ignore_case ? toupper(*reginput) : *reginput) == nextch
                                        )
                                                if(regmatch(next))
                                                        return 1;
                                        /* Couldn't or didn't -- back up. */
                                        no--;
                                        reginput = save + direction * no;
                                }
                                return 0;
                        }
                                /* NOTREACHED */
                                break;
                        case END:
                                return 1;       /* Success! */
                        case ANCHEND:
                                if(reginput != reginputeol)
                                        return 0;
                                return 1;       /* Success! */
                        default:
                                raise_error("corrupted compiled pattern");
                                return 0;
                }

                scan = next;
        }

        /*
         * We get here only if there's trouble -- normally "case END" or
         * "case ANCHEND" is the terminating point.
         */
        raise_error("corrupted compiled pattern");
        return 0;
}

static int regrepeat(char *p)
{
        /* Repeatedly match something simple, report how many */

        register1 int count = 0;
        register2 char *scan;
        register3 char *opnd;
        register4 int dir = direction;

        scan = reginput;
        opnd = OPERAND(p);
        switch(OP(p)) {
                case ANY:
                        if(multi_line) 
                                while(scan != reginputend && ! IS_NEWLINE_AT(scan)) {
                                        count++;
                                        scan += dir;
                                }
                        else {
                                count = dir * (reginputend - scan);
                                scan = reginputend;
                        }
                        break;
                case EXACTLYNUL:
                        while(scan != reginputend && *scan == '\0') {
                                count++;
                                scan += dir;
                        }
                        break;
                case EXACTLY:
                        if(ignore_case)
                                while(scan != reginputend && *opnd == toupper(*scan)) {
                                        count++;
                                        scan += dir;
                                }
                        else
                                while(scan != reginputend && *opnd == *scan) {
                                        count++;
                                        scan += dir;
                                }
                        break;
                case ANYOF:
                        while(
                                scan != reginputend && *scan != '\0' && 
                                strchr(opnd, (ignore_case ? toupper(*scan) : *scan)) != NULL
                        ) {
                                count++;
                                scan += dir;
                        }
                        break;
                case ANYOFNUL:
                        while(
                                scan != reginputend &&
                                (*scan == '\0' || 
                                strchr(opnd, (ignore_case ? toupper(*scan) : *scan)) != NULL)
                        ) {
                                count++;
                                scan += dir;
                        }
                        break;
                case ANYBUT:
                        if(multi_line)
                                while(
                                        scan != reginputend && ! IS_NEWLINE_AT(scan) && 
                                        (
                                                *scan == '\0' ||
                                                strchr(opnd,
                                                        (ignore_case ? toupper(*scan) : *scan)) == NULL
                                        )
                                ) {
                                        count++;
                                        scan += dir;
                                }
                        else
                                while(
                                        scan != reginputend &&
                                        (
                                                *scan == '\0' ||
                                                strchr(opnd,
                                                        (ignore_case ? toupper(*scan) : *scan)) == NULL
                                        )
                                ) {
                                        count++;
                                        scan += dir;
                                }
                        break;
                case ANYBUTNUL:
                        if (multi_line)
                                while(
                                        scan != reginputend && ! IS_NEWLINE_AT(scan) && 
                                        *scan != '\0' &&
                                        strchr(opnd, (ignore_case ? toupper(*scan) : *scan)) == NULL
                                ) {
                                        count++;
                                        scan += dir;
                                }
                        else
                                while(
                                        scan != reginputend && *scan != '\0' &&
                                        strchr(opnd, (ignore_case ? toupper(*scan) : *scan)) == NULL
                                ) {
                                        count++;
                                        scan += dir;
                                }
                        break;
                default:                /* Oh dear.  Called inappropriately. */
                        raise_error("corrupted compiled pattern");
                        count = 0;      /* Best compromise. */
                        break;
        }
        reginput = scan;

        return count;
}

static char *regnext(register1 char *p)
{
        /* Determine the "next" pointer from a node */

        register2 int offset;

        if(regsize > regmaxsize)
                return NULL;

        offset = NEXT(p);
        if(offset == 0)
                return NULL;

        if(OP(p) == BACK)
                return p - offset;
        else
                return p + offset;
}

static char *contains   (
        register1 char *s1, long s1len, register3 char *s2, register4 long s2len
)
{
        /* Does `s1' contain `s2'? 
         * Global variable `ignore_case' says, if we shall ignore case. If we do,
         * we depend on `s2' having only upper case letters.
         * Global variable `direction' says, if we have a "backward" compiled
         * pattern.
         * Return position of `s2' in `s1' (pointer to first character in
         * search direction) or NULL if not found.
         */

        register5 long i;
        register6 char *se;

        if(direction < 0)       /* backward */
                if(ignore_case)
                        for(se = s1 + s2len - 1, s1 += s1len - 1; s1 >= se; s1--) {
                                while(toupper(*s1) != *s2 && s1 > se)
                                        s1--;
                                for(i = 0; i < s2len && toupper(s1[-i]) == s2[i]; i++)
                                        /* empty */ ;
                                if(i == s2len)
                                        return s1;
                        }
                else
                        for(se = s1 + s2len - 1, s1 += s1len - 1; s1 >= se; s1--) {
                                while(*s1 != *s2 && s1 > se)
                                        s1--;
                                for(i = 0; i < s2len && s1[-i] == s2[i]; i++)
                                        /* empty */ ;
                                if(i == s2len)
                                        return s1;
                        }
        else                            /* forward */
                if (ignore_case)
                        for(se = s1 + s1len - s2len; s1 <= se; s1++) {
                                while(toupper(*s1) != *s2 && s1 < se)
                                        s1++;
                                for(i = 0; i < s2len && toupper(s1[i]) == s2[i]; i++)
                                        /* empty */ ;
                                if(i == s2len)
                                        return s1;
                        }
                else
                        for(se = s1 + s1len - s2len; s1 <= se; s1++) {
                                while(*s1 != *s2 && s1 < se)
                                        s1++;
                                for(i = 0; i < s2len && s1[i] == s2[i]; i++)
                                        /* empty */ ;
                                if(i == s2len)
                                        return s1;
                        }

        return NULL;
}

static int strncspn(char *s1, int s1len, char *s2)
{
        /* Find length of initial segment of `s1' consisting entirely
         * of characters not from `s2'. This differs from the standard
         * strcspn() that length of s1 is given by `s1len', not by
         * a '\0' character, so `s1' may include any '\0' characters.
         */

        register1 char *scan1;
        register2 char *scan2;
        register3 char *scan1end;
        register4 int count;

        count = 0;
        scan1end = s1 + s1len;
        for(scan1 = s1; scan1 < scan1end; scan1++) {
                for(scan2 = s2; *scan2 != '\0';)        /* ++ moved down. */
                        if(*scan1 == *scan2++)
                                return count;
                count++;
        }
        return count;
}

#ifdef DEBUG

/* The following code enables you to have a look on the structure of your
 * automaton. It is very useful for debugging purposes. If you are in doubt
 * wether a specific pattern should (or should not) match, while in fact it
 * does not (resp. does), compile this code with DEBUG set and call regdump()
 * after compilation of the pattern.
 */

static char *regprop(char *op);

void regdump(char *prog)
{
        /* Dump a regexp onto stdout in vaguely comprehensible form */

        register1 Regexp_T *r;
        register2 char *s;
        register3 char op = EXACTLY;    /* Arbitrary non-END op. */
        register4 char *next;


        r = (Regexp_T *) prog;
        s = r->program + 1;
        while(op != END && op != ANCHEND) {
                /* While that wasn't END last time... */
                op = OP(s);
                printf("%2d%s", s-r->program, regprop(s));      /* Where, what. */
                next = regnext(s);
                if(next == NULL)                /* Next ptr. */
                        printf("(0)");
                else 
                        printf("(%d)", (s - r->program) + (next - s));
                s += 3;
                if(op == ANYOF || op == ANYBUT || op == EXACTLY || 
                        op == ANYOFNUL || op == ANYBUTNUL) {
                        /* Literal string, where present. */
                        while(*s != '\0') {
                                putchar(*s);
                                s++;
                        }
                        s++;
                }
                else if(op == OPEN || op == CLOSE)
                        s++;
                putchar('\n');
        }

        /* Header fields of interest. */
        if(r->backward)
                printf("bachward compiled, ");
        if(r->ignorecase)
                printf("case insensitive, ");
        if(r->firstchar != '\0')
                printf("start `%c' ", r->firstchar);
        if(r->anchored)
                printf("anchored ");
        s = r->program + MUST(r->musthave);
        if(s > r->program)
                printf("must have \"%s\"", s);
        printf("\n");
}

static char *regprop(char *op)
{
        /* Printable representation of opcode */

        register1 char *p;
        static char buf[50];

        (void) strcpy(buf, ":");

        switch(OP(op)) {
                case BOL:
                        p = "BOL";
                        break;
                case EOL:
                        p = "EOL";
                        break;
                case BOW:
                        p = "BOW";
                        break;
                case EOW:
                        p = "EOW";
                        break;
                case ANY:
                        p = "ANY";
                        break;
                case ANYOFNUL:
                        p = "ANYOFNUL";
                        break;
                case ANYBUTNUL:
                        p = "ANYBUTNUL";
                        break;
                case ANYOF:
                        p = "ANYOF";
                        break;
                case ANYBUT:
                        p = "ANYBUT";
                        break;
                case BRANCH:
                        p = "BRANCH";
                        break;
                case EXACTLYNUL:
                        p = "EXACTLYNUL";
                        break;
                case EXACTLY:
                        p = "EXACTLY";
                        break;
                case NOTHING:
                        p = "NOTHING";
                        break;
                case BACK:
                        p = "BACK";
                        break;
                case END:
                        p = "END";
                        break;
                case OPEN:
                        sprintf(buf + strlen(buf), "OPEN%d", UCHARAT(OPERAND(op)));
                        p = NULL;
                        break;
                case CLOSE:
                        sprintf(buf + strlen(buf), "CLOSE%d", UCHARAT(OPERAND(op)));
                        p = NULL;
                        break;
                case STAR:
                        p = "STAR";
                        break;
                case PLUS:
                        p = "PLUS";
                        break;
                case ANCHEND:
                        p = "ANCHEND";
                        break;
                default:
                        raise_error("corrupted compiled pattern");
                        break;
        }
        if(p != NULL)
                (void) strcat(buf, p);

        return buf;
}
#endif
