/* LINTLIBRARY */

#include <ctype.h>

#include "eif_match.h"

typedef unsigned char *DTableC;	/* type of the small shift table */
typedef EIF_INTEGER *DTableI; 	/* type of the large shift table */

/* routines for small shift table */

	/* Search backward case sensitive */
static EIF_POINTER s_search_b_cs	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableC dt
);
	/* Search forward case sensitive */
static EIF_POINTER s_search_f_cs	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableC dt
);
	/* Search backward case insensitive */
static EIF_POINTER s_search_b_ic	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableC dt
);
	/* Search forward case insensitive */
static EIF_POINTER s_search_f_ic	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableC dt
);

/* routines for large shift table */

	/* Search backward case sensitive */
static EIF_POINTER l_search_b_cs	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableI dt
);
	/* Search forward case sensitive */
static EIF_POINTER l_search_f_cs	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableI dt
);
	/* Search backward case insensitive */
static EIF_POINTER l_search_b_ic	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableI dt
);
	/* Search forward case insensitive */
static EIF_POINTER l_search_f_ic	(
	unsigned char *text, int tlen, unsigned char *pattern, int plen, DTableI dt
);

void eif_match_sub_compile	(
	EIF_POINTER pattern, 			/* The pattern we want to compile */
	register1 EIF_INTEGER plen,		/* The length of the pattern */
	EIF_POINTER dtable,				/* The shifting table */
	register2 EIF_BOOLEAN ignore,	/* Is the search case insensitive? */
	EIF_BOOLEAN backward			/* Compile for backward search? */
)
{
	/* Compile `pattern' by computing the shifting table and store it in
	 * dtable for later use. `ignore' says, if we shall ignore letter case.
	 * If that is the case, we depend on pattern being already downshifted!
	 * `backward' says, if the shifting table is for a backward search.
	 * Note: A shifting table can only be used for letter case and search
	 * direction it was compiled for.
	 */

	register4 int j;
	register5 int direction = 1;	/* default: forward = increment */
	register3 unsigned char *p;

	if(backward) {
		direction = -1;			/* backward = decrement */
		pattern = (char *) pattern + plen - 1;	/* go to end of pattern */
	}
	if(plen++ <= SMALLPLEN) {	/* increment `plen'!!! */

		/* If `plen' is small enough, shifting values will fit in a character,
		 * so we (hopefully) got a character array from the Eiffel side.
		 */
		register6 DTableC dt = (DTableC) dtable;

		for(j = 0; j < ALPHASIZE; j++) 
			dt[j] = (unsigned char) plen;	/* init with pattern length + 1 */
		plen--;				/* restore pattern length */
		for(
				p = (unsigned char *) pattern, j = plen;
				j > 0;
				p += direction, j--
		) {
			dt[*p] = (unsigned char) j;
			if(ignore)
				dt[toupper(*p)] = (unsigned char) j;
		}
	} else {
		/* Length of `pattern' is to large, that shifting values can be stored
		 * in a character array, so we assume that we got an integer array.
		 */
		register6 DTableI dt = (DTableI)dtable;

		for(j = 0; j < ALPHASIZE; j++) 
			dt[j] = plen;
		plen--;
		for(
				p = (unsigned char *) pattern, j = plen; 
				j > 0;
				p += direction, j--
		) {
			dt[*p] = j;
			if(ignore)
				dt[toupper(*p)] = j;
		}
	}
}

EIF_INTEGER eif_match_sub_search	(
	EIF_POINTER text, 		/* The text we are searching through */
	EIF_INTEGER from,		/* The relevant range inside text */
	EIF_INTEGER to, 		/* (from .. to) */
	EIF_POINTER pattern,	/* The pattern we are looking for */
	EIF_INTEGER plen, 		/* The length of the pattern */
	EIF_POINTER dtable,		/* The pattern in compiled form */
	EIF_BOOLEAN ignore, 	/* Is the search case insensitive? */
	EIF_BOOLEAN backward	/* Search backward? */
)
{
	/* Search in `text' for `pattern'. Search only in range `from' - `to'.
	 * We depend on `from' and `to' denoting a valid range (which will be
	 * guaranteed by an appropriate precondition on the Eiffel side).
	 * `dtable' is the shifting table of the pattern, and as you remeber
	 * from the comment in eif_match_sub_compile, must be consistent with
	 * the search direction and letter case.
	 */

	char *p;

	from--;					/* Convert index to C one (start at 0) */

	/* Dispatch function call */

	if(plen <= SMALLPLEN)
		if(ignore) 
			if(backward) 
				p = s_search_b_ic((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableC) dtable);
			else 
				p = s_search_f_ic((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableC) dtable);
		else 
			if(backward) 
				p = s_search_b_cs((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableC) dtable);
			else 
				p = s_search_f_cs((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableC) dtable);
	else
		if(ignore) 
			if(backward) 
				p = l_search_b_ic((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableI) dtable);
			else 
				p = l_search_f_ic((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableI) dtable);
		else
			if(backward)
				p = l_search_b_cs((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableI) dtable);
			else
				p = l_search_f_cs((unsigned char *) text + from, to - from,
					(unsigned char *) pattern, plen, (DTableI) dtable);

	if(p == NULL)
		return 0;	/* not found */

	return (1 + p - (char *) text);	/* return index, starting at 1 */
}

/* Because there are lots of searching functions and they are all so similar we
 * generate them from a template instead of writing them by hand. This is much
 * more convenient, since we have to make changes only once in the template.
 */

/*
 * Generation of searching routines
 */

/* For small shifting tables ... */

#define DTABLE			DTableC
#define CASE_ADAPT(c)	(c)
#define BACKWARD
#define SEARCH			s_search_b_cs
#include "substring.t"
#undef BACKWARD
#undef SEARCH
#define SEARCH			s_search_f_cs
#include "substring.t"
#undef CASE_ADAPT
#define CASE_ADAPT(c)	(tolower(c))
#define BACKWARD
#undef SEARCH
#define SEARCH			s_search_b_ic
#include "substring.t"
#undef BACKWARD
#undef SEARCH
#define SEARCH			s_search_f_ic
#include "substring.t"

/* ... and again for the large ones ... */

#undef DTABLE
#define DTABLE			DTableI
#undef CASE_ADAPT
#define CASE_ADAPT(c)	(c)
#define BACKWARD
#undef SEARCH
#define SEARCH			l_search_b_cs
#include "substring.t"
#undef BACKWARD
#undef SEARCH
#define SEARCH			l_search_f_cs
#include "substring.t"
#undef CASE_ADAPT
#define CASE_ADAPT(c)	(tolower(c))
#define BACKWARD
#undef SEARCH
#define SEARCH			l_search_b_ic
#include "substring.t"
#undef BACKWARD
#undef SEARCH
#define SEARCH			l_search_f_ic
#include "substring.t"

