/*
 * The search template for substring.
 */

/* Searching has to distinguish lots of cases, eg:
 *	- forward or backward search
 *	- case sensitive or insensitive search
 *	- small distance table (based on chars) or large one (based on ints)
 * Since it's speed what counts here, we cannot distinuish the cases in every
 * simple step (inside the searching routine) but must dispatch to different
 * searching routines at the top level. 
 * Since all cases must be combined we get a lot of routines, which are almost
 * identical. To ease maintenance of the code, all routines are generated
 * from the following template.
 *
 * To build the routines, you have to define the following macros:
 *	- define SEARCH to the name of the search routine
 *	- define DTABLE to the type of distance table used
 *	- define BACKWARD if routine is for backward search
 *	- define CASE_ADAPT() as null macro, if case sensitive, else as tolower()
 */

static EIF_POINTER SEARCH	(
	unsigned char *text, 		/* The text we are searching through */
	int tlen,					/* Length of the text */
	unsigned char *pattern, 	/* The pattern string */
	register5 int plen,			/* The length of the pattern */
	DTABLE dt					/* The shift table for the pattern */
)
{
	/* The search algorithm. It returns the position of `pattern' in `text'
	 * if pattern is found, else NULL.
	 */

	register1 unsigned char *p;		/* Pattern string pointer */
	register2 unsigned char *t;		/* Text pointer */
	register4 unsigned char *tx;	/* Another text pointer (start of search) */
	register3 int i;				/* Position within pattern string */

#ifdef BACKWARD
	tx = text + tlen - 1;	/* Where we start scanning */
	text += plen - 1;		/* Last address where matching can occur */
	while(tx >= text) {		/* While enough text still left */

		for(p = pattern + plen - 1, t = tx, i = 0; i < plen; p--, t--, i++)
			if(*p != CASE_ADAPT(*t)) 	/* Scan pattern string */
				break;
		if(i == plen) 				/* Got substring (whole pattern matched) */
			return (EIF_POINTER) (t + 1);
		tx -= dt[*(tx - plen)];		/* Shift to next text location */
	}
#else
	tx = text;				/* Where we start scanning */
	text += tlen - plen;	/* Last address where matching can occur */
	while(tx <= text) {	/* While enough text still left */

		for(p = pattern, t = tx, i = 0; i < plen; p++, t++, i++)
			if(*p != CASE_ADAPT(*t)) 	/* Scan pattern string */
				break;
		if(i == plen) 				/* Got substring (whole pattern matched) */
			return (EIF_POINTER) tx;
		tx += dt[*(tx + plen)];		/* Shift to next text location */
	}
#endif
	return NULL;
}
