indexing

	description:
		"Misspelled substring matcher";
	store_and_retrieve: "yes"

class MISSPELLED

inherit

	MATCHER
		rename
			make as matcher_make,
			make_sensitive as matcher_make_sensitive,
			make_insensitive as matcher_make_insensitive
		export
			{ NONE } matcher_make, matcher_make_sensitive,
					 matcher_make_insensitive
		end

	MATCHER
		redefine
			make, make_sensitive, make_insensitive
		select
			make, make_sensitive, make_insensitive
		end

creation
	make, make_sensitive, make_insensitive,
	make_misspelled, make_sensitive_misspelled, make_insensitive_misspelled

feature -- Initialization

	make (p: STRING) is
			-- Make a misspelled matcher from search pattern `p'.
			-- Case sensitivity (which is enabled at object creation)
			-- is not modified. The maximum number of misspelled characters
			-- depends on the patterns length and is taken from `default_blur'.
		do
			make_misspelled (p, default_blur_of(p))
		ensure then
			compiled: compiled
		end

	make_sensitive (p: STRING) is
			-- Make a misspelled matcher from search pattern `p' that is case
			-- sensitive. The blurredness is taken from `default_blur'.
		do
			make_sensitive_misspelled (p, default_blur_of(p))
		ensure then
			compiled: compiled
		end

	make_insensitive (p: STRING) is
			-- Make a misspelled matcher from search pattern `p' that is case
			-- insensitive. The blurredness is taken from `default_blur'.
		do
			make_insensitive_misspelled (p, default_blur_of(p))
		ensure then
			compiled: compiled
		end

	make_misspelled (p: STRING; mm: INTEGER) is
			-- Make a misspelled matcher from search pattern `p' with at most
			-- `mm' misspelled characters. Case sensitivity (which is enabled
			-- at object creation) is not modified.
		require
			pattern_defined: p /= Void
			blur_defined: mm >= 0 and mm <= p.count
		do
			if mm > blur then
				pattern := Void		-- Force recompilation
			end
			blur := mm
			matcher_make (p)
		ensure
			compiled: compiled
		end

	make_sensitive_misspelled (p: STRING; mm: INTEGER) is
			-- Make a misspelled matcher from search pattern `p' with at most
			-- `mm' misspelled characters. The matcher is case sensitive.
		require
			pattern_defined: p /= Void
			blur_defined: mm >= 0 and mm <= p.count
		do
			if mm > blur then
				pattern := Void		-- Force recompilation
			end
			blur := mm
			matcher_make_sensitive (p)
		ensure
			compiled: compiled
		end

	make_insensitive_misspelled (p: STRING; mm: INTEGER) is
			-- Make a misspelled matcher from search pattern `p' with at most
			-- `mm' misspelled characters. The matcher is case insensitive.
		require
			pattern_defined: p /= Void
			blur_defined: mm >= 0 and mm <= p.count
		do
			if mm > blur then
				pattern := Void		-- Force recompilation
			end
			blur := mm
			matcher_make_insensitive (p)
		ensure
			compiled: compiled
		end

feature -- Modification

	set_blur (mm: INTEGER) is
			-- Set the maximum number of allowed mismatches.
		require
			blur_defined: mm >= 0 and mm <= pattern.count
		do
			make_misspelled (pattern, mm)
		ensure
			blur_set: blur = mm
		end

	set_sensitive is
			-- Make searcher case-sensitive.
		local
			cp: like dtable
		do
			if insensitive then
				insensitive := false
				cp := saved_dtable
				saved_dtable := dtable
				dtable := cp
				cp := saved_dtable_b
				saved_dtable_b := dtable_b
				dtable_b := cp
				if dtable = Void then
					compile
				end
			end
		end

	set_insensitive is
			-- Make searcher case-insensitive.
		local
			cp: like dtable
		do
			if not insensitive then
				insensitive := true
				cp := saved_dtable
				saved_dtable := dtable
				dtable := cp
				cp := saved_dtable_b
				saved_dtable_b := dtable_b
				dtable_b := cp
				if dtable = Void then
					compile
				end
			end
		end

	set_multi_line is
			-- Set matcher in multi-line mode. Mode is kept until matcher is
			-- explicitely set to single-line mode.
		do
			multi_line := true
		end

	set_single_line is
			-- Set matcher in single-line mode. Mode is kept until matcher is
			-- explicitely set to multi-line mode.
		do
			multi_line := false
		end

feature  -- Status report

	multi_line: BOOLEAN
			-- Is matcher in multi-line mode?

	single_line: BOOLEAN is
			-- Is matcher in single-line mode? This is the default on object
			-- creation.
		do
			Result := not multi_line
		end

	blur: INTEGER
			-- Maximum number of allowed mismatches.

	default_blur: INTEGER is
			-- Default blurredness of the matcher.
		do
			Result := default_blur_of(pattern)
		end

	mismatches: INTEGER
			-- If `found', actual number of mismatches.

feature -- Searching

    search_forward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_from' and search until the pattern matches (while
            -- ignoring any characters behind `s_to'). If pattern matches, set
            -- `found' to true, `mismatches' to the number of mismatches that
			-- occured and `start_position' and `end_position' to the
            -- corresponding index positions (relative to the beginning
            -- of the string). If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position', `end_position' or
			-- `mismatches'.
		require else
			empty_range: s_from - 1 = s_to;
        local
            pattern_area: POINTER
            flag: INTEGER
        do
            if insensitive then
                pattern_area := downshifted_pattern.to_c
				flag := Mis_ignore_case
            else
                pattern_area := pattern.to_c
            end
			if multi_line then
				flag := flag + Mis_multi_line
			end
			found := mis_search_forw(s.to_c, s_from, s_to,
            			pattern_area, pattern.count,
                        dtable.to_c, blur, flag,
                        $start_position, $end_position, $mismatches)
        end

    search_backward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_to' and search backward until the pattern matches
            -- (ignoring any characters before `s_from'). If pattern matches,
            -- set `found' to true, `mismatches' to the number of mismatches
			-- that occured and `start_position' and `end_position' to the
            -- corresponding index positions (relative to the beginning
            -- of the string). If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position', `end_position' or
			-- `mismatches'.
		require else
			empty_range: s_from - 1 = s_to;
        local
            pattern_area: POINTER
            flag: INTEGER
        do
            if dtable_b = Void then
                compile_backw
            end
            if insensitive then
                pattern_area := downshifted_pattern.to_c
					 flag := Mis_ignore_case
            else
                pattern_area := pattern.to_c
            end
			if multi_line then
				flag := flag + Mis_multi_line
			end
			found := mis_search_back(s.to_c, s_from, s_to,
            			pattern_area, pattern.count,
                        dtable_b.to_c, blur, flag,
                        $start_position, $end_position, $mismatches)
        end

feature  -- Matching actions

    match (s: STRING) is
            -- Match `pattern' against string `s'. The whole string must
            -- match. Set attributes `found' and - if match succeeded -
            -- `start_position', `end_position' and `mismatches'.
        do
			if s.count /= pattern.count then
				found := false
			else
				search_forward(s, 1, s.count)
			end
        end

    match_prefix (s: STRING) is
            -- Match `pattern' against prefix of string `s'.
            -- Matching is forward in `s'. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position' and
			-- `mismatches'.
        do
			if s.count < pattern.count then
				found := false
			else
				search_forward(s, 1, pattern.count)
			end
        end

    match_suffix (s: STRING) is
            -- Match `pattern' against suffix of string `s'.
            -- Matching is backward in `s'. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position' and
			-- `mismatches'.
        do
			if s.count < pattern.count then
				found := false
			else
				search_backward(s, s.count - pattern.count + 1, s.count)
			end
        end

    match_left_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
            -- Matching is from left to right. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position' and
			-- `mismatches'.
        do
			search_forward(s, 1, s.count)
        end

    match_right_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
            -- Matching is from right to left. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position' and
			-- `mismatches'.
        do
			search_backward(s, 1, s.count)
        end

feature -- Matching questions

    matches (s: STRING): BOOLEAN is
            -- Does string `s' matches with pattern? The whole string must
            -- match. None of the position attributes nor `found' is modified.
        local
            pattern_area: POINTER
            flag, plen: INTEGER
        do
            plen := pattern.count
			if plen /= s.count then
				Result := false
			else
				if insensitive then
					pattern_area := downshifted_pattern.to_c
					flag := Mis_ignore_case
				else
					pattern_area := pattern.to_c
				end
				if multi_line then
					flag := flag + Mis_multi_line
				end
				Result := mis_search_forw(s.to_c, 1, plen,
                		pattern_area, plen,
                        dtable.to_c, blur, flag,
                        default_pointer, default_pointer, default_pointer)
			end
        end

    matches_prefix (s: STRING): BOOLEAN is
            -- Does any prefix of string `s' matches with pattern?
            -- Neither a position attribute nor `found' is modified.
        local
            pattern_area: POINTER
            flag, plen: INTEGER
        do
            plen := pattern.count
			if s.count < plen then
				found := false
			else
				if insensitive then
					pattern_area := downshifted_pattern.to_c
					flag := Mis_ignore_case
				else
					pattern_area := pattern.to_c
				end
				if multi_line then
					flag := flag + Mis_multi_line
				end
				Result := mis_search_forw(s.to_c, 1, plen,
                		pattern_area, plen,
                        dtable.to_c, blur, flag,
                        default_pointer, default_pointer, default_pointer)
			end
        end

    matches_suffix (s: STRING): BOOLEAN is
            -- Does any suffix of string `s' matches with pattern?
            -- Neither a position attribute nor `found' is modified.
        local
            pattern_area: POINTER
            flag, slen, plen: INTEGER
        do
            plen := pattern.count
			slen := s.count
			if slen < plen then
				Result := false
			else
				if dtable_b = Void then
					compile_backw
				end
				if insensitive then
					pattern_area := downshifted_pattern.to_c
					flag := Mis_ignore_case
				else
					pattern_area := pattern.to_c
				end
				if multi_line then
					flag := flag + Mis_multi_line
				end
				Result := mis_search_back(s.to_c, slen - plen + 1, slen,
						pattern_area, plen,
                        dtable_b.to_c, blur, flag,
						default_pointer, default_pointer, default_pointer)
            end
        end

    matches_left_part (s: STRING): BOOLEAN is
            -- Does any substring of `s' matches with pattern?
            -- Neither a position attribute nor `found' is modified.
        local
            pattern_area: POINTER
			flag: INTEGER
        do
			if insensitive then
				pattern_area := downshifted_pattern.to_c
				flag := Mis_ignore_case
			else
				pattern_area := pattern.to_c
			end
			if multi_line then
				flag := flag + Mis_multi_line
			end
			Result := mis_search_forw(s.to_c, 1, s.count,
            			pattern_area, pattern.count,
                        dtable.to_c, blur, flag,
						default_pointer, default_pointer, default_pointer)
        end

    matches_right_part (s: STRING): BOOLEAN is
            -- Does any substring of `s' matches with pattern?
            -- Neither a position attribute nor `found' is modified.
        local
            pattern_area: POINTER
			flag: INTEGER
        do
			if dtable_b = Void then
				compile_backw
			end
			if insensitive then
				pattern_area := downshifted_pattern.to_c
				flag := Mis_ignore_case
			else
				pattern_area := pattern.to_c
			end
			if multi_line then
				flag := flag + Mis_multi_line
			end
			Result := mis_search_back(s.to_c, 1, s.count,
            			pattern_area, pattern.count,
                        dtable_b.to_c, blur, flag,
						default_pointer, default_pointer, default_pointer)
        end

feature { NONE }  -- Implementation

	dtable, dtable_b, saved_dtable, saved_dtable_b: MEMORY_BLOCK
			-- Distance tables areas
            -- These are actually 2-dimensional arrays, row-major,
            -- with blur+1 rows and Alpha_size columns.
            -- We don't need to deal with them as 2-dimensional arrays,
            -- and defining them as ARRAY2[G] introduces nuisances
            -- depending on whether the ISE EiffelBase cluster is
            -- part of our system or not.  Let's just avoid that...

	downshifted_pattern: like pattern
			-- Pattern for case-insensitive search

	wipe_out is
		do
			dtable := Void
			dtable_b := Void
			saved_dtable := Void
			saved_dtable_b := Void
			downshifted_pattern := Void
			compiled := false
		end

	compile is
			-- Compile `pattern' in forward-search mode.
		do
			compile_it(false)
		end

	compile_backw is
			-- Compile `pattern' in backward-search mode.
		do
			compile_it(true)
		end

	compile_it(backw: BOOLEAN) is
			-- Compile `pattern'.
		local
			pattern_area: POINTER
			table: MEMORY_BLOCK
		do
			if insensitive then
				if downshifted_pattern = Void then
					downshifted_pattern := clone(pattern)
					downshifted_pattern.to_lower
				end
				pattern_area := downshifted_pattern.to_c
			else
				pattern_area := pattern.to_c
			end
			if (pattern.count <= Small_pattern_size) then
				!ARRAY_MEMORY_BLOCK[CHARACTER]!table.make((blur + 1) * Alpha_size)
			else
				!ARRAY_MEMORY_BLOCK[INTEGER]!table.make((blur + 1) * Alpha_size)
			end
            if (backw) then
            	dtable_b := table
            else
            	dtable := table
            end
			mis_compile(pattern_area, pattern.count,
            	table.to_c, blur, insensitive, backw)
            compiled := true;
		end

	default_blur_of (p: STRING): INTEGER is
			-- Default blur of a pattern `p'.
		local
			l: INTEGER
		do
			l := p.count
			if l <= 3 then
				Result := 0
			else
				Result := (l * 0.11).ceiling
			end
		end

feature {NONE}  -- Externals

    Alpha_size: INTEGER is
            -- Number of characters in the alphabet
        external
            "C"
        alias
            "eif_match_ALPHASIZE"
        end

    Small_pattern_size: INTEGER is
            -- Maximum pattern length for character array usage
        external
            "C"
        alias
            "eif_match_SMALLPLEN"
        end

    Mis_ignore_case: INTEGER is
            -- Flag for case insensitive search
        external
            "C"
        alias
            "eif_match_IGN_CASE"
        end

    Mis_multi_line: INTEGER is
            -- Flag for multi-line mode
        external
            "C"
        alias
            "eif_match_MULTILINE"
        end

    mis_compile (patt: POINTER; plen: INTEGER; dt: POINTER; b: INTEGER;
        ign, back: BOOLEAN) is
            -- Compile substring pattern `patt', which has length `plen'
            -- and store result in area pointed to by `dt'. `ign' says
            -- if we shall compile for case insensite, `back' if we compile
            -- for a backward search distance table.
        external
            "C"
		alias
			"eif_match_mis_compile"
        end

    mis_search_forw (text: POINTER; t_from, t_to: INTEGER; patt: POINTER;
        plen: INTEGER; dt: POINTER; b: INTEGER; flag: INTEGER;
		sp, ep, mm: POINTER): BOOLEAN is
            -- Search in range `t_from' to `t_to' in `text' for pattern
            -- pointed to by `patt' (with length `plen') using distance
            -- table `dt'. `flag' says, if we shall ignore letter case,
			-- and if we are in multi-line mode;
            -- `back' says, if we shall do a backward search.
            -- Return index in `text' where pattern starts (if found) else 0.
        external
            "C"
		alias
			"eif_match_mis_search_forw"
        end

    mis_search_back (text: POINTER; t_from, t_to: INTEGER; patt: POINTER;
        plen: INTEGER; dt: POINTER; b: INTEGER; flag: INTEGER;
		sp, ep, mm: POINTER): BOOLEAN is
            -- Search in range `t_from' to `t_to' in `text' for pattern
            -- pointed to by `patt' (with length `plen') using distance
            -- table `dt'. `flag' says, if we shall ignore letter case,
			-- and if we are in multi-line mode;
            -- `back' says, if we shall do a backward search.
            -- Return index in `text' where pattern starts (if found) else 0.
        external
            "C"
		alias
			"eif_match_mis_search_back"
        end

end  -- class MISSPELLED
