indexing

   description:
      "Regular expression matcher";
   store_and_retrieve: "yes"

class REGEXP

inherit

   MATCHER

creation

   make, make_sensitive, make_insensitive

feature -- Modification

   set_sensitive is
         -- Make the matcher case-sensitive.
      local
         cp: like cpatt
      do
         if insensitive then
            insensitive := false
            cp := saved_cpatt
            saved_cpatt := cpatt
            cpatt := cp
            cp := saved_cpatt_b
            saved_cpatt_b := cpatt_b
            cpatt_b := cp
            if cpatt = Void then
               compile
            end
         end
      end

   set_insensitive is
         -- Make the matcher case-insensitive.
      local
         cp: like cpatt
      do
         if not insensitive then
            insensitive := true
            cp := saved_cpatt
            saved_cpatt := cpatt
            cpatt := cp
            cp := saved_cpatt_b
            saved_cpatt_b := cpatt_b
            cpatt_b := cp
            if cpatt = Void then
               compile
            end
         end
      end

   set_multi_line is
         -- Set matcher in multi-line mode. Mode is kept until matcher is
         -- explicitely set to single-line mode.
      do
         multi_line := true
      end

   set_single_line is
         -- Set matcher in single-line mode. Mode is kept until matcher is
         -- explicitely set to multi-line mode.
      do
         multi_line := false
      end

feature -- Status report

   multi_line: BOOLEAN
         -- Is matcher in multi-line mode?

   single_line: BOOLEAN is
         -- Is matcher in single-line mode? This is the default on object
         -- creation.
      do
         Result := not multi_line
      end

   number_of_subexpr: INTEGER
         -- If `compiled' is true, number of subexpressions in `pattern'.

   start_of_subexpr(i: INTEGER): INTEGER is
         -- Start position of the `i'th subexpression in the search string.
         -- If `i'th subexpression was not part of the match 0 is returned.
      require
         successful_search: found
         valid_index: i >= 1 and i <= number_of_subexpr
      do
         Result := start_loc.item(i)
      ensure
         part_of_match: Result = 0 or else
            (Result >= start_position and Result <= end_position)
      end

   end_of_subexpr(i: INTEGER): INTEGER is
         -- End position of the `i'th subexpression in the search string.
         -- If `i'th subexpression was not part of the match 0 is returned.
      require
         successful_search: found
         valid_index: i >= 1 and i <= number_of_subexpr
      do
         Result := end_loc.item(i)
      ensure
         part_of_match: Result = 0 or else
            (Result >= start_position and Result <= end_position)
      end

feature -- Searching

   search_forward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_from' and search until the pattern matches (while
            -- ignoring any characters behind `s_to'). If pattern matches, set
            -- `found' to true and `start_position' and `end_position'
            -- to the corresponding index positions (relative to the beginning
            -- of the string). If the pattern contains subexpressions (r.e.'s
         -- enclosed in `()') also record their start and end positions
         -- in `s'. If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position' or `end_position'.
      do
         search(s, s_from, s_to, cpatt)
      end

   search_backward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_to' and search backward until the pattern matches
            -- (ignoring any characters before `s_from'). If pattern matches,
            -- set `found' to true and `start_position' and `end_position'
            -- to the corresponding index positions (relative to the beginning
            -- of the string). If the pattern contains subexpressions (r.e.'s
            -- enclosed in `()') also record their start and end positions
            -- in `s'. If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position' or `end_position'.
      do
         if cpatt_b = Void then
            compile_backw
         end
         search(s, s_from, s_to, cpatt_b)
      end

feature -- Matching actions

   match (s: STRING) is
            -- Match `pattern' against string `s'. The whole string must
            -- match. Set attributes `found' and - if match succeeded -
            -- `start_position' and `end_position', and record start and end
         -- position of any matched subexpression.
      do
         match_against(s, cpatt, Reg_prefix + Reg_suffix)
      end

   match_prefix (s: STRING) is
         -- Match `pattern' against any prefix of string `s'.
         -- Matching is forward in `s'. Set attributes `found' and -
         -- if match succeeded - `start_position' and `end_position'
         -- and record start and end position of any matched subexpression.
      do
         match_against(s, cpatt, Reg_prefix)
      end

   match_suffix (s: STRING) is
         -- Match `pattern' against any suffix of string `s'.
            -- Matching is backward in `s'. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'
            -- and record start and end position of any matched subexpression.
      do
         if cpatt_b = Void then
            compile_backw
         end
         match_against(s, cpatt_b, Reg_suffix)
      end

   match_left_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
         -- Matching is from left to right. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'
            -- and record start and end position of any matched subexpression.
      do
         match_against(s, cpatt, 0)
      end

   match_right_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
         -- Matching is from right to left. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'
            -- and record start and end position of any matched subexpression.
      do
         if cpatt_b = Void then
            compile_backw
         end
         match_against(s, cpatt_b, 0)
      end

feature -- Matching questions

   matches (s: STRING): BOOLEAN is
            -- Does string `s' matches with pattern? The whole string must
         -- match. None of the position attributes nor `found' is modified.
      do
         Result := matches_against(s, cpatt, Reg_prefix + Reg_suffix)
      end

   matches_prefix (s: STRING): BOOLEAN is
         -- Does any prefix of string `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matches_against(s, cpatt, Reg_prefix)
      end

   matches_suffix (s: STRING): BOOLEAN is
         -- Does any suffix of string `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         if cpatt_b = Void then
            compile_backw
         end
         Result := matches_against(s, cpatt_b, Reg_suffix)
      end

   matches_left_part (s: STRING): BOOLEAN is
         -- Does any substring of `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matches_against(s, cpatt, 0)
      end

   matches_right_part (s: STRING): BOOLEAN is
         -- Does any substring of `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         if cpatt_b = Void then
            compile_backw
         end
         Result := matches_against(s, cpatt_b, 0)
      end

feature {NONE}  -- Implementation

   cpatt, cpatt_b, saved_cpatt, saved_cpatt_b: STRING    -- Compiled patterns

   start_loc, end_loc: ARRAY[INTEGER]     -- Location arrays

   wipe_out is
      do
         cpatt := Void
         cpatt_b := Void
         saved_cpatt := Void
         saved_cpatt_b := Void
         compiled := false
      end

   compile is
         -- Compile `pattern'.
      do
         !!cpatt.make_filled ('%U', 8 + pattern.count * 5) -- heuristic value
         error_code := reg_compile(pattern.to_c, pattern.count,
                  cpatt.to_c, cpatt.count,
               insensitive, $number_of_subexpr)
         if error_code < 0 then
            cpatt.resize(- error_code);
            error_code := reg_compile(pattern.to_c, pattern.count,
                     cpatt.to_c, cpatt.count,
                     insensitive, $number_of_subexpr)
         end
         if error_code = 0 then
            compiled := true;
            if start_loc = Void  then
               !!start_loc.make(0, number_of_subexpr)
               !!end_loc.make(0, number_of_subexpr)
            elseif start_loc.upper < number_of_subexpr then
               start_loc.resize(0, number_of_subexpr)
               end_loc.resize(0, number_of_subexpr)
            end
         else
            compiled := false;
            cpatt := Void;
         end
      end

   compile_backw is
         -- Reverse the compiled pattern `cpatt', so it can be used in a
         -- backward search. A forward-compiled pattern is always
         -- reversable.
      require
         pattern_compiled: compiled;
      do
         !!cpatt_b.make_filled ('%U', cpatt.count + 9)  -- heuristic value
         error_code := reg_reverse(cpatt.to_c, cpatt_b.to_c, cpatt_b.count)
         if error_code < 0 then
            cpatt_b.resize(- error_code);
            error_code := reg_reverse(cpatt.to_c, cpatt_b.to_c, cpatt_b.count)
         end
         if error_code /= 0 then
            -- this should never occur
            compiled := false;
            cpatt_b := Void
         end
      end


   search(s: STRING; s_from, s_to: INTEGER; code: like cpatt) is
         -- Search in `s' for pattern `code'.
      local
         flag: INTEGER
      do
         if multi_line then
            flag := Reg_multi_line
         end
         if reg_execute(code.to_c, s.to_c, s.count,
                  s_from, s_to,
                    start_loc.to_c, end_loc.to_c, start_loc.count,
                    flag) then
            found := true
            start_position := start_loc.item(0)
            end_position := end_loc.item(0)
         else
            found := false
         end
      end


   match_against(s: STRING; code: like cpatt; flag: INTEGER) is
         -- Match `s' against the compiled pattern in `code'.
      local
         flags: INTEGER
      do
         if multi_line then
            flags := flag + Reg_multi_line
         else
            flags := flag
         end
         if reg_execute(code.to_c, s.to_c, s.count,
                  1, s.count, start_loc.to_c,
               end_loc.to_c, start_loc.count, flags) then
            found := true
            start_position := start_loc.item(0)
            end_position := end_loc.item(0)
         else
            found := false
         end
      end


   matches_against(s: STRING; code: like cpatt; flag: INTEGER): BOOLEAN is
         -- Does `s' match against the compiled pattern in `code'?
      local
         flags: INTEGER
      do
         if multi_line then
            flags := flag + Reg_multi_line
         else
            flags := flag
         end
         if reg_execute(code.to_c, s.to_c, s.count,
                  1, s.count, default_pointer,
               default_pointer, 0, flags) then
            Result := true
         end
      end


feature {NONE} -- Externals

   Reg_prefix: INTEGER is
         -- Flag for anchoring a pattern to beginning of a string
      external
         "C"
      alias
         "eif_match_MATCH_PREFIX"
      end

   Reg_suffix: INTEGER is
         -- Flag for anchoring a pattern to end of a string
      external
         "C"
      alias
         "eif_match_MATCH_SUFFIX"
      end

   Reg_multi_line: INTEGER is
         -- Flag for multi-line mode
      external
         "C"
      alias
         "eif_match_MULTI_LINE"
      end

   reg_compile(patt: POINTER; plen: INTEGER; code: POINTER; clen: INTEGER;
            ignore: BOOLEAN; nsubs: POINTER): INTEGER is
         -- Compile pattern `patt', which has length `plen' and store the
         -- result in `code', which has length `clen'. `ignore' says,
         -- if we want to be case insensitive. `nsubs' can be used to get
         -- the number of subexpressions back.
      external
         "C"
      alias
         "eif_match_reg_compile"
      end

   reg_reverse(code, rev_code: POINTER; rev_len: INTEGER): INTEGER is
         -- Reverse the compiled pattern in `code' and store the result
         -- in `rev_code', which has length `rev_len'.
      external
         "C"
      alias
         "eif_match_reg_reverse"
      end

   reg_execute(code, str: POINTER; strlen, s_from, s_to: INTEGER;
            startloc, endloc: POINTER; lenloc, flags: INTEGER): BOOLEAN is
         -- Match compiled pattern in `code' against string `str', which
         -- is of length `strlen'. The relevant part for the matching is
         -- between indices `s_from' and `s_to', the matching positions are
         -- stored in `startloc' and `endloc', which are integer arrays of
         -- size `lenloc'. Flags can be used to anchor the pattern to the
         -- beginning or end of the string or to switch between multi- and
         -- single-line mode.
      external
         "C"
      alias
         "eif_match_reg_execute"
      end

end  -- class REGEXP
