indexing

	description:
		"General concept of a searching machine."

deferred class SEARCHER

feature -- Initialization

	make (p: STRING) is
			-- Make a searcher from search pattern `p'. 
			-- Case sensitivity (which is enabled at object creation) 
			-- is not modified.
		require
			pattern_defined: p /= Void
		do
			if pattern = Void or else not p.is_equal(pattern) then
				pattern := clone(p)
				wipe_out
				compile
			end
		end

	make_sensitive (p: STRING) is
			-- Make a case sensitive searcher from search pattern `p'.
		require
			pattern_defined: p /= Void
		do
			if pattern = Void or else not p.is_equal(pattern) then
				insensitive := false
				make(p)
			else
				set_sensitive
			end
		ensure
			case_sensitivity: sensitive
		end

	make_insensitive (p: STRING) is
			-- Make a case insensitive searcher from search pattern `p'.
		require
			pattern_defined: p /= Void
		do
			if pattern = Void or else not p.is_equal(pattern) then
				insensitive := true
				make(p)
			else
				set_insensitive
			end
		ensure
			case_insensitivity: insensitive
		end


feature -- Modification

	set_sensitive is
			-- Make searcher case-sensitive.
		deferred
		ensure
			case_sensitivity: sensitive
		end

	set_insensitive is
			-- Make searcher case-insensitive.
		deferred
		ensure
			case_insensitivity: insensitive
		end

feature -- Status report

	pattern: STRING			
			-- The search pattern

	compiled: BOOLEAN		
			-- Is the pattern compiled successfully?

    error_code: INTEGER
            -- If not compiled, reason why not, else 0.

	insensitive: BOOLEAN	
			-- Is the searcher case insensitive?

	sensitive: BOOLEAN is
			-- Is the searcher case insensitive?
		do
			Result := not insensitive
		end

	found: BOOLEAN			
			-- Was last search successful?

	start_position, end_position: INTEGER
			-- If last search was successful, position of the pattern in the
			-- string. If it was not successful, the old values are kept.

feature -- Actions

	search_forward(s: STRING; s_from, s_to: INTEGER) is
			-- Search for `pattern' in string `s'. Start searching at index
			-- position `s_from' and search until the pattern matches (while
			-- ignoring any characters behind `s_to'). If pattern matches, set
			-- `found' to true and `start_position' and `end_position' 
			-- to the corresponding index positions (relative to the beginning
			-- of the string). If the pattern doesn't match, set `found' to
			-- false, but do not modify `start_position' or `end_position'.
		require
			s_exists: s /= Void;
			from_in_range: s_from >=1 and s_from <= s.count;
			to_in_range: s_to >= 1 and s_to <= s.count;
			pattern_compiled: compiled;
		deferred
		ensure
			position_set: found implies 
					(start_position >= s_from and start_position <= s_to + 1 and
					 end_position >= s_from - 1 and end_position <= s_to)
			position_kept: (not found) implies
					(start_position = old start_position and
					 end_position = old end_position)
		end

	search_backward(s: STRING; s_from, s_to: INTEGER) is
			-- Search for `pattern' in string `s'. Start searching at index
			-- position `s_to' and search backward until the pattern matches 
			-- (ignoring any characters before `s_from'). If pattern matches, 
			-- set `found' to true and `start_position' and `end_position' 
			-- to the corresponding index positions (relative to the beginning
			-- of the string). If the pattern doesn't match, set `found' to
			-- false, but do not modify `start_position' or `end_position'.
		require
			s_exists: s /= Void;
			from_in_range: s_from >=1 and s_from <= s.count;
			to_in_range: s_to >= 1 and s_to <= s.count;
			pattern_compiled: compiled;
		deferred
		ensure
			position_set: found implies 
					(start_position >= s_from and start_position <= s_to + 1 and
					 end_position >= s_from - 1 and end_position <= s_to)
			position_kept: (not found) implies
					(start_position = old start_position and
					 end_position = old end_position)
		end

feature {NONE} -- Implementation

	wipe_out is
			-- Clear the searcher.
		deferred
		end

	compile is
			-- Compile the pattern.
		require
			pattern_exists: pattern /= Void
		deferred
		end

end -- class SEARCHER
