indexing

	description:
		"An object which splits a string into parts."

class SPLITTER

inherit

	ARRAY [STRING]
		rename
			make as array_make
		end

creation
	make

feature -- Initialization

	make (sep_searcher: SEARCHER) is
			-- Make a splitter which splits strings at separators
			-- found by `sep_searcher'.
		require
			searcher_defined: sep_searcher /= Void
			searcher_ready: sep_searcher.compiled
		do
			separation_searcher := sep_searcher;
			array_make (lower, lower - 1);
		ensure
			searcher_set: separation_searcher = sep_searcher
		end

	separation_searcher: SEARCHER
			-- The searcher that finds the separators.

	ignore_separators (leading, trailing: BOOLEAN) is
			-- Ignore or recognize leading (trailing) separators. If not
			-- ignored and input string contains a leading (trailing)
			-- separator, an empty string will be assumed for the first
			-- (last) position.
		do
			leading_separator_ignored := leading
			trailing_separator_ignored := trailing
		end

	leading_separator_ignored, trailing_separator_ignored: BOOLEAN
			-- Shall leading / trailing separators be ignored?

feature -- Action

	split (s: STRING) is
			-- Split `s' into parts. Each part is accessible by its index
			-- position. The lower index position is 0, unless explicitely
			-- specified by `array_make'.
		require
			s_exists: s /= Void
		do
			split_substring (s, 1, s.count)
		end

	split_substring (s: STRING; s_from, s_to: INTEGER) is
			-- Split a substring of `s' into parts. The relevant part of `s'
			-- is denoted by the index positions `s_from' and `s_to'.
			-- Leading / trailing separators are taken into account with
			-- respect to the substring.
		require
			s_exists: s /= Void
			valid_range: s_from >= 1 and s_to <= s.count and s_from <= s_to + 1
		local
			s_f, i: INTEGER
			stop: BOOLEAN
		do
			i := lower
			if s_from > s_to then
				-- Nothing to split, so do nothing
			else
				separation_searcher.search_forward (s, s_from, s_to)
				if not separation_searcher.found then
					-- Exactly one part
					force (s.substring (s_from, s_to), i)
					i := i + 1;
				else
					if separation_searcher.start_position > s_from or else
							not leading_separator_ignored then
						force (substring (s, s_from,
									separation_searcher.start_position - 1), i)
						i := i + 1;
					end
					from
						s_f := separation_searcher.end_position + 1
						if s_f <= s_to then
							separation_searcher.search_forward (s, s_f, s_to)
						else
							stop := true
						end
					until
						stop or else not separation_searcher.found
					loop
						force (substring (s, s_f,
									separation_searcher.start_position - 1), i)
						i := i + 1;
						s_f := separation_searcher.end_position + 1
						if s_f <= s_to then
							separation_searcher.search_forward (s, s_f, s_to)
						else
							stop := true
						end
					end
					if separation_searcher.end_position < s_to or else
							not trailing_separator_ignored then
						force (substring (s, s_f, s_to), i)
						i := i + 1;
					end
				end
			end
			count := i - lower
		end

feature {NONE} -- Implementation

	substring (s: STRING; s_from, s_to: INTEGER): STRING is
		do
			if s_from > s_to then
				Result := ""
			else
				Result := s.substring (s_from, s_to)
			end
		end

end -- class SPLITTER
