indexing

   description:
      "Wildcard matcher"

class WILDCARD

inherit
   MATCHER
   REGEXP_ERRORS

creation

   make, make_sensitive, make_insensitive

feature -- Modification

   set_sensitive is
         -- Make the matcher case-sensitive.
      do
         insensitive := false
         matcher.set_sensitive
      end

   set_insensitive is
         -- Make the matcher case-insensitive.
      do
         insensitive := true
         matcher.set_insensitive
      end

   set_multi_line is
         -- Set matcher in multi-line mode. Mode is kept until matcher is
         -- explicitely set to single-line mode.
      do
         matcher.set_multi_line
      end

   set_single_line is
         -- Set matcher in single-line mode. Mode is kept until matcher is
         -- explicitely set to multi-line mode.
      do
         matcher.set_single_line
      end

feature -- Status report

   multi_line: BOOLEAN is
         -- Is matcher in multi-line mode?
      do
         Result := matcher.multi_line
      end

   single_line: BOOLEAN is
         -- Is matcher in single-line mode? This is the default on object
         -- creation.
      do
         Result := not multi_line
      end

feature  -- Searching

   search_forward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_from' and search until the pattern matches (while
            -- ignoring any characters behind `s_to'). If pattern matches, set
            -- `found' to true and `start_position' and `end_position'
            -- to the corresponding index positions (relative to the beginning
            -- of the string). If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position' or `end_position'.
      do
         matcher.search_forward(s, s_from, s_to)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

   search_backward(s: STRING; s_from, s_to: INTEGER) is
            -- Search for `pattern' in string `s'. Start searching at index
            -- position `s_to' and search backward until the pattern matches
            -- (ignoring any characters before `s_from'). If pattern matches,
            -- set `found' to true and `start_position' and `end_position'
            -- to the corresponding index positions (relative to the beginning
            -- of the string). If the pattern doesn't match, set `found' to
            -- false, but do not modify `start_position' or `end_position'.
      do
         matcher.search_backward(s, s_from, s_to)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

feature -- Matching actions

   match (s: STRING) is
            -- Match `pattern' against string `s'. The whole string must
            -- match. Set attributes `found' and - if match succeeded -
            -- `start_position' and `end_position'.
      do
         matcher.match(s)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

   match_prefix (s: STRING) is
         -- Match `pattern' against any prefix of string `s'.
         -- Matching is forward in `s'. Set attributes `found' and -
         -- if match succeeded - `start_position' and `end_position'.
      do
         matcher.match_prefix(s)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

   match_suffix (s: STRING) is
         -- Match `pattern' against any suffix of string `s'.
            -- Matching is backward in `s'. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'.
      do
         matcher.match_suffix(s)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

   match_left_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
         -- Matching is from left to right. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'.
      do
         matcher.match_left_part(s)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

   match_right_part (s: STRING) is
            -- Match `pattern' against any substring of string `s'.
         -- Matching is from right to left. Set attributes `found' and -
            -- if match succeeded - `start_position' and `end_position'.
      do
         matcher.match_right_part(s)
         if matcher.found then
            found := true;
            start_position := matcher.start_position
            end_position := matcher.end_position
         else
            found := false
         end
      end

feature -- Matching questions

   matches (s: STRING): BOOLEAN is
            -- Does string `s' matches with pattern? The whole string must
         -- match. None of the position attributes nor `found' is modified.
      do
         Result := matcher.matches(s)
      end

   matches_prefix (s: STRING): BOOLEAN is
         -- Does any prefix of string `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matcher.matches_prefix(s)
      end

   matches_suffix (s: STRING): BOOLEAN is
         -- Does any suffix of string `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matcher.matches_suffix(s)
      end

   matches_left_part (s: STRING): BOOLEAN is
         -- Does any substring of `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matcher.matches_left_part(s)
      end

   matches_right_part (s: STRING): BOOLEAN is
         -- Does any substring of `s' matches with pattern?
         -- Neither a position attribute nor `found' is modified.
      do
         Result := matcher.matches_right_part(s)
      end

feature {NONE} -- Implementation

   matcher: REGEXP

   wipe_out is
      do
         compiled := false
      end

   compile is
         -- Compile `pattern' into a r.e. pattern.
      local
         re_pattern: STRING
         plen, i, j: INTEGER
         stopped: BOOLEAN
         c: CHARACTER
      do
         compiled := false
         error_code := 0
         plen := pattern.count
         !!re_pattern.make_empty -- make (2 * plen)
         from
            i := 1
         until
            i > plen
         loop
            c := pattern.item(i)
            inspect
               c
            when '*' then
               re_pattern.append_string(".*")
            when '?' then
               re_pattern.append_character('.')
            when '{' then
               re_pattern.append_character('(')
            when '}' then
               re_pattern.append_character(')')
            when ',' then
               re_pattern.append_character('|')
            when '^', '$', '+', '(', ')', '.', '<', '>' then
               re_pattern.append_character('\')
               re_pattern.append_character(c)
            when '\' then
               re_pattern.append_character('\')
               if i < plen then
                  i := i + 1
                  re_pattern.append_character(pattern.item(i))
               end
            when '[' then
               re_pattern.append_character('[')
               from
                  stopped := false
                  j := i + 1
               until
                  j > plen or stopped
               loop
                  c := pattern.item(j)
                  inspect
                     c
                  when '\' then
                     re_pattern.append_character('\')
                     if j < plen then
                        j := j + 1
                        re_pattern.append_character(pattern.item(j))
                     end
                  when ']' then
                     if j > i + 2 or else pattern.item(i + 1) /= '^' then
                        stopped := true
                     else
                        re_pattern.append_character('\')
                     end
                     re_pattern.append_character(']')
                  when '{', '}', '$', '+', '(', ')', '.', '<', '>' then
                     re_pattern.append_character('\')
                     re_pattern.append_character(c)
                  else
                     re_pattern.append_character(c)
                  end
                  j := j + 1
               end
               if not stopped then
                  error_code := Re_unmatched_brackets
               end
               i := j - 1
            else  -- outer inspect
               re_pattern.append_character(c)
            end
            i := i + 1
         end
         if error_code = 0 then
            if matcher = Void then
               if insensitive then
                  !!matcher.make_insensitive(re_pattern)
               else
                  !!matcher.make_sensitive(re_pattern)
               end
            else
               matcher.make(re_pattern)
            end
            error_code := matcher.error_code
            if error_code = 0 then
               compiled := true
            end
         end
      end

end  -- class WILDCARD
