indexing

   title:       "Date class implementation";
   cluster:     "timedate";
   project:     "The Universal Time & Date UTDC";
   date:        "$Date: 2005/02/01 10:21:55 $";
   revision:    "$Revision: 1.1 $";
   copyright:   "Copyright (C) 1996-2005 Object Tools Group";
   license:     "http://visual-eiffel.com/license";
   disclaimer:  "[
      This program is free software; you can redistribute it and/or modify it 
      under the terms of the GNU General Public License as published by the 
      Free Software Foundation; either version 2 of the License, or (at your 
      option) any later version. This program is distributed in the hope that 
      it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
      See the GNU General Public License for more details. 
      You should have received a copy of the GNU General Public License along 
      with this program; if not, write to the Free Software Foundation, Inc., 
      59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   ]";
   restrictions: "[
      if you want to use this program under a different license type, contact 
      Object Tools http://www.object-tools.com
   ]";

class DATE

inherit

   SIMPLE_TIMESTAMP
      redefine
         out
      end

creation

   clear,
   set_date,
   set_from_absolute_days,
   set_to_now,
   set_local_time

feature -- Queries: change-over date
        -- This is the date when Julian calendar was switched to Gregorian one
        -- The date is different for different countries
        -- Here it is the date used by Pope Gregory XIII

   change_over_year: INTEGER is 1582
         -- Year of change-over date

   change_over_month: INTEGER is 10
         -- Month of change-over date

   change_over_day: INTEGER is 15
         -- Day of change-over date

feature -- Status

   absolute_days: INTEGER is
         -- Julian days: the number of days since January 1, 4713 B.C.
      local
         years: INTEGER
         months: INTEGER
         days: INTEGER
      do
         years := year
         months := month
         days := day
         if years < change_over_year or else
            (years = change_over_year and then
             (months < change_over_month or else
              (months = change_over_month and then days < change_over_day)))
         then
            -- Julian calendar
            Result := tt.julian_date_days (years, months, days)
         else
            -- Gregorian calendar
            Result := tt.gregorian_date_days (years, months, days)
         end
      ensure
         Result >= 0
      end

   set_date (yr, mh, dy: INTEGER) is
         -- Set the date to the given value
      require
         date_valid: tt.valid_date (yr, mh, dy)
      do
         year := yr
         month := mh
         day := dy
      ensure
         date_set: year = yr and month = mh and day = dy
      end

   set_from_absolute_days (i: INTEGER) is
      require
         absolute_days_valid: i >= 0
      local
         years: INTEGER
         months: INTEGER
         century: INTEGER
         temp: INTEGER
         day_of_year: INTEGER
      do
         if i < 2_299_161 then
            -- Julian calendar

            temp := (i + tt.Julian_absolute_days_offset) * 4 - 1

            -- Calculate the year and day of year (1 <= day_of_year <= 366)
            years := temp // tt.Days_per_4_years
         else
            -- Gregorian calendar

            temp := (i + tt.Gregorian_absolute_days_offset) * 4 - 1

            -- Calculate the century (year/100)
            century := temp // tt.Days_per_400_years

            -- Calculate the year and day of year (1 <= day_of_year <= 366)
            temp := ((temp \\ tt.Days_per_400_years) // 4) * 4 + 3
            years := (century * 100) + (temp // tt.Days_per_4_years)
         end

         day_of_year := (temp \\ tt.Days_per_4_years) // 4 + 1

         -- Calculate the month and day of month
         temp := day_of_year * 5 - 3
         months := temp // tt.Days_per_5_months
         day := (temp \\ tt.Days_per_5_months) // 5 + 1

         -- Convert to the normal beginning of the year
         if months < 10 then
            months := months + 3
         else
            years := years + 1
            months := months - 9
         end

         -- Adjust to the B.C./A.D. type numbering
         years := years - 4800
         if years <= 0 then
            years := years - 1
         end

         year := years
         month := months
      ensure
         absolute_days_set: absolute_days = i
      end

   set_to_now is
         -- Set the current date
      local
         ts: TIMESTAMP
      do
         !!ts.set_to_now
         set_date (ts.year, ts.month, ts.day)
      end

   set_local_time is
         -- Set the current local date
      local
         ts: TIMESTAMP
      do
         !!ts.set_local_time
         set_date (ts.year, ts.month, ts.day)
      end

   clear is
      do
         set_from_absolute_days (0)
      end

feature -- Queries

   infix "<" (ts: like Current): BOOLEAN is
      do
         Result := absolute_days < ts.absolute_days
      end

   infix "+",
   plus (d: LONG_DURATION): like Current is
      require else
         date_valid: (absolute_days + d.as_days) > 0
      do
         !!Result.set_from_absolute_days (absolute_days + d.absolute_days);
      end

   infix "-",
   minus (d: LONG_DURATION): like Current is
      require else
         date_valid: (absolute_days - d.whole_days) > 0
      do
         !!Result.set_from_absolute_days (absolute_days - d.absolute_days);
      end

   difference (t: like Current): LONG_DURATION is
      do
         !!Result.set_days (absolute_days - t.absolute_days)
      end

   day_of_week: INTEGER is
      do
         Result := absolute_days \\ tt.days_in_week + 1
      end

   is_leap_year: BOOLEAN is
         -- Is the year leap?
      do
         Result := tt.is_leap_year (year)
      end

feature -- Status

   year: INTEGER

   month: INTEGER

   day: INTEGER

feature -- Status: compatibility features

   year_part: INTEGER is
      obsolete "This feature will be removed soon: use year instead"
         -- Year
      do
         Result := year
      end

   month_part: INTEGER is
      obsolete "This feature will be removed soon: use month instead"
         -- Month
      do
         Result := month
      end

   day_part: INTEGER is
      obsolete "This feature will be removed soon: use day instead"
         -- Day
      do
         Result := day
      end

feature -- Output

   out: STRING is
         -- New string containing printable represantation
         -- Format is "YYYY-MM-DD" or "YYYY-MM-DD B.C."
      local
         s: STRING
      do
         Result := "0000-00-00"
         s := year.abs.out
         Result.replace_substring (s, 5 - s.count, 4)
         s := month.out
         Result.replace_substring (s, 8 - s.count, 7)
         s := day.out
         Result.replace_substring (s, 11 - s.count, 10)
         if year < 0 then
            Result.append_string (" B.C.")
         end
      end

invariant

   valid_date: tt.valid_date (year, month, day)

end